//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/dynamicCompiler.h"

void
showOptions(const char *name,
            const StrVector *options)
{
Platform_msg("%s:\n",name);
for(intptr_t i=0,i_end=StrVector_size(options);i<i_end;++i)
  {
  Platform_msg("  %s\n",StrVector_get(options,i));
  }
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
const char *fnctName="testDynamicCompiler";
if(Platform_argc()>1)
  {
  fnctName=Platform_argv(1);
  }
DynamicCompiler dc;
DynamicCompiler_init(&dc);
Text code;
Text_init(&code);
Text_append(&code,
  "#include \"TransProg/platform.h\"\n"
  "EXTERN_C_FUNCTION EXPORT_SYMBOL\n"
  "int\n"
  "%s(const char *str)\n"
  "{\n"
  "Platform_msg(\"from %s ~~~~ %%s ~~~~\\n\",str);\n"
  "return (int)strlen(str);\n"
  "}\n",
  fnctName,fnctName);
Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
             "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
Platform_msg("%s",Text_get(&code));
Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
             "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
showOptions("cc",DynamicCompiler_cc(&dc));
showOptions("cflags_only_I",DynamicCompiler_cflags_only_I(&dc));
showOptions("cflags_only_D",DynamicCompiler_cflags_only_D(&dc));
showOptions("cflags_only_other",DynamicCompiler_cflags_only_other(&dc));
showOptions("libs_only_L",DynamicCompiler_libs_only_L(&dc));
showOptions("libs_only_l",DynamicCompiler_libs_only_l(&dc));
showOptions("libs_only_other",DynamicCompiler_libs_only_other(&dc));
showOptions("oflags",DynamicCompiler_oflags(&dc));
Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
             "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
PluginHandle plugin=DynamicCompiler_build(&dc,Text_get(&code));
if(DynamicCompiler_buildLog(&dc))
  {
  Platform_msg("%s\n",DynamicCompiler_buildLog(&dc));
  Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
               "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  }
Text_destroy(&code);
if(plugin)
  {
  void (*f1)(void);
  Plugin_function(&f1,plugin,"this_does_not_exist");
  if(f1)
    {
    (*f1)();
    }
  else
    {
    char msg[0x100];
    Plugin_lastErrorMsg(msg,0x100);
    Platform_msg("Plugin_function() failure: %s\n",msg);
    }
  int (*f2)(const char *);
  Plugin_function(&f2,plugin,fnctName);
  if(f2)
    {
    Platform_msg("%s() --> %d\n",
                 fnctName,(*f2)("Hello Dynamic World!"));
    }
  else
    {
    char msg[0x100];
    Plugin_lastErrorMsg(msg,0x100);
    Platform_msg("Plugin_function() failure: %s\n",msg);
    }
  Plugin_close(plugin);
  }
else
  {
  Platform_msg("code compilation failure\n");
  }
DynamicCompiler_destroy(&dc);
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
