//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/path.h"

void
showContent(const char *path,
            int indent)
{
char spc[0x100];
for(int i=0;i<indent;++i)
  {
  spc[i]=' ';
  }
spc[indent]='\0';
switch(Path_kind(path))
  {
  case PATH_FILE:
    {
    Platform_msg("%s%s FILE (size=%"F64"d)\n",
                 spc,path,Path_fileSize(path));
    break;
    }
  case PATH_DIR:
    {
    Platform_msg("%s%s DIRECTORY\n",spc,path);
    PathDir *dir=Path_openDir(path);
    if(dir)
      {
      for(;;)
        {
        const char *e=Path_readDir(dir);
        if(!e) break;
        char s[0x400];
        Path_join(s,0x400,path,e);
        showContent(s,indent+4);
        }
      Path_closeDir(dir);
      }
    break;
    }
  case PATH_SPECIAL:
    {
    Platform_msg("%s%s SPECIAL\n",spc,path);
    break;
    }
  case PATH_ERROR:
    {
    Platform_msg("%s%s ERROR\n",spc,path);
    break;
    }
  }
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
for(int i=0;i<4;++i)
  {
  char name[0x400];
  if(Path_createTmpDir(name,0x400,".","tmp_??????_dir"))
    {
    for(int j=0;j<5;++j)
      {
      char name2[0x400];
      Path_createTmpFile(name2,0x400,name,"tmp_??????_sub_file");
      }
    }
  Path_createTmpFile(name,0x400,".","tmp_??????_file");
  }
showContent(Platform_argc()>1 ? Platform_argv(1) : ".",0);
PathDir *dir=Path_openDir(".");
if(dir)
  {
  for(;;)
    {
    const char *e=Path_readDir(dir);
    if(!e) break;
    if(!strncmp(e,"tmp_",4))
      {
      Path_remove(e);
      }
    }
  Path_closeDir(dir);
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
