//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef SIMU_H
#define SIMU_H 1

#include "TransProg/scheduler.h"
#include "TransProg/random.h"
#include "TransProg/spatialUtils.h"
#include "TransProg/coordIndex.h"

#define DT 0.001
#define DIAMETER 1.0
// #define AVOID_RENDERING

typedef struct _MyEntity
{
Activity act;
SpinSynchro sync;
bool dead;
real64_t x,y,z;
real64_t sx,sy,sz;
intptr_t key;
} MyEntity;

//----------------------------------------------------------------------------

typedef struct _Cell
{
SpinSynchro sync;
PtrVector content;
} Cell;

INLINE
void
Cell_add(Cell *cell,
         void *element)
{
SpinSynchro_lockW(&cell->sync);
PtrVector_push(&cell->content,element);
SpinSynchro_unlockW(&cell->sync);
}

INLINE
void
Cell_remove(Cell *cell,
            void *element)
{
SpinSynchro_lockW(&cell->sync);
PtrVector_fastErase(&cell->content,
                    PtrVector_find(&cell->content,
                                   (PtrVector_CmpFnct)0,
                                   element));
SpinSynchro_unlockW(&cell->sync);
}

//----------------------------------------------------------------------------

typedef struct _Simulation
{
intptr_t entityCount;
int64_t stepCount;
bool neighbourAffinity;
bool spatialAffinity;
#if !defined AVOID_RENDERING
  bool needsRedraw[2];
#endif
bool mustQuit;
int numaMigration;
real64_t size;
SpatialAffinity spAff;
Scheduler sched;
SpatialGrid grid;
Cell *cells;
} Simulation;

extern Simulation simu;

#if !defined AVOID_RENDERING

void
Rendering_showEntity(MyEntity *e);

void
Rendering_activate(int cpuCtx,
                   void *data);

void
Rendering_init(void);

#endif

#endif // SIMU_H

//----------------------------------------------------------------------------
