//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/skeleton.h"
#include "TransGl/viewer.h"
#include "TransGl/glDraw.h"
#include "TransProg/r32Vector.h"
#include "TransProg/surface.h"

int seenCount;
real64_t xSeen,ySeen,zSeen;
Viewer *viewer;

Skeleton skel;
SpatialBasis location;
SpatialBasis *bases;
real64_t curTime;

bool animate;
bool rotateX;
bool rotateY;
bool rotateZ;
int recordW;
int recordH;
int recordCount;
uint8_t *recordStorage;
int highlight;

void
computeBoundingBox(BoundingBox *bboxOut)
{
BoundingBox_invalidate(bboxOut);
for(int i=0,i_end=Skeleton_nodeCount(&skel);i<i_end;++i)
  {
  real64_t x,y,z;
  SpatialBasis_getPosition(bases+i,&x,&y,&z);
  BoundingBox_addVertex(bboxOut,x,y,z);
  }
}

void
applyPose(void)
{
Platform_msg("frame %d\n",(int)(curTime/Skeleton_frameTime(&skel)));
SpatialBasis_setPosition(bases,0.0,0.0,0.0);
SpatialBasis_cancelRotation(bases);
Skeleton_applyTimeToNodes(&skel,0,curTime,bases);
}

void
drawSkel(const real32_t *projMtx,
         const real32_t *viewMtx)
{
static R32Vector lines;
static GlShape *boxSh=(GlShape *)0;
static GlShape *lineSh=(GlShape *)0;
if(!boxSh)
  {
  R32Vector_init(&lines);
  BoundingBox bbox;
  BoundingBox_init(&bbox);
  computeBoundingBox(&bbox);
  real64_t nodeSz=0.01*BoundingBox_maxSize(&bbox);
  int indexCount,vertexCount;
  int *indices;
  real32_t *vertices;
  real32_t *normals;
  Surface_allocBox(&indexCount,&indices,&vertexCount,&vertices,
                   &normals,(real32_t **)0,true,nodeSz,nodeSz,nodeSz);
  boxSh=GlShape_new(GLSHAPE_HAS_NORMALS,GL_TRIANGLES,
                    indexCount,indices,vertexCount,vertices,normals,
                    (const real32_t *)0,(const real32_t *)0);
  Memory_free(indices); Memory_free(vertices); Memory_free(normals);
  lineSh=GlShape_new(0,GL_LINES,0,(const int *)0,
                     2*Skeleton_nodeCount(&skel),(const real32_t *)0,
                     (const real32_t *)0,(const real32_t *)0,
                     (const real32_t *)0);
  }
real32_t ambient=0.2f;
real32_t specular=0.75f;
real32_t shininess=20.0f;
const real32_t *fog=(const real32_t *)0;
real32_t headLight[3]={0.0f,0.0f,0.0f};
real32_t boxRgba[4]={0.7f,0.4f,0.2f,1.0f};
real32_t lineRgba[4]={1.0f,1.0f,1.0f,1.0f};
real32_t highlightRgba[4]={0.1f,0.8f,0.1f,1.0f};
real32_t modelViewMtx[16];
Gfx_computeModelViewMatrix(modelViewMtx,viewMtx,location.matrix);
R32Vector_clear(&lines);
for(int i=0,i_end=Skeleton_nodeCount(&skel);i<i_end;++i)
  {
  real32_t nodeModelViewMtx[16];
  Gfx_computeModelViewMatrix(nodeModelViewMtx,modelViewMtx,bases[i].matrix);
  GlShape_draw(boxSh,
               boxRgba,ambient,specular,shininess,
               headLight,fog,0,projMtx,nodeModelViewMtx);
  if(i!=highlight)
    {
    int p=Skeleton_parent(&skel,i);
    if(p!=-1)
      {
      real64_t x,y,z;
      SpatialBasis_getPosition(bases+i,&x,&y,&z);
      R32Vector_push(&lines,(real32_t)x);
      R32Vector_push(&lines,(real32_t)y);
      R32Vector_push(&lines,(real32_t)z);
      SpatialBasis_getPosition(bases+p,&x,&y,&z);
      R32Vector_push(&lines,(real32_t)x);
      R32Vector_push(&lines,(real32_t)y);
      R32Vector_push(&lines,(real32_t)z);
      }
    }
  }
lineSh->vertexCount=R32Vector_size(&lines)/3;
glBindBuffer(GL_ARRAY_BUFFER,GlShape_vertexBuffer(lineSh));
glBufferSubData(GL_ARRAY_BUFFER,0,
                lineSh->vertexCount*3*sizeof(real32_t),
                lines.data);
glBindBuffer(GL_ARRAY_BUFFER,0);
glLineWidth(1.0);
GlShape_draw(lineSh,
             lineRgba,ambient,specular,shininess,
             headLight,fog,0,projMtx,modelViewMtx);
int p=Skeleton_parent(&skel,highlight);
if(p!=-1)
  {
  R32Vector_clear(&lines);
  real64_t x,y,z;
  SpatialBasis_getPosition(bases+highlight,&x,&y,&z);
  R32Vector_push(&lines,(real32_t)x);
  R32Vector_push(&lines,(real32_t)y);
  R32Vector_push(&lines,(real32_t)z);
  SpatialBasis_getPosition(bases+p,&x,&y,&z);
  R32Vector_push(&lines,(real32_t)x);
  R32Vector_push(&lines,(real32_t)y);
  R32Vector_push(&lines,(real32_t)z);
  lineSh->vertexCount=R32Vector_size(&lines)/3;
  glBindBuffer(GL_ARRAY_BUFFER,GlShape_vertexBuffer(lineSh));
  glBufferSubData(GL_ARRAY_BUFFER,0,
                  lineSh->vertexCount*3*sizeof(real32_t),
                  lines.data);
  glBindBuffer(GL_ARRAY_BUFFER,0);
  glLineWidth(3.0);
  GlShape_draw(lineSh,
               highlightRgba,ambient,specular,shininess,
               headLight,fog,0,projMtx,modelViewMtx);
  }
}

void
viewerCB(Viewer *v)
{
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
real64_t d=seenCount ? 1.0/seenCount : 0.0;
Viewer_setRefDistanceFromPoint(v,d*xSeen,d*ySeen,d*zSeen);
bool close=false;
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"\x1B"))
        {
        close=true;
        }
      else if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"+"))
        {
        if(highlight<Skeleton_nodeCount(&skel)-1) ++highlight;
        Platform_msg("highlight=%d %s\n",highlight,
                     Skeleton_nodeName(&skel,highlight));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"-"))
        {
        if(highlight) --highlight;
        Platform_msg("highlight=%d %s\n",highlight,
                     Skeleton_nodeName(&skel,highlight));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"r"))
        {
        if(recordStorage)
          {
          Memory_free(recordStorage);
          recordStorage=(uint8_t *)0;
          }
        else
          {
          recordW=GfxState_width(state);
          recordH=GfxState_height(state);
          recordStorage=Memory_alloc(recordW*recordH*3,uint8_t);
          }
        }
      else if(!strcmp(GfxState_key(state),"a"))
        {
        animate=!animate;
        }
      else if(!strcmp(GfxState_key(state),"Right"))
        {
        curTime+=Skeleton_frameTime(&skel);
        if(curTime>=Skeleton_frameCount(&skel)*Skeleton_frameTime(&skel))
          {
          curTime=0.0;
          }
        applyPose();
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"Left"))
        {
        curTime-=Skeleton_frameTime(&skel);
        if(curTime<0.0)
          {
          curTime=Skeleton_frameTime(&skel)*(Skeleton_frameCount(&skel)-1);
          }
        applyPose();
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"x"))
        {
        rotateX=!rotateX;
        }
      else if(!strcmp(GfxState_key(state),"y"))
        {
        rotateY=!rotateY;
        }
      else if(!strcmp(GfxState_key(state),"z"))
        {
        rotateZ=!rotateZ;
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    close=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    xSeen=ySeen=zSeen=0.0;
    seenCount=0;
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    BoundingBox bbox;
    BoundingBox_init(&bbox);
    computeBoundingBox(&bbox);
    if(Perspective_boxVisibility(Viewer_perspective(v),
                                 Viewer_viewPoint(v),
                                 &bbox,&location,
                                 (SpatialTransform *)0))
      {
      ++seenCount;
      real64_t xc=BoundingBox_centerX(&bbox);
      real64_t yc=BoundingBox_centerY(&bbox);
      real64_t zc=BoundingBox_centerZ(&bbox);
      SpatialBasis_localToGlobalPosition(&location,&xc,&yc,&zc);
      xSeen+=xc;
      ySeen+=yc;
      zSeen+=zc;
      drawSkel(projMtx,viewMtx);
      }
    BoundingBox_destroy(&bbox);
    break;
    }
  default:
    {
    break;
    }
  }
if(close)
  {
  Viewer_delete(v);
  viewer=(Viewer *)0;
  }
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
if(argc<2)
  {
  char msg[0x100];
  sprintf(msg,"usage: %s skeleton_file\n",argv[0]);
  DROP_DEAD(msg);
  }
Skeleton_init(&skel);
if(!Skeleton_loadFile(&skel,argv[1]))
  {
  char msg[0x100];
  sprintf(msg,"Cannot load file %s\n",argv[1]);
  DROP_DEAD(msg);
  }
Platform_msg("%s nodeCount=%d channelCount=%d frameCount=%d\n",
             argv[1],Skeleton_nodeCount(&skel),
             Skeleton_channelCount(&skel),Skeleton_frameCount(&skel));
bases=Memory_alloc(Skeleton_nodeCount(&skel),SpatialBasis);
for(int i=0,i_end=Skeleton_nodeCount(&skel);i<i_end;++i)
  {
  SpatialBasis_init(bases+i);
  }
SpatialBasis_init(&location);
applyPose();
GlView_init();
GlDraw_init();
int w=640+14;
int h=480+32;
viewer=Viewer_new(GlView_screenWidth()-w,0,w,h,false,&viewerCB,(void *)0);
BoundingBox bbox;
BoundingBox_init(&bbox);
computeBoundingBox(&bbox);
real64_t maxSz=BoundingBox_maxSize(&bbox);
Platform_msg("sx=%g\n",BoundingBox_sizeX(&bbox));
Platform_msg("sy=%g\n",BoundingBox_sizeY(&bbox));
Platform_msg("sz=%g\n",BoundingBox_sizeZ(&bbox));
Perspective_setNearDistance(Viewer_perspective(viewer),0.1*maxSz);
Perspective_setFarDistance(Viewer_perspective(viewer),20.0*maxSz);
SpatialBasis_setPosition(Viewer_viewPoint(viewer),
                         BoundingBox_centerX(&bbox),
                         BoundingBox_centerY(&bbox),
                         BoundingBox_centerZ(&bbox));
SpatialBasis_translate(Viewer_viewPoint(viewer),-1.0*maxSz,0.0,0.0);
BoundingBox_destroy(&bbox);
int64_t lastTime=Platform_time();
for(;;)
  {
  int64_t dt=Platform_time()-lastTime;
  GlView_processEvents(dt<40000 ? 40000-dt : 0);
  GlView_sync();
  if(!viewer) break;
  dt=Platform_time()-lastTime;
  lastTime+=dt;
  if(animate)
    {
    curTime+=1e-6*dt;
    if(curTime>=Skeleton_frameCount(&skel)*Skeleton_frameTime(&skel))
      {
      curTime=0.0;
      }
    applyPose();
    GlView_postRedraw(Viewer_view(viewer));
    }
  if(rotateX)
    {
    SpatialBasis_roll(&location,0.01);
    GlView_postRedraw(Viewer_view(viewer));
    }
  if(rotateY)
    {
    SpatialBasis_pitch(&location,0.01);
    GlView_postRedraw(Viewer_view(viewer));
    }
  if(rotateZ)
    {
    SpatialBasis_yaw(&location,0.01);
    GlView_postRedraw(Viewer_view(viewer));
    }
  if(recordStorage)
    {
    char buffer[0x20];
    sprintf(buffer,"snap_%6.6d.tga",recordCount++);
    Platform_msg("%s\n",buffer);
    GlView_makeCurrentGlTarget(viewer->view);
    GlDraw_snapshot(0,0,recordW,recordH,recordStorage);
    Gfx_saveTgaImage(buffer,recordW,recordH,3,recordStorage);
    }
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
