//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSGL_VIEWER_H
#define TRANSGL_VIEWER_H 1

#include "TransGl/glView.h"

#include "TransProg/examiner.h"

typedef struct _Viewer Viewer;

typedef
void
(*ViewerCallback)(Viewer *v);

struct _Viewer
{
GlView *view;
ViewerCallback callback;
void *data;
SpatialBasis viewPoint;
Perspective perspective;
Examiner examiner;
real64_t refDistance;
bool right;
real32_t leftProjMtx[16],rightProjMtx[16];
real32_t leftViewMtx[16],rightViewMtx[16];
struct { int x,y,w,h; } viewport;
};

TRANSGL_API
void
Viewer_init(Viewer *v,
            int initX,
            int initY,
            int initWidth,
            int initHeight,
            bool borderless,
            ViewerCallback callback,
            void *data);

INLINE
Viewer *
Viewer_new(int initX,
           int initY,
           int initWidth,
           int initHeight,
           bool borderless,
           ViewerCallback callback,
           void *data)
{
Viewer *v=Memory_alloc(1,Viewer);
Viewer_init(v,initX,initY,initWidth,initHeight,borderless,callback,data);
return v;
}

TRANSGL_API
void
Viewer_destroy(Viewer *v);

INLINE
void
Viewer_delete(Viewer *v)
{
Viewer_destroy(v);
Memory_free(v);
}

INLINE
void *
Viewer_data(const Viewer *v)
{
return v->data;
}

INLINE
GlView *
Viewer_view(const Viewer *v)
{
return v->view;
}

INLINE
const GfxState *
Viewer_state(const Viewer *v)
{
return GlView_state(v->view);
}

INLINE
SpatialBasis *
Viewer_viewPoint(const Viewer *v)
{
return (SpatialBasis *)&v->viewPoint;
}

INLINE
Perspective *
Viewer_perspective(const Viewer *v)
{
return (Perspective *)&v->perspective;
}

INLINE
Examiner *
Viewer_examiner(const Viewer *v)
{
return (Examiner *)&v->examiner;
}

INLINE
void
Viewer_setRefDistance(Viewer *v,
                      real64_t refDistance)
{
v->refDistance=refDistance;
}

TRANSGL_API
void
Viewer_setRefDistanceFromPoint(Viewer *v,
                               real64_t px,
                               real64_t py,
                               real64_t pz);

INLINE
real64_t
Viewer_refDistance(const Viewer *v)
{
return v->refDistance;
}

INLINE
const real32_t *
Viewer_projectionMatrix(const Viewer *v)
{
return v->right ? v->rightProjMtx : v->leftProjMtx;
}

INLINE
const real32_t *
Viewer_viewMatrix(const Viewer *v)
{
return v->right ? v->rightViewMtx : v->leftViewMtx;
}

INLINE
void
Viewer_viewport(const Viewer *v,
                int *out_x,
                int *out_y,
                int *out_width,
                int *out_height)
{
*out_x=v->viewport.x;
*out_y=v->viewport.y;
*out_width=v->viewport.w;
*out_height=v->viewport.h;
}

#endif // TRANSGL_VIEWER_H

//----------------------------------------------------------------------------
