//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_ASYNCCALL_H
#define TRANSPROG_ASYNCCALL_H 1

#include "TransProg/memory.h"

typedef struct _AsyncCall AsyncCall;

#define AsyncCall_typedef(rt,name)                \
        typedef struct _##name                    \
        {                                         \
        void (*stub)(struct _##name *);           \
        rt (*fnct)(void);                         \
        rt result;                                \
        } name;                                   \
        INLINE                                    \
        void                                      \
        _##name##_stub(name *ac)                  \
        {                                         \
        ac->result=(*ac->fnct)();                 \
        }                                         \
        INLINE                                    \
        name *                                    \
        name##_new(rt (*fnct)(void))              \
        {                                         \
        name *ac=Memory_alloc(1,name);            \
        ac->stub=_##name##_stub;                  \
        ac->fnct=fnct;                            \
        memset(&ac->result,0,sizeof(ac->result)); \
        return ac;                                \
        }

#define AsyncCall_typedef_void(name)    \
        typedef struct _##name          \
        {                               \
        void (*stub)(struct _##name *); \
        void (*fnct)(void);             \
        } name;                         \
        INLINE                          \
        void                            \
        _##name##_stub(name *ac)        \
        {                               \
        (*ac->fnct)();                  \
        }                               \
        INLINE                          \
        name *                          \
        name##_new(void (*fnct)(void))  \
        {                               \
        name *ac=Memory_alloc(1,name);  \
        ac->stub=_##name##_stub;        \
        ac->fnct=fnct;                  \
        return ac;                      \
        }

#define AsyncCall_typedef_1(rt,name,              \
                            t1)                   \
        typedef struct _##name                    \
        {                                         \
        void (*stub)(struct _##name *);           \
        rt (*fnct)(t1);                           \
        rt result;                                \
        t1 p1;                                    \
        } name;                                   \
        INLINE                                    \
        void                                      \
        _##name##_stub(name *ac)                  \
        {                                         \
        ac->result=(*ac->fnct)(ac->p1);           \
        }                                         \
        INLINE                                    \
        name *                                    \
        name##_new(rt (*fnct)(t1),                \
                   t1 p1)                         \
        {                                         \
        name *ac=Memory_alloc(1,name);            \
        ac->stub=_##name##_stub;                  \
        ac->fnct=fnct;                            \
        memset(&ac->result,0,sizeof(ac->result)); \
        ac->p1=p1;                                \
        return ac;                                \
        }

#define AsyncCall_typedef_1_void(name,     \
                                 t1)       \
        typedef struct _##name             \
        {                                  \
        void (*stub)(struct _##name *);    \
        void (*fnct)(t1);                  \
        t1 p1;                             \
        } name;                            \
        INLINE                             \
        void                               \
        _##name##_stub(name *ac)           \
        {                                  \
        (*ac->fnct)(ac->p1);               \
        }                                  \
        INLINE                             \
        name *                             \
        name##_new(void (*fnct)(t1),       \
                   t1 p1)                  \
        {                                  \
        name *ac=Memory_alloc(1,name);     \
        ac->stub=_##name##_stub;           \
        ac->fnct=fnct;                     \
        ac->p1=p1;                         \
        return ac;                         \
        }

#define AsyncCall_typedef_2(rt,name,              \
                            t1,t2)                \
        typedef struct _##name                    \
        {                                         \
        void (*stub)(struct _##name *);           \
        rt (*fnct)(t1,t2);                        \
        rt result;                                \
        t1 p1; t2 p2;                             \
        } name;                                   \
        INLINE                                    \
        void                                      \
        _##name##_stub(name *ac)                  \
        {                                         \
        ac->result=(*ac->fnct)(ac->p1,ac->p2);    \
        }                                         \
        INLINE                                    \
        name *                                    \
        name##_new(rt (*fnct)(t1,t2),             \
                   t1 p1,t2 p2)                   \
        {                                         \
        name *ac=Memory_alloc(1,name);            \
        ac->stub=_##name##_stub;                  \
        ac->fnct=fnct;                            \
        memset(&ac->result,0,sizeof(ac->result)); \
        ac->p1=p1; ac->p2=p2;                     \
        return ac;                                \
        }

#define AsyncCall_typedef_2_void(name,  \
                                 t1,t2) \
        typedef struct _##name          \
        {                               \
        void (*stub)(struct _##name *); \
        void (*fnct)(t1,t2);            \
        t1 p1; t2 p2;                   \
        } name;                         \
        INLINE                          \
        void                            \
        _##name##_stub(name *ac)        \
        {                               \
        (*ac->fnct)(ac->p1,ac->p2);     \
        }                               \
        INLINE                          \
        name *                          \
        name##_new(void (*fnct)(t1,t2), \
                   t1 p1,t2 p2)         \
        {                               \
        name *ac=Memory_alloc(1,name);  \
        ac->stub=_##name##_stub;        \
        ac->fnct=fnct;                  \
        ac->p1=p1; ac->p2=p2;           \
        return ac;                      \
        }

#define AsyncCall_typedef_3(rt,name,                  \
                            t1,t2,t3)                 \
        typedef struct _##name                        \
        {                                             \
        void (*stub)(struct _##name *);               \
        rt (*fnct)(t1,t2,t3);                         \
        rt result;                                    \
        t1 p1; t2 p2; t3 p3;                          \
        } name;                                       \
        INLINE                                        \
        void                                          \
        _##name##_stub(name *ac)                      \
        {                                             \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3); \
        }                                             \
        INLINE                                        \
        name *                                        \
        name##_new(rt (*fnct)(t1,t2,t3),              \
                   t1 p1,t2 p2,t3 p3)                 \
        {                                             \
        name *ac=Memory_alloc(1,name);                \
        ac->stub=_##name##_stub;                      \
        ac->fnct=fnct;                                \
        memset(&ac->result,0,sizeof(ac->result));     \
        ac->p1=p1; ac->p2=p2; ac->p3=p3;              \
        return ac;                                    \
        }

#define AsyncCall_typedef_3_void(name,      \
                                 t1,t2,t3)  \
        typedef struct _##name              \
        {                                   \
        void (*stub)(struct _##name *);     \
        void (*fnct)(t1,t2,t3);             \
        t1 p1; t2 p2; t3 p3;                \
        } name;                             \
        INLINE                              \
        void                                \
        _##name##_stub(name *ac)            \
        {                                   \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3);  \
        }                                   \
        INLINE                              \
        name *                              \
        name##_new(void (*fnct)(t1,t2,t3),  \
                   t1 p1,t2 p2,t3 p3)       \
        {                                   \
        name *ac=Memory_alloc(1,name);      \
        ac->stub=_##name##_stub;            \
        ac->fnct=fnct;                      \
        ac->p1=p1; ac->p2=p2; ac->p3=p3;    \
        return ac;                          \
        }

#define AsyncCall_typedef_4(rt,name,                         \
                            t1,t2,t3,t4)                     \
        typedef struct _##name                               \
        {                                                    \
        void (*stub)(struct _##name *);                      \
        rt (*fnct)(t1,t2,t3,t4);                             \
        rt result;                                           \
        t1 p1; t2 p2; t3 p3; t4 p4;                          \
        } name;                                              \
        INLINE                                               \
        void                                                 \
        _##name##_stub(name *ac)                             \
        {                                                    \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4); \
        }                                                    \
        INLINE                                               \
        name *                                               \
        name##_new(rt (*fnct)(t1,t2,t3,t4),                  \
                   t1 p1,t2 p2,t3 p3,t4 p4)                  \
        {                                                    \
        name *ac=Memory_alloc(1,name);                       \
        ac->stub=_##name##_stub;                             \
        ac->fnct=fnct;                                       \
        memset(&ac->result,0,sizeof(ac->result));            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;          \
        return ac;                                           \
        }

#define AsyncCall_typedef_4_void(name,              \
                                 t1,t2,t3,t4)       \
        typedef struct _##name                      \
        {                                           \
        void (*stub)(struct _##name *);             \
        void (*fnct)(t1,t2,t3,t4);                  \
        t1 p1; t2 p2; t3 p3; t4 p4;                 \
        } name;                                     \
        INLINE                                      \
        void                                        \
        _##name##_stub(name *ac)                    \
        {                                           \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4);   \
        }                                           \
        INLINE                                      \
        name *                                      \
        name##_new(void (*fnct)(t1,t2,t3,t4),       \
                   t1 p1,t2 p2,t3 p3,t4 p4)         \
        {                                           \
        name *ac=Memory_alloc(1,name);              \
        ac->stub=_##name##_stub;                    \
        ac->fnct=fnct;                              \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4; \
        return ac;                                  \
        }

#define AsyncCall_typedef_5(rt,name,                        \
                            t1,t2,t3,t4,t5)                 \
        typedef struct _##name                              \
        {                                                   \
        void (*stub)(struct _##name *);                     \
        rt (*fnct)(t1,t2,t3,t4,t5);                         \
        rt result;                                          \
        t1 p1; t2 p2; t3 p3; t4 p4;                         \
        t5 p5;                                              \
        } name;                                             \
        INLINE                                              \
        void                                                \
        _##name##_stub(name *ac)                            \
        {                                                   \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4, \
                               ac->p5);                     \
        }                                                   \
        INLINE                                              \
        name *                                              \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5),              \
                   t1 p1,t2 p2,t3 p3,t4 p4,                 \
                   t5 p5)                                   \
        {                                                   \
        name *ac=Memory_alloc(1,name);                      \
        ac->stub=_##name##_stub;                            \
        ac->fnct=fnct;                                      \
        memset(&ac->result,0,sizeof(ac->result));           \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;         \
        ac->p5=p5;                                          \
        return ac;                                          \
        }

#define AsyncCall_typedef_5_void(name,              \
                                 t1,t2,t3,t4,t5)    \
        typedef struct _##name                      \
        {                                           \
        void (*stub)(struct _##name *);             \
        void (*fnct)(t1,t2,t3,t4,t5);               \
        t1 p1; t2 p2; t3 p3; t4 p4;                 \
        t5 p5;                                      \
        } name;                                     \
        INLINE                                      \
        void                                        \
        _##name##_stub(name *ac)                    \
        {                                           \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,    \
                    ac->p5);                        \
        }                                           \
        INLINE                                      \
        name *                                      \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5),    \
                   t1 p1,t2 p2,t3 p3,t4 p4,         \
                   t5 p5)                           \
        {                                           \
        name *ac=Memory_alloc(1,name);              \
        ac->stub=_##name##_stub;                    \
        ac->fnct=fnct;                              \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4; \
        ac->p5=p5;                                  \
        return ac;                                  \
        }

#define AsyncCall_typedef_6(rt,name,                        \
                            t1,t2,t3,t4,t5,t6)              \
        typedef struct _##name                              \
        {                                                   \
        void (*stub)(struct _##name *);                     \
        rt (*fnct)(t1,t2,t3,t4,t5,t6);                      \
        rt result;                                          \
        t1 p1; t2 p2; t3 p3; t4 p4;                         \
        t5 p5; t6 p6;                                       \
        } name;                                             \
        INLINE                                              \
        void                                                \
        _##name##_stub(name *ac)                            \
        {                                                   \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4, \
                               ac->p5,ac->p6);              \
        }                                                   \
        INLINE                                              \
        name *                                              \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6),           \
                   t1 p1,t2 p2,t3 p3,t4 p4,                 \
                   t5 p5,t6 p6)                             \
        {                                                   \
        name *ac=Memory_alloc(1,name);                      \
        ac->stub=_##name##_stub;                            \
        ac->fnct=fnct;                                      \
        memset(&ac->result,0,sizeof(ac->result));           \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;         \
        ac->p5=p5; ac->p6=p6;                               \
        return ac;                                          \
        }

#define AsyncCall_typedef_6_void(name,              \
                                 t1,t2,t3,t4,t5,t6) \
        typedef struct _##name                      \
        {                                           \
        void (*stub)(struct _##name *);             \
        void (*fnct)(t1,t2,t3,t4,t5,t6);            \
        t1 p1; t2 p2; t3 p3; t4 p4;                 \
        t5 p5; t6 p6;                               \
        } name;                                     \
        INLINE                                      \
        void                                        \
        _##name##_stub(name *ac)                    \
        {                                           \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,    \
                    ac->p5,ac->p6);                 \
        }                                           \
        INLINE                                      \
        name *                                      \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6), \
                   t1 p1,t2 p2,t3 p3,t4 p4,         \
                   t5 p5,t6 p6)                     \
        {                                           \
        name *ac=Memory_alloc(1,name);              \
        ac->stub=_##name##_stub;                    \
        ac->fnct=fnct;                              \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4; \
        ac->p5=p5; ac->p6=p6;                       \
        return ac;                                  \
        }

#define AsyncCall_typedef_7(rt,name,                        \
                            t1,t2,t3,t4,t5,t6,t7)           \
        typedef struct _##name                              \
        {                                                   \
        void (*stub)(struct _##name *);                     \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7);                   \
        rt result;                                          \
        t1 p1; t2 p2; t3 p3; t4 p4;                         \
        t5 p5; t6 p6; t7 p7;                                \
        } name;                                             \
        INLINE                                              \
        void                                                \
        _##name##_stub(name *ac)                            \
        {                                                   \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4, \
                               ac->p5,ac->p6,ac->p7);       \
        }                                                   \
        INLINE                                              \
        name *                                              \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7),        \
                   t1 p1,t2 p2,t3 p3,t4 p4,                 \
                   t5 p5,t6 p6,t7 p7)                       \
        {                                                   \
        name *ac=Memory_alloc(1,name);                      \
        ac->stub=_##name##_stub;                            \
        ac->fnct=fnct;                                      \
        memset(&ac->result,0,sizeof(ac->result));           \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;         \
        ac->p5=p5; ac->p6=p6; ac->p7=p7;                    \
        return ac;                                          \
        }

#define AsyncCall_typedef_7_void(name,                 \
                                 t1,t2,t3,t4,t5,t6,t7) \
        typedef struct _##name                         \
        {                                              \
        void (*stub)(struct _##name *);                \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7);            \
        t1 p1; t2 p2; t3 p3; t4 p4;                    \
        t5 p5; t6 p6; t7 p7;                           \
        } name;                                        \
        INLINE                                         \
        void                                           \
        _##name##_stub(name *ac)                       \
        {                                              \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,       \
                    ac->p5,ac->p6,ac->p7);             \
        }                                              \
        INLINE                                         \
        name *                                         \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7), \
                   t1 p1,t2 p2,t3 p3,t4 p4,            \
                   t5 p5,t6 p6,t7 p7)                  \
        {                                              \
        name *ac=Memory_alloc(1,name);                 \
        ac->stub=_##name##_stub;                       \
        ac->fnct=fnct;                                 \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;    \
        ac->p5=p5; ac->p6=p6; ac->p7=p7;               \
        return ac;                                     \
        }

#define AsyncCall_typedef_8(rt,name,                         \
                            t1,t2,t3,t4,t5,t6,t7,t8)         \
        typedef struct _##name                               \
        {                                                    \
        void (*stub)(struct _##name *);                      \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8);                 \
        rt result;                                           \
        t1 p1; t2 p2; t3 p3; t4 p4;                          \
        t5 p5; t6 p6; t7 p7; t8 p8;                          \
        } name;                                              \
        INLINE                                               \
        void                                                 \
        _##name##_stub(name *ac)                             \
        {                                                    \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,  \
                               ac->p5,ac->p6,ac->p7,ac->p8); \
        }                                                    \
        INLINE                                               \
        name *                                               \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8),      \
                   t1 p1,t2 p2,t3 p3,t4 p4,                  \
                   t5 p5,t6 p6,t7 p7,t8 p8)                  \
        {                                                    \
        name *ac=Memory_alloc(1,name);                       \
        ac->stub=_##name##_stub;                             \
        ac->fnct=fnct;                                       \
        memset(&ac->result,0,sizeof(ac->result));            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;          \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;          \
        return ac;                                           \
        }

#define AsyncCall_typedef_8_void(name,                    \
                                 t1,t2,t3,t4,t5,t6,t7,t8) \
        typedef struct _##name                            \
        {                                                 \
        void (*stub)(struct _##name *);                   \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8);            \
        t1 p1; t2 p2; t3 p3; t4 p4;                       \
        t5 p5; t6 p6; t7 p7; t8 p8;                       \
        } name;                                           \
        INLINE                                            \
        void                                              \
        _##name##_stub(name *ac)                          \
        {                                                 \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,          \
                    ac->p5,ac->p6,ac->p7,ac->p8);         \
        }                                                 \
        INLINE                                            \
        name *                                            \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8), \
                   t1 p1,t2 p2,t3 p3,t4 p4,               \
                   t5 p5,t6 p6,t7 p7,t8 p8)               \
        {                                                 \
        name *ac=Memory_alloc(1,name);                    \
        ac->stub=_##name##_stub;                          \
        ac->fnct=fnct;                                    \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;       \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;       \
        return ac;                                        \
        }

#define AsyncCall_typedef_9(rt,name,                         \
                            t1,t2,t3,t4,t5,t6,t7,t8,         \
                            t9)                              \
        typedef struct _##name                               \
        {                                                    \
        void (*stub)(struct _##name *);                      \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                  \
                   t9);                                      \
        rt result;                                           \
        t1 p1; t2 p2; t3 p3; t4 p4;                          \
        t5 p5; t6 p6; t7 p7; t8 p8;                          \
        t9 p9;                                               \
        } name;                                              \
        INLINE                                               \
        void                                                 \
        _##name##_stub(name *ac)                             \
        {                                                    \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,  \
                               ac->p5,ac->p6,ac->p7,ac->p8,  \
                               ac->p9);                      \
        }                                                    \
        INLINE                                               \
        name *                                               \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,       \
                              t9),                           \
                   t1 p1,t2 p2,t3 p3,t4 p4,                  \
                   t5 p5,t6 p6,t7 p7,t8 p8,                  \
                   t9 p9)                                    \
        {                                                    \
        name *ac=Memory_alloc(1,name);                       \
        ac->stub=_##name##_stub;                             \
        ac->fnct=fnct;                                       \
        memset(&ac->result,0,sizeof(ac->result));            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;          \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;          \
        ac->p9=p9;                                           \
        return ac;                                           \
        }

#define AsyncCall_typedef_9_void(name,                    \
                                 t1,t2,t3,t4,t5,t6,t7,t8, \
                                 t9)                      \
        typedef struct _##name                            \
        {                                                 \
        void (*stub)(struct _##name *);                   \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,             \
                     t9);                                 \
        t1 p1; t2 p2; t3 p3; t4 p4;                       \
        t5 p5; t6 p6; t7 p7; t8 p8;                       \
        t9 p9;                                            \
        } name;                                           \
        INLINE                                            \
        void                                              \
        _##name##_stub(name *ac)                          \
        {                                                 \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,          \
                    ac->p5,ac->p6,ac->p7,ac->p8,          \
                    ac->p9);                              \
        }                                                 \
        INLINE                                            \
        name *                                            \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,  \
                                t9),                      \
                   t1 p1,t2 p2,t3 p3,t4 p4,               \
                   t5 p5,t6 p6,t7 p7,t8 p8,               \
                   t9 p9)                                 \
        {                                                 \
        name *ac=Memory_alloc(1,name);                    \
        ac->stub=_##name##_stub;                          \
        ac->fnct=fnct;                                    \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;       \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;       \
        ac->p9=p9;                                        \
        return ac;                                        \
        }

#define AsyncCall_typedef_10(rt,name,                        \
                             t1,t2,t3,t4,t5,t6,t7,t8,        \
                             t9,t10)                         \
        typedef struct _##name                               \
        {                                                    \
        void (*stub)(struct _##name *);                      \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                  \
                   t9,t10);                                  \
        rt result;                                           \
        t1 p1; t2 p2; t3 p3; t4 p4;                          \
        t5 p5; t6 p6; t7 p7; t8 p8;                          \
        t9 p9; t10 p10;                                      \
        } name;                                              \
        INLINE                                               \
        void                                                 \
        _##name##_stub(name *ac)                             \
        {                                                    \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,  \
                               ac->p5,ac->p6,ac->p7,ac->p8,  \
                               ac->p9,ac->p10);              \
        }                                                    \
        INLINE                                               \
        name *                                               \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,       \
                              t9,t10),                       \
                   t1 p1,t2 p2,t3 p3,t4 p4,                  \
                   t5 p5,t6 p6,t7 p7,t8 p8,                  \
                   t9 p9,t10 p10)                            \
        {                                                    \
        name *ac=Memory_alloc(1,name);                       \
        ac->stub=_##name##_stub;                             \
        ac->fnct=fnct;                                       \
        memset(&ac->result,0,sizeof(ac->result));            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;          \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;          \
        ac->p9=p9; ac->p10=p10;                              \
        return ac;                                           \
        }

#define AsyncCall_typedef_10_void(name,                    \
                                  t1,t2,t3,t4,t5,t6,t7,t8, \
                                  t9,t10)                  \
        typedef struct _##name                             \
        {                                                  \
        void (*stub)(struct _##name *);                    \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,              \
                     t9,t10);                              \
        t1 p1; t2 p2; t3 p3; t4 p4;                        \
        t5 p5; t6 p6; t7 p7; t8 p8;                        \
        t9 p9; t10 p10;                                    \
        } name;                                            \
        INLINE                                             \
        void                                               \
        _##name##_stub(name *ac)                           \
        {                                                  \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,           \
                    ac->p5,ac->p6,ac->p7,ac->p8,           \
                    ac->p9,ac->p10);                       \
        }                                                  \
        INLINE                                             \
        name *                                             \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,   \
                                t9,t10),                   \
                   t1 p1,t2 p2,t3 p3,t4 p4,                \
                   t5 p5,t6 p6,t7 p7,t8 p8,                \
                   t9 p9,t10 p10)                          \
        {                                                  \
        name *ac=Memory_alloc(1,name);                     \
        ac->stub=_##name##_stub;                           \
        ac->fnct=fnct;                                     \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;        \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;        \
        ac->p9=p9; ac->p10=p10;                            \
        return ac;                                         \
        }

#define AsyncCall_typedef_11(rt,name,                        \
                             t1,t2,t3,t4,t5,t6,t7,t8,        \
                             t9,t10,t11)                     \
        typedef struct _##name                               \
        {                                                    \
        void (*stub)(struct _##name *);                      \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                  \
                   t9,t10,t11);                              \
        rt result;                                           \
        t1 p1; t2 p2; t3 p3; t4 p4;                          \
        t5 p5; t6 p6; t7 p7; t8 p8;                          \
        t9 p9; t10 p10; t11 p11;                             \
        } name;                                              \
        INLINE                                               \
        void                                                 \
        _##name##_stub(name *ac)                             \
        {                                                    \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,  \
                               ac->p5,ac->p6,ac->p7,ac->p8,  \
                               ac->p9,ac->p10,ac->p11);      \
        }                                                    \
        INLINE                                               \
        name *                                               \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,       \
                              t9,t10,t11),                   \
                   t1 p1,t2 p2,t3 p3,t4 p4,                  \
                   t5 p5,t6 p6,t7 p7,t8 p8,                  \
                   t9 p9,t10 p10,t11 p11)                    \
        {                                                    \
        name *ac=Memory_alloc(1,name);                       \
        ac->stub=_##name##_stub;                             \
        ac->fnct=fnct;                                       \
        memset(&ac->result,0,sizeof(ac->result));            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;          \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;          \
        ac->p9=p9; ac->p10=p10; ac->p11=p11;                 \
        return ac;                                           \
        }

#define AsyncCall_typedef_11_void(name,                    \
                                  t1,t2,t3,t4,t5,t6,t7,t8, \
                                  t9,t10,t11)              \
        typedef struct _##name                             \
        {                                                  \
        void (*stub)(struct _##name *);                    \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,              \
                     t9,t10,t11);                          \
        t1 p1; t2 p2; t3 p3; t4 p4;                        \
        t5 p5; t6 p6; t7 p7; t8 p8;                        \
        t9 p9; t10 p10; t11 p11;                           \
        } name;                                            \
        INLINE                                             \
        void                                               \
        _##name##_stub(name *ac)                           \
        {                                                  \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,           \
                    ac->p5,ac->p6,ac->p7,ac->p8,           \
                    ac->p9,ac->p10,ac->p11);               \
        }                                                  \
        INLINE                                             \
        name *                                             \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,   \
                                t9,t10,t11),               \
                   t1 p1,t2 p2,t3 p3,t4 p4,                \
                   t5 p5,t6 p6,t7 p7,t8 p8,                \
                   t9 p9,t10 p10,t11 p11)                  \
        {                                                  \
        name *ac=Memory_alloc(1,name);                     \
        ac->stub=_##name##_stub;                           \
        ac->fnct=fnct;                                     \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;        \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;        \
        ac->p9=p9; ac->p10=p10; ac->p11=p11;               \
        return ac;                                         \
        }

#define AsyncCall_typedef_12(rt,name,                           \
                             t1,t2,t3,t4,t5,t6,t7,t8,           \
                             t9,t10,t11,t12)                    \
        typedef struct _##name                                  \
        {                                                       \
        void (*stub)(struct _##name *);                         \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                   t9,t10,t11,t12);                             \
        rt result;                                              \
        t1 p1; t2 p2; t3 p3; t4 p4;                             \
        t5 p5; t6 p6; t7 p7; t8 p8;                             \
        t9 p9; t10 p10; t11 p11; t12 p12;                       \
        } name;                                                 \
        INLINE                                                  \
        void                                                    \
        _##name##_stub(name *ac)                                \
        {                                                       \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,     \
                               ac->p5,ac->p6,ac->p7,ac->p8,     \
                               ac->p9,ac->p10,ac->p11,ac->p12); \
        }                                                       \
        INLINE                                                  \
        name *                                                  \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                              t9,t10,t11,t12),                  \
                   t1 p1,t2 p2,t3 p3,t4 p4,                     \
                   t5 p5,t6 p6,t7 p7,t8 p8,                     \
                   t9 p9,t10 p10,t11 p11,t12 p12)               \
        {                                                       \
        name *ac=Memory_alloc(1,name);                          \
        ac->stub=_##name##_stub;                                \
        ac->fnct=fnct;                                          \
        memset(&ac->result,0,sizeof(ac->result));               \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;             \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;             \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;       \
        return ac;                                              \
        }

#define AsyncCall_typedef_12_void(name,                    \
                                  t1,t2,t3,t4,t5,t6,t7,t8, \
                                  t9,t10,t11,t12)          \
        typedef struct _##name                             \
        {                                                  \
        void (*stub)(struct _##name *);                    \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,              \
                     t9,t10,t11,t12);                      \
        t1 p1; t2 p2; t3 p3; t4 p4;                        \
        t5 p5; t6 p6; t7 p7; t8 p8;                        \
        t9 p9; t10 p10; t11 p11; t12 p12;                  \
        } name;                                            \
        INLINE                                             \
        void                                               \
        _##name##_stub(name *ac)                           \
        {                                                  \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,           \
                    ac->p5,ac->p6,ac->p7,ac->p8,           \
                    ac->p9,ac->p10,ac->p11,ac->p12);       \
        }                                                  \
        INLINE                                             \
        name *                                             \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,   \
                                t9,t10,t11,t12),           \
                   t1 p1,t2 p2,t3 p3,t4 p4,                \
                   t5 p5,t6 p6,t7 p7,t8 p8,                \
                   t9 p9,t10 p10,t11 p11,t12 p12)          \
        {                                                  \
        name *ac=Memory_alloc(1,name);                     \
        ac->stub=_##name##_stub;                           \
        ac->fnct=fnct;                                     \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;        \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;        \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;  \
        return ac;                                         \
        }

#define AsyncCall_typedef_13(rt,name,                          \
                             t1,t2,t3,t4,t5,t6,t7,t8,          \
                             t9,t10,t11,t12,t13)               \
        typedef struct _##name                                 \
        {                                                      \
        void (*stub)(struct _##name *);                        \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                    \
                   t9,t10,t11,t12,t13);                        \
        rt result;                                             \
        t1 p1; t2 p2; t3 p3; t4 p4;                            \
        t5 p5; t6 p6; t7 p7; t8 p8;                            \
        t9 p9; t10 p10; t11 p11; t12 p12;                      \
        t13 p13;                                               \
        } name;                                                \
        INLINE                                                 \
        void                                                   \
        _##name##_stub(name *ac)                               \
        {                                                      \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,    \
                               ac->p5,ac->p6,ac->p7,ac->p8,    \
                               ac->p9,ac->p10,ac->p11,ac->p12, \
                               ac->p13);                       \
        }                                                      \
        INLINE                                                 \
        name *                                                 \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,         \
                              t9,t10,t11,t12,t13),             \
                   t1 p1,t2 p2,t3 p3,t4 p4,                    \
                   t5 p5,t6 p6,t7 p7,t8 p8,                    \
                   t9 p9,t10 p10,t11 p11,t12 p12,              \
                   t13 p13)                                    \
        {                                                      \
        name *ac=Memory_alloc(1,name);                         \
        ac->stub=_##name##_stub;                               \
        ac->fnct=fnct;                                         \
        memset(&ac->result,0,sizeof(ac->result));              \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;            \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;            \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;      \
        ac->p13=p13;                                           \
        return ac;                                             \
        }

#define AsyncCall_typedef_13_void(name,                    \
                                  t1,t2,t3,t4,t5,t6,t7,t8, \
                                  t9,t10,t11,t12,t13)      \
        typedef struct _##name                             \
        {                                                  \
        void (*stub)(struct _##name *);                    \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,              \
                     t9,t10,t11,t12,t13);                  \
        t1 p1; t2 p2; t3 p3; t4 p4;                        \
        t5 p5; t6 p6; t7 p7; t8 p8;                        \
        t9 p9; t10 p10; t11 p11; t12 p12;                  \
        t13 p13;                                           \
        } name;                                            \
        INLINE                                             \
        void                                               \
        _##name##_stub(name *ac)                           \
        {                                                  \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,           \
                    ac->p5,ac->p6,ac->p7,ac->p8,           \
                    ac->p9,ac->p10,ac->p11,ac->p12,        \
                    ac->p13);                              \
        }                                                  \
        INLINE                                             \
        name *                                             \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,   \
                                t9,t10,t11,t12,t13),       \
                   t1 p1,t2 p2,t3 p3,t4 p4,                \
                   t5 p5,t6 p6,t7 p7,t8 p8,                \
                   t9 p9,t10 p10,t11 p11,t12 p12,          \
                   t13 p13)                                \
        {                                                  \
        name *ac=Memory_alloc(1,name);                     \
        ac->stub=_##name##_stub;                           \
        ac->fnct=fnct;                                     \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;        \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;        \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;  \
        ac->p13=p13;                                       \
        return ac;                                         \
        }

#define AsyncCall_typedef_14(rt,name,                          \
                             t1,t2,t3,t4,t5,t6,t7,t8,          \
                             t9,t10,t11,t12,t13,t14)           \
        typedef struct _##name                                 \
        {                                                      \
        void (*stub)(struct _##name *);                        \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                    \
                   t9,t10,t11,t12,t13,t14);                    \
        rt result;                                             \
        t1 p1; t2 p2; t3 p3; t4 p4;                            \
        t5 p5; t6 p6; t7 p7; t8 p8;                            \
        t9 p9; t10 p10; t11 p11; t12 p12;                      \
        t13 p13; t14 p14;                                      \
        } name;                                                \
        INLINE                                                 \
        void                                                   \
        _##name##_stub(name *ac)                               \
        {                                                      \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,    \
                               ac->p5,ac->p6,ac->p7,ac->p8,    \
                               ac->p9,ac->p10,ac->p11,ac->p12, \
                               ac->p13,ac->p14);               \
        }                                                      \
        INLINE                                                 \
        name *                                                 \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,         \
                              t9,t10,t11,t12,t13,t14),         \
                   t1 p1,t2 p2,t3 p3,t4 p4,                    \
                   t5 p5,t6 p6,t7 p7,t8 p8,                    \
                   t9 p9,t10 p10,t11 p11,t12 p12,              \
                   t13 p13,t14 p14)                            \
        {                                                      \
        name *ac=Memory_alloc(1,name);                         \
        ac->stub=_##name##_stub;                               \
        ac->fnct=fnct;                                         \
        memset(&ac->result,0,sizeof(ac->result));              \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;            \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;            \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;      \
        ac->p13=p13; ac->p14=p14;                              \
        return ac;                                             \
        }

#define AsyncCall_typedef_14_void(name,                    \
                                  t1,t2,t3,t4,t5,t6,t7,t8, \
                                  t9,t10,t11,t12,t13,t14)  \
        typedef struct _##name                             \
        {                                                  \
        void (*stub)(struct _##name *);                    \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,              \
                     t9,t10,t11,t12,t13,t14);              \
        t1 p1; t2 p2; t3 p3; t4 p4;                        \
        t5 p5; t6 p6; t7 p7; t8 p8;                        \
        t9 p9; t10 p10; t11 p11; t12 p12;                  \
        t13 p13; t14 p14;                                  \
        } name;                                            \
        INLINE                                             \
        void                                               \
        _##name##_stub(name *ac)                           \
        {                                                  \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,           \
                    ac->p5,ac->p6,ac->p7,ac->p8,           \
                    ac->p9,ac->p10,ac->p11,ac->p12,        \
                    ac->p13,ac->p14);                      \
        }                                                  \
        INLINE                                             \
        name *                                             \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,   \
                                t9,t10,t11,t12,t13,t14),   \
                   t1 p1,t2 p2,t3 p3,t4 p4,                \
                   t5 p5,t6 p6,t7 p7,t8 p8,                \
                   t9 p9,t10 p10,t11 p11,t12 p12,          \
                   t13 p13,t14 p14)                        \
        {                                                  \
        name *ac=Memory_alloc(1,name);                     \
        ac->stub=_##name##_stub;                           \
        ac->fnct=fnct;                                     \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;        \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;        \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;  \
        ac->p13=p13; ac->p14=p14;                          \
        return ac;                                         \
        }

#define AsyncCall_typedef_15(rt,name,                          \
                             t1,t2,t3,t4,t5,t6,t7,t8,          \
                             t9,t10,t11,t12,t13,t14,t15)       \
        typedef struct _##name                                 \
        {                                                      \
        void (*stub)(struct _##name *);                        \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                    \
                   t9,t10,t11,t12,t13,t14,t15);                \
        rt result;                                             \
        t1 p1; t2 p2; t3 p3; t4 p4;                            \
        t5 p5; t6 p6; t7 p7; t8 p8;                            \
        t9 p9; t10 p10; t11 p11; t12 p12;                      \
        t13 p13; t14 p14; t15 p15;                             \
        } name;                                                \
        INLINE                                                 \
        void                                                   \
        _##name##_stub(name *ac)                               \
        {                                                      \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,    \
                               ac->p5,ac->p6,ac->p7,ac->p8,    \
                               ac->p9,ac->p10,ac->p11,ac->p12, \
                               ac->p13,ac->p14,ac->p15);       \
        }                                                      \
        INLINE                                                 \
        name *                                                 \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,         \
                              t9,t10,t11,t12,t13,t14,t15),     \
                   t1 p1,t2 p2,t3 p3,t4 p4,                    \
                   t5 p5,t6 p6,t7 p7,t8 p8,                    \
                   t9 p9,t10 p10,t11 p11,t12 p12,              \
                   t13 p13,t14 p14,t15 p15)                    \
        {                                                      \
        name *ac=Memory_alloc(1,name);                         \
        ac->stub=_##name##_stub;                               \
        ac->fnct=fnct;                                         \
        memset(&ac->result,0,sizeof(ac->result));              \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;            \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;            \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;      \
        ac->p13=p13; ac->p14=p14; ac->p15=p15;                 \
        return ac;                                             \
        }

#define AsyncCall_typedef_15_void(name,                       \
                                  t1,t2,t3,t4,t5,t6,t7,t8,    \
                                  t9,t10,t11,t12,t13,t14,t15) \
        typedef struct _##name                                \
        {                                                     \
        void (*stub)(struct _##name *);                       \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                 \
                     t9,t10,t11,t12,t13,t14,t15);             \
        t1 p1; t2 p2; t3 p3; t4 p4;                           \
        t5 p5; t6 p6; t7 p7; t8 p8;                           \
        t9 p9; t10 p10; t11 p11; t12 p12;                     \
        t13 p13; t14 p14; t15 p15;                            \
        } name;                                               \
        INLINE                                                \
        void                                                  \
        _##name##_stub(name *ac)                              \
        {                                                     \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,              \
                    ac->p5,ac->p6,ac->p7,ac->p8,              \
                    ac->p9,ac->p10,ac->p11,ac->p12,           \
                    ac->p13,ac->p14,ac->p15);                 \
        }                                                     \
        INLINE                                                \
        name *                                                \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,      \
                                t9,t10,t11,t12,t13,t14,t15),  \
                   t1 p1,t2 p2,t3 p3,t4 p4,                   \
                   t5 p5,t6 p6,t7 p7,t8 p8,                   \
                   t9 p9,t10 p10,t11 p11,t12 p12,             \
                   t13 p13,t14 p14,t15 p15)                   \
        {                                                     \
        name *ac=Memory_alloc(1,name);                        \
        ac->stub=_##name##_stub;                              \
        ac->fnct=fnct;                                        \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;           \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;           \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;     \
        ac->p13=p13; ac->p14=p14; ac->p15=p15;                \
        return ac;                                            \
        }

#define AsyncCall_typedef_16(rt,name,                            \
                             t1,t2,t3,t4,t5,t6,t7,t8,            \
                             t9,t10,t11,t12,t13,t14,t15,t16)     \
        typedef struct _##name                                   \
        {                                                        \
        void (*stub)(struct _##name *);                          \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                      \
                   t9,t10,t11,t12,t13,t14,t15,t16);              \
        rt result;                                               \
        t1 p1; t2 p2; t3 p3; t4 p4;                              \
        t5 p5; t6 p6; t7 p7; t8 p8;                              \
        t9 p9; t10 p10; t11 p11; t12 p12;                        \
        t13 p13; t14 p14; t15 p15; t16 p16;                      \
        } name;                                                  \
        INLINE                                                   \
        void                                                     \
        _##name##_stub(name *ac)                                 \
        {                                                        \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,      \
                               ac->p5,ac->p6,ac->p7,ac->p8,      \
                               ac->p9,ac->p10,ac->p11,ac->p12,   \
                               ac->p13,ac->p14,ac->p15,ac->p16); \
        }                                                        \
        INLINE                                                   \
        name *                                                   \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,           \
                              t9,t10,t11,t12,t13,t14,t15,t16),   \
                   t1 p1,t2 p2,t3 p3,t4 p4,                      \
                   t5 p5,t6 p6,t7 p7,t8 p8,                      \
                   t9 p9,t10 p10,t11 p11,t12 p12,                \
                   t13 p13,t14 p14,t15 p15,t16 p16)              \
        {                                                        \
        name *ac=Memory_alloc(1,name);                           \
        ac->stub=_##name##_stub;                                 \
        ac->fnct=fnct;                                           \
        memset(&ac->result,0,sizeof(ac->result));                \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;              \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;              \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;        \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;      \
        return ac;                                               \
        }

#define AsyncCall_typedef_16_void(name,                           \
                                  t1,t2,t3,t4,t5,t6,t7,t8,        \
                                  t9,t10,t11,t12,t13,t14,t15,t16) \
        typedef struct _##name                                    \
        {                                                         \
        void (*stub)(struct _##name *);                           \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                     t9,t10,t11,t12,t13,t14,t15,t16);             \
        t1 p1; t2 p2; t3 p3; t4 p4;                               \
        t5 p5; t6 p6; t7 p7; t8 p8;                               \
        t9 p9; t10 p10; t11 p11; t12 p12;                         \
        t13 p13; t14 p14; t15 p15; t16 p16;                       \
        } name;                                                   \
        INLINE                                                    \
        void                                                      \
        _##name##_stub(name *ac)                                  \
        {                                                         \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,                  \
                    ac->p5,ac->p6,ac->p7,ac->p8,                  \
                    ac->p9,ac->p10,ac->p11,ac->p12,               \
                    ac->p13,ac->p14,ac->p15,ac->p16);             \
        }                                                         \
        INLINE                                                    \
        name *                                                    \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                                t9,t10,t11,t12,t13,t14,t15,t16),  \
                   t1 p1,t2 p2,t3 p3,t4 p4,                       \
                   t5 p5,t6 p6,t7 p7,t8 p8,                       \
                   t9 p9,t10 p10,t11 p11,t12 p12,                 \
                   t13 p13,t14 p14,t15 p15,t16 p16)               \
        {                                                         \
        name *ac=Memory_alloc(1,name);                            \
        ac->stub=_##name##_stub;                                  \
        ac->fnct=fnct;                                            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;               \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;               \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;         \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;       \
        return ac;                                                \
        }

#define AsyncCall_typedef_17(rt,name,                           \
                             t1,t2,t3,t4,t5,t6,t7,t8,           \
                             t9,t10,t11,t12,t13,t14,t15,t16,    \
                             t17)                               \
        typedef struct _##name                                  \
        {                                                       \
        void (*stub)(struct _##name *);                         \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                   t9,t10,t11,t12,t13,t14,t15,t16,              \
                   t17);                                        \
        rt result;                                              \
        t1 p1; t2 p2; t3 p3; t4 p4;                             \
        t5 p5; t6 p6; t7 p7; t8 p8;                             \
        t9 p9; t10 p10; t11 p11; t12 p12;                       \
        t13 p13; t14 p14; t15 p15; t16 p16;                     \
        t17 p17;                                                \
        } name;                                                 \
        INLINE                                                  \
        void                                                    \
        _##name##_stub(name *ac)                                \
        {                                                       \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,     \
                               ac->p5,ac->p6,ac->p7,ac->p8,     \
                               ac->p9,ac->p10,ac->p11,ac->p12,  \
                               ac->p13,ac->p14,ac->p15,ac->p16, \
                               ac->p17);                        \
        }                                                       \
        INLINE                                                  \
        name *                                                  \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                              t9,t10,t11,t12,t13,t14,t15,t16,   \
                              t17),                             \
                   t1 p1,t2 p2,t3 p3,t4 p4,                     \
                   t5 p5,t6 p6,t7 p7,t8 p8,                     \
                   t9 p9,t10 p10,t11 p11,t12 p12,               \
                   t13 p13,t14 p14,t15 p15,t16 p16,             \
                   t17 p17)                                     \
        {                                                       \
        name *ac=Memory_alloc(1,name);                          \
        ac->stub=_##name##_stub;                                \
        ac->fnct=fnct;                                          \
        memset(&ac->result,0,sizeof(ac->result));               \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;             \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;             \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;       \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;     \
        ac->p17=p17;                                            \
        return ac;                                              \
        }

#define AsyncCall_typedef_17_void(name,                           \
                                  t1,t2,t3,t4,t5,t6,t7,t8,        \
                                  t9,t10,t11,t12,t13,t14,t15,t16, \
                                  t17)                            \
        typedef struct _##name                                    \
        {                                                         \
        void (*stub)(struct _##name *);                           \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                     t9,t10,t11,t12,t13,t14,t15,t16,              \
                     t17);                                        \
        t1 p1; t2 p2; t3 p3; t4 p4;                               \
        t5 p5; t6 p6; t7 p7; t8 p8;                               \
        t9 p9; t10 p10; t11 p11; t12 p12;                         \
        t13 p13; t14 p14; t15 p15; t16 p16;                       \
        t17 p17;                                                  \
        } name;                                                   \
        INLINE                                                    \
        void                                                      \
        _##name##_stub(name *ac)                                  \
        {                                                         \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,                  \
                    ac->p5,ac->p6,ac->p7,ac->p8,                  \
                    ac->p9,ac->p10,ac->p11,ac->p12,               \
                    ac->p13,ac->p14,ac->p15,ac->p16,              \
                    ac->p17);                                     \
        }                                                         \
        INLINE                                                    \
        name *                                                    \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                                t9,t10,t11,t12,t13,t14,t15,t16,   \
                                t17),                             \
                   t1 p1,t2 p2,t3 p3,t4 p4,                       \
                   t5 p5,t6 p6,t7 p7,t8 p8,                       \
                   t9 p9,t10 p10,t11 p11,t12 p12,                 \
                   t13 p13,t14 p14,t15 p15,t16 p16,               \
                   t17 p17)                                       \
        {                                                         \
        name *ac=Memory_alloc(1,name);                            \
        ac->stub=_##name##_stub;                                  \
        ac->fnct=fnct;                                            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;               \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;               \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;         \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;       \
        ac->p17=p17;                                              \
        return ac;                                                \
        }

#define AsyncCall_typedef_18(rt,name,                           \
                             t1,t2,t3,t4,t5,t6,t7,t8,           \
                             t9,t10,t11,t12,t13,t14,t15,t16,    \
                             t17,t18)                           \
        typedef struct _##name                                  \
        {                                                       \
        void (*stub)(struct _##name *);                         \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                   t9,t10,t11,t12,t13,t14,t15,t16,              \
                   t17,t18);                                    \
        rt result;                                              \
        t1 p1; t2 p2; t3 p3; t4 p4;                             \
        t5 p5; t6 p6; t7 p7; t8 p8;                             \
        t9 p9; t10 p10; t11 p11; t12 p12;                       \
        t13 p13; t14 p14; t15 p15; t16 p16;                     \
        t17 p17; t18 p18;                                       \
        } name;                                                 \
        INLINE                                                  \
        void                                                    \
        _##name##_stub(name *ac)                                \
        {                                                       \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,     \
                               ac->p5,ac->p6,ac->p7,ac->p8,     \
                               ac->p9,ac->p10,ac->p11,ac->p12,  \
                               ac->p13,ac->p14,ac->p15,ac->p16, \
                               ac->p17,ac->p18);                \
        }                                                       \
        INLINE                                                  \
        name *                                                  \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                              t9,t10,t11,t12,t13,t14,t15,t16,   \
                              t17,t18),                         \
                   t1 p1,t2 p2,t3 p3,t4 p4,                     \
                   t5 p5,t6 p6,t7 p7,t8 p8,                     \
                   t9 p9,t10 p10,t11 p11,t12 p12,               \
                   t13 p13,t14 p14,t15 p15,t16 p16,             \
                   t17 p17,t18 p18)                             \
        {                                                       \
        name *ac=Memory_alloc(1,name);                          \
        ac->stub=_##name##_stub;                                \
        ac->fnct=fnct;                                          \
        memset(&ac->result,0,sizeof(ac->result));               \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;             \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;             \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;       \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;     \
        ac->p17=p17; ac->p18=p18;                               \
        return ac;                                              \
        }

#define AsyncCall_typedef_18_void(name,                           \
                                  t1,t2,t3,t4,t5,t6,t7,t8,        \
                                  t9,t10,t11,t12,t13,t14,t15,t16, \
                                  t17,t18)                        \
        typedef struct _##name                                    \
        {                                                         \
        void (*stub)(struct _##name *);                           \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                     t9,t10,t11,t12,t13,t14,t15,t16,              \
                     t17,t18);                                    \
        t1 p1; t2 p2; t3 p3; t4 p4;                               \
        t5 p5; t6 p6; t7 p7; t8 p8;                               \
        t9 p9; t10 p10; t11 p11; t12 p12;                         \
        t13 p13; t14 p14; t15 p15; t16 p16;                       \
        t17 p17; t18 p18;                                         \
        } name;                                                   \
        INLINE                                                    \
        void                                                      \
        _##name##_stub(name *ac)                                  \
        {                                                         \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,                  \
                    ac->p5,ac->p6,ac->p7,ac->p8,                  \
                    ac->p9,ac->p10,ac->p11,ac->p12,               \
                    ac->p13,ac->p14,ac->p15,ac->p16,              \
                    ac->p17,ac->p18);                             \
        }                                                         \
        INLINE                                                    \
        name *                                                    \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                                t9,t10,t11,t12,t13,t14,t15,t16,   \
                                t17,t18),                         \
                   t1 p1,t2 p2,t3 p3,t4 p4,                       \
                   t5 p5,t6 p6,t7 p7,t8 p8,                       \
                   t9 p9,t10 p10,t11 p11,t12 p12,                 \
                   t13 p13,t14 p14,t15 p15,t16 p16,               \
                   t17 p17,t18 p18)                               \
        {                                                         \
        name *ac=Memory_alloc(1,name);                            \
        ac->stub=_##name##_stub;                                  \
        ac->fnct=fnct;                                            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;               \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;               \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;         \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;       \
        ac->p17=p17; ac->p18=p18;                                 \
        return ac;                                                \
        }

#define AsyncCall_typedef_19(rt,name,                           \
                             t1,t2,t3,t4,t5,t6,t7,t8,           \
                             t9,t10,t11,t12,t13,t14,t15,t16,    \
                             t17,t18,t19)                       \
        typedef struct _##name                                  \
        {                                                       \
        void (*stub)(struct _##name *);                         \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                   t9,t10,t11,t12,t13,t14,t15,t16,              \
                   t17,t18,t19);                                \
        rt result;                                              \
        t1 p1; t2 p2; t3 p3; t4 p4;                             \
        t5 p5; t6 p6; t7 p7; t8 p8;                             \
        t9 p9; t10 p10; t11 p11; t12 p12;                       \
        t13 p13; t14 p14; t15 p15; t16 p16;                     \
        t17 p17; t18 p18; t19 p19;                              \
        } name;                                                 \
        INLINE                                                  \
        void                                                    \
        _##name##_stub(name *ac)                                \
        {                                                       \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,     \
                               ac->p5,ac->p6,ac->p7,ac->p8,     \
                               ac->p9,ac->p10,ac->p11,ac->p12,  \
                               ac->p13,ac->p14,ac->p15,ac->p16, \
                               ac->p17,ac->p18,ac->p19);        \
        }                                                       \
        INLINE                                                  \
        name *                                                  \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                              t9,t10,t11,t12,t13,t14,t15,t16,   \
                              t17,t18,t19),                     \
                   t1 p1,t2 p2,t3 p3,t4 p4,                     \
                   t5 p5,t6 p6,t7 p7,t8 p8,                     \
                   t9 p9,t10 p10,t11 p11,t12 p12,               \
                   t13 p13,t14 p14,t15 p15,t16 p16,             \
                   t17 p17,t18 p18,t19 p19)                     \
        {                                                       \
        name *ac=Memory_alloc(1,name);                          \
        ac->stub=_##name##_stub;                                \
        ac->fnct=fnct;                                          \
        memset(&ac->result,0,sizeof(ac->result));               \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;             \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;             \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;       \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;     \
        ac->p17=p17; ac->p18=p18; ac->p19=p19;                  \
        return ac;                                              \
        }

#define AsyncCall_typedef_19_void(name,                           \
                                  t1,t2,t3,t4,t5,t6,t7,t8,        \
                                  t9,t10,t11,t12,t13,t14,t15,t16, \
                                  t17,t18,t19)                    \
        typedef struct _##name                                    \
        {                                                         \
        void (*stub)(struct _##name *);                           \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                     t9,t10,t11,t12,t13,t14,t15,t16,              \
                     t17,t18,t19);                                \
        t1 p1; t2 p2; t3 p3; t4 p4;                               \
        t5 p5; t6 p6; t7 p7; t8 p8;                               \
        t9 p9; t10 p10; t11 p11; t12 p12;                         \
        t13 p13; t14 p14; t15 p15; t16 p16;                       \
        t17 p17; t18 p18; t19 p19;                                \
        } name;                                                   \
        INLINE                                                    \
        void                                                      \
        _##name##_stub(name *ac)                                  \
        {                                                         \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,                  \
                    ac->p5,ac->p6,ac->p7,ac->p8,                  \
                    ac->p9,ac->p10,ac->p11,ac->p12,               \
                    ac->p13,ac->p14,ac->p15,ac->p16,              \
                    ac->p17,ac->p18,ac->p19);                     \
        }                                                         \
        INLINE                                                    \
        name *                                                    \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                                t9,t10,t11,t12,t13,t14,t15,t16,   \
                                t17,t18,t19),                     \
                   t1 p1,t2 p2,t3 p3,t4 p4,                       \
                   t5 p5,t6 p6,t7 p7,t8 p8,                       \
                   t9 p9,t10 p10,t11 p11,t12 p12,                 \
                   t13 p13,t14 p14,t15 p15,t16 p16,               \
                   t17 p17,t18 p18,t19 p19)                       \
        {                                                         \
        name *ac=Memory_alloc(1,name);                            \
        ac->stub=_##name##_stub;                                  \
        ac->fnct=fnct;                                            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;               \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;               \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;         \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;       \
        ac->p17=p17; ac->p18=p18; ac->p19=p19;                    \
        return ac;                                                \
        }

#define AsyncCall_typedef_20(rt,name,                            \
                             t1,t2,t3,t4,t5,t6,t7,t8,            \
                             t9,t10,t11,t12,t13,t14,t15,t16,     \
                             t17,t18,t19,t20)                    \
        typedef struct _##name                                   \
        {                                                        \
        void (*stub)(struct _##name *);                          \
        rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                      \
                   t9,t10,t11,t12,t13,t14,t15,t16,               \
                   t17,t18,t19,t20);                             \
        rt result;                                               \
        t1 p1; t2 p2; t3 p3; t4 p4;                              \
        t5 p5; t6 p6; t7 p7; t8 p8;                              \
        t9 p9; t10 p10; t11 p11; t12 p12;                        \
        t13 p13; t14 p14; t15 p15; t16 p16;                      \
        t17 p17; t18 p18; t19 p19; t20 p20;                      \
        } name;                                                  \
        INLINE                                                   \
        void                                                     \
        _##name##_stub(name *ac)                                 \
        {                                                        \
        ac->result=(*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,      \
                               ac->p5,ac->p6,ac->p7,ac->p8,      \
                               ac->p9,ac->p10,ac->p11,ac->p12,   \
                               ac->p13,ac->p14,ac->p15,ac->p16,  \
                               ac->p17,ac->p18,ac->p19,ac->p20); \
        }                                                        \
        INLINE                                                   \
        name *                                                   \
        name##_new(rt (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,           \
                              t9,t10,t11,t12,t13,t14,t15,t16,    \
                              t17,t18,t19,t20),                  \
                   t1 p1,t2 p2,t3 p3,t4 p4,                      \
                   t5 p5,t6 p6,t7 p7,t8 p8,                      \
                   t9 p9,t10 p10,t11 p11,t12 p12,                \
                   t13 p13,t14 p14,t15 p15,t16 p16,              \
                   t17 p17,t18 p18,t19 p19,t20 p20)              \
        {                                                        \
        name *ac=Memory_alloc(1,name);                           \
        ac->stub=_##name##_stub;                                 \
        ac->fnct=fnct;                                           \
        memset(&ac->result,0,sizeof(ac->result));                \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;              \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;              \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;        \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;      \
        ac->p17=p17; ac->p18=p18; ac->p19=p19; ac->p20=p20;      \
        return ac;                                               \
        }

#define AsyncCall_typedef_20_void(name,                           \
                                  t1,t2,t3,t4,t5,t6,t7,t8,        \
                                  t9,t10,t11,t12,t13,t14,t15,t16, \
                                  t17,t18,t19,t20)                \
        typedef struct _##name                                    \
        {                                                         \
        void (*stub)(struct _##name *);                           \
        void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,                     \
                     t9,t10,t11,t12,t13,t14,t15,t16,              \
                     t17,t18,t19,t20);                            \
        t1 p1; t2 p2; t3 p3; t4 p4;                               \
        t5 p5; t6 p6; t7 p7; t8 p8;                               \
        t9 p9; t10 p10; t11 p11; t12 p12;                         \
        t13 p13; t14 p14; t15 p15; t16 p16;                       \
        t17 p17; t18 p18; t19 p19; t20 p20;                       \
        } name;                                                   \
        INLINE                                                    \
        void                                                      \
        _##name##_stub(name *ac)                                  \
        {                                                         \
        (*ac->fnct)(ac->p1,ac->p2,ac->p3,ac->p4,                  \
                    ac->p5,ac->p6,ac->p7,ac->p8,                  \
                    ac->p9,ac->p10,ac->p11,ac->p12,               \
                    ac->p13,ac->p14,ac->p15,ac->p16,              \
                    ac->p17,ac->p18,ac->p19,ac->p20);             \
        }                                                         \
        INLINE                                                    \
        name *                                                    \
        name##_new(void (*fnct)(t1,t2,t3,t4,t5,t6,t7,t8,          \
                                t9,t10,t11,t12,t13,t14,t15,t16,   \
                                t17,t18,t19,t20),                 \
                   t1 p1,t2 p2,t3 p3,t4 p4,                       \
                   t5 p5,t6 p6,t7 p7,t8 p8,                       \
                   t9 p9,t10 p10,t11 p11,t12 p12,                 \
                   t13 p13,t14 p14,t15 p15,t16 p16,               \
                   t17 p17,t18 p18,t19 p19,t20 p20)               \
        {                                                         \
        name *ac=Memory_alloc(1,name);                            \
        ac->stub=_##name##_stub;                                  \
        ac->fnct=fnct;                                            \
        ac->p1=p1; ac->p2=p2; ac->p3=p3; ac->p4=p4;               \
        ac->p5=p5; ac->p6=p6; ac->p7=p7; ac->p8=p8;               \
        ac->p9=p9; ac->p10=p10; ac->p11=p11; ac->p12=p12;         \
        ac->p13=p13; ac->p14=p14; ac->p15=p15; ac->p16=p16;       \
        ac->p17=p17; ac->p18=p18; ac->p19=p19; ac->p20=p20;       \
        return ac;                                                \
        }

#define AsyncCall_execute(asyncCall)                                       \
        do                                                                 \
          {                                                                \
          (**(void (**)(AsyncCall *))asyncCall)((AsyncCall *)(asyncCall)); \
          } while(0)

#define AsyncCall_result(asyncCall) \
        ((asyncCall)->result)

#define AsyncCall_delete(asyncCall) \
        do                          \
          {                         \
          Memory_free(asyncCall);   \
          } while(0)

#endif // TRANSPROG_ASYNCCALL_H

//----------------------------------------------------------------------------
