//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_PATH_H
#define TRANSPROG_PATH_H 1

#include "TransProg/platform.h"

typedef enum { PATH_ERROR,
               PATH_FILE,
               PATH_DIR,
               PATH_SPECIAL } PathKind;

typedef struct _PathDir PathDir;

TRANSPROG_API
PathKind
Path_kind(const char *path);

TRANSPROG_API
int64_t // file size or -1
Path_fileSize(const char *fileName);

TRANSPROG_API
intptr_t // amount read or -1
Path_readFile(const char *fileName,
              void *out_buffer,
              intptr_t size,
              int64_t offset);

TRANSPROG_API
intptr_t // amount written or -1
Path_writeFile(const char *fileName,
               const void *buffer,
               intptr_t size,
               bool append);

TRANSPROG_API
FILE * // open file or 0
Path_fopen(const char *fileName,
           const char *mode);

INLINE
char
Path_dirSep(void)
{
#if defined _WIN32
  return '\\';
#else
  return '/';
#endif
}

TRANSPROG_API
void
Path_normalize(char *inout_path);

TRANSPROG_API
bool // success
Path_join(char *out_path,
          int pathSize,
          const char *path1,
          const char *path2);

TRANSPROG_API
const char * // basename in path or 0
Path_findBasename(const char *path);

TRANSPROG_API
const char * // extension in path or 0
Path_findExtension(const char *path);

TRANSPROG_API
bool // success
Path_currentDir(char *out_currentDir,
                int currentDirSize);

TRANSPROG_API
bool // success (lookedFor was absolute, relative or relative to executable)
Path_lookFor(char *out_path,
             int pathSize,
             const char *lookedFor);

TRANSPROG_API
PathDir * // directory or 0
Path_openDir(const char *path);

TRANSPROG_API
const char * // next element in directory or 0
Path_readDir(PathDir *dir);

TRANSPROG_API
void
Path_closeDir(PathDir *dir);

TRANSPROG_API
bool // success
Path_createDir(const char *path);

TRANSPROG_API
bool // success
Path_createTmpDir(char *out_dirName,
                  int dirNameSize,
                  const char *path,
                  const char *pattern);

TRANSPROG_API
bool // success
Path_createTmpFile(char *out_fileName,
                   int fileNameSize,
                   const char *path,
                   const char *pattern);

TRANSPROG_API
bool // success
Path_remove(const char *path);

#endif // TRANSPROG_PATH_H

//----------------------------------------------------------------------------
