//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_SOCKET_H
#define TRANSPROG_SOCKET_H 1

#include "TransProg/platform.h"

typedef SOCKET SocketHandle;
typedef uint32_t Ipv4Address;

TRANSPROG_API
bool // success
Socket_parseIpv4Address(Ipv4Address *out_address,
                        const char *txt);

TRANSPROG_API
void
Socket_formatIpv4Address(char *out_txt,
                         Ipv4Address address);

TRANSPROG_API
bool // success
Socket_hostName(char *out_hostName,
                int capacity);

TRANSPROG_API
Ipv4Address // IPV4 address of host name or 0
Socket_ipv4Address(const char *hostName);

TRANSPROG_API
SocketHandle // socket or INVALID_SOCKET
Socket_open(int domain,
            int type,
            int protocol);

TRANSPROG_API
bool // success
Socket_close(SocketHandle s);

TRANSPROG_API
bool // success
Socket_shutdown(SocketHandle s,
                int how);

TRANSPROG_API
bool // success
Socket_setReuseAddrOption(SocketHandle s);

TRANSPROG_API
bool // success
Socket_setBroadcastOption(SocketHandle s);

TRANSPROG_API
bool // success
Socket_setRcvBufOption(SocketHandle s,
                       int rcvBuf);

TRANSPROG_API
bool // success
Socket_bindIpv4(SocketHandle s,
                Ipv4Address address,
                uint16_t port);

TRANSPROG_API
bool // success
Socket_localIpv4(SocketHandle s,
                 Ipv4Address *out_address,
                 uint16_t *out_port);

TRANSPROG_API
bool // success
Socket_distantIpv4(SocketHandle s,
                   Ipv4Address *out_address,
                   uint16_t *out_port);

TRANSPROG_API
bool // success
Socket_listen(SocketHandle s,
              int backlog);

TRANSPROG_API
SocketHandle // dialog socket or INVALID_SOCKET
Socket_acceptIpv4(SocketHandle listenSocket,
                  Ipv4Address *out_address,
                  uint16_t *out_port);

TRANSPROG_API
bool // success
Socket_connectIpv4(SocketHandle s,
                   Ipv4Address address,
                   uint16_t port);

TRANSPROG_API
int // sent bytes or SOCKET_ERROR
Socket_send(SocketHandle s,
            const void *data,
            int size,
            int flags);

TRANSPROG_API
int // sent bytes or SOCKET_ERROR
Socket_sendToIpv4(SocketHandle s,
                  const void *data,
                  int size,
                  int flags,
                  Ipv4Address address,
                  uint16_t port);

TRANSPROG_API
int // received bytes or 0 (EOF) or SOCKET_ERROR
Socket_recv(SocketHandle s,
            void *buffer,
            int capacity,
            int flags);

TRANSPROG_API
int // received bytes or 0 (EOF) or SOCKET_ERROR
Socket_recvFromIpv4(SocketHandle s,
                    void *buffer,
                    int capacity,
                    int flags,
                    Ipv4Address *out_address,
                    uint16_t *out_port);

TRANSPROG_API
int // number of socket in ready-state or SOCKET_ERROR
Socket_select(SocketHandle *inout_rdSet,
              SocketHandle *inout_wrSet,
              int64_t usecTimeout);

INLINE
bool // found
Socket_inSet(SocketHandle s,
             const SocketHandle *socketSet)
{
for(int i=0;socketSet[i]!=INVALID_SOCKET;++i)
  {
  if(socketSet[i]==s) return true;
  }
return false;
}

#endif // TRANSPROG_SOCKET_H

//----------------------------------------------------------------------------
