#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
#   http://www.enib.fr/~harrouet/transprog.html
#
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import os
import sys

class Structure:
  def __init__(self,**kw):
    self.__dict__=kw

coreArbLines=[]
extensions=[]
curExt=None

def findWord(l,e,startswith=True,endswith=True):
  for i in range(len(l)):
    found=True
    if startswith and not l[i].startswith(e):
      found=False
    if endswith and not l[i].endswith(e):
      found=False
    if found:
      return i
  return -1

def findApiEntry(typeName,ext=None):
  if ext:
    ext=[ext]
  else:
    ext=extensions
  for e in ext:
    for a in e.apiEntries:
      if a.typeName==typeName:
        return a
  return None

for l in open('glcorearb.h','r'):
  coreArbLines.append(l[:])
  l=l.split()
  if len(l)==2 and l[0]=='#ifndef' and l[1].startswith('GL_'):
    extension=l[1][3:]
    if curExt:
      raise Exception('extension %s detected while in %s'
                      %(extension,curExt.name))
    curExt=Structure(name=extension,apiEntries=[])
    extensions.append(curExt)
    continue
  if len(l)==4 and l[0]=='#endif' and l[1]=='/*' and \
     l[2].startswith('GL_') and l[3]=='*/':
    extension=l[2][3:]
    if not curExt:
      raise Exception('closing extension %s out of any extension'%extension)
    if curExt.name!=extension:
      raise Exception('closing extension %s out in extension %s'
                      %(extension,curExt.name))
    curExt=None
    continue
  i=findWord(l,'typedef')
  if i==0:
    i=findWord(l,'APIENTRYP',startswith=False)
    if i!=-1 and len(l)>i+1:
      typeName=l[i+1]
      if typeName.startswith('PFNGL') and typeName.endswith('PROC)'):
        typeName=typeName[5:-5]
        if not curExt:
          raise Exception('type %s detected out of any extension'%typeName)
        if findApiEntry(typeName):
          raise Exception('multiple occurences of type %s'%typeName)
        curExt.apiEntries.append(Structure(typeName=typeName,name=None))
        continue
  i=findWord(l,'GLAPI')
  if i==0:
    i=findWord(l,'APIENTRY',startswith=False)
    if i!=-1 and len(l)>i+1:
      name=l[i+1]
      if name.startswith('gl'):
        name=name[2:]
        typeName=name.upper()
        if not curExt:
          raise Exception('entry %s detected out of any extension'%name)
        e=findApiEntry(typeName)
        if not e:
          raise Exception('no type %s for entry %s'%(typeName,name))
        if not e in curExt.apiEntries:
          raise Exception('type %s should be in extension %s'
                          %(typeName,curExt.name))
        if e.name:
          raise Exception('multiple names for type %s'%typeName)
        e.name=name
        continue

#-----------------------------------------------------------------------------

header=open('coreGl.h','w')
source=open('coreGl.c','w')

preamble='//'+'-'*76+r'''
// Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

// This Core OpenGL wrapper is generated from:
//    http://www.opengl.org/registry/api/GL/glcorearb.h
// Do not edit or your changes will be lost!
'''

header.write(preamble)
source.write(preamble)

header.write(r'''
#ifndef TRANSGL_COREGL_H
#define TRANSGL_COREGL_H 1

#include "TransGl/api.h"
//
// The preceding inclusion can be commented out in order to make this code
// independant from the TransProg library.
// In this case, the following settings are required.
// They may need some adjustments on some platforms.
//

#if !defined TRANSGL_API_H // TransProg library was not used here

#include "platformAdapter.h"

# if !defined TRANSGL_API_VAR
#   define TRANSGL_API_VAR // visibility range is current binary
# endif

# if !defined TRANSGL_API
#   define TRANSGL_API EXTERN_C_FUNCTION TRANSGL_API_VAR
# endif

#endif // not TRANSGL_API_H

//%s

'''%('-'*76))
for l in coreArbLines:
  header.write(l)
header.write(r'''
//%s

#if defined NDEBUG
# define ON_GL_ERROR(action) \
         do { ; } while(0)
#else
# define ON_GL_ERROR(action)                       \
         do                                        \
           {                                       \
           const char *__glErr=CoreGl_lastError(); \
           ON_ERROR(__glErr,__glErr,action);       \
           } while(0)
#endif

typedef struct _CoreGlComputeProperties
{
int maxWorkGroupCount[3];
int maxWorkGroupSize[3];
int maxWorkGroupInvocations;
int maxUniformBlocks;
int maxTextureImageUnits;
int maxAtomicCounterBuffers;
int maxAtomicCounters;
int maxSharedMemorySize;
int maxUniformComponents;
int maxImageUniforms;
int maxCombinedUniformComponents;
int maxShaderStorageBlocks;
int warpSize;
int warpsPerStreamingMultiprocessor;
int streamingMultiprocessorCount;
bool hasBallot;
bool hasShuffle;
} CoreGlComputeProperties;

struct _CoreGlManager
{
bool initDone;
bool coreProfile;
bool embedded;
int majorVersion,minorVersion;
int shadingLanguageVersion;
int extensionCount;
char ** extensions;
bool hasTessControlShaders;
bool hasTessEvaluationShaders;
bool hasGeometryShaders;
bool hasComputeShaders;
CoreGlComputeProperties computeProperties;'''%('-'*76))

for e in extensions:
  header.write(r'''
bool has_%s;'''%e.name)

header.write(r'''
};

extern TRANSGL_API_VAR struct _CoreGlManager _CoreGl_mgr;

TRANSGL_API
bool // success
CoreGl_init(void);

INLINE
bool // OpenGL core profile is supported
CoreGl_coreProfile(void)
{
return _CoreGl_mgr.coreProfile;
}

INLINE
bool // OpenGL is actually OpenGL ES
CoreGl_embedded(void)
{
return _CoreGl_mgr.embedded;
}

INLINE
int // OpenGL major version
CoreGl_majorVersion(void)
{
return _CoreGl_mgr.majorVersion;
}

INLINE
int // OpenGL minor version
CoreGl_minorVersion(void)
{
return _CoreGl_mgr.minorVersion;
}

INLINE
bool // success
CoreGl_isAtLeastVersion(int major,
                        int minor)
{
return (_CoreGl_mgr.majorVersion>major)||
       ((_CoreGl_mgr.majorVersion==major)&&
        (_CoreGl_mgr.minorVersion>=minor));
}

INLINE
int // GLSL version
CoreGl_shadingLanguageVersion(void)
{
return _CoreGl_mgr.shadingLanguageVersion;
}

INLINE
bool // tessellation evaluation shaders are available
CoreGl_hasTessEvaluationShaders(void)
{
return _CoreGl_mgr.hasTessEvaluationShaders;
}

INLINE
bool // tessellation control shaders are available
CoreGl_hasTessControlShaders(void)
{
return _CoreGl_mgr.hasTessControlShaders;
}

INLINE
bool // geometry shaders are available
CoreGl_hasGeometryShaders(void)
{
return _CoreGl_mgr.hasGeometryShaders;
}

INLINE
bool // compute shaders are available
CoreGl_hasComputeShaders(void)
{
return _CoreGl_mgr.hasComputeShaders;
}

INLINE
const CoreGlComputeProperties * // compute shader properties or 0
CoreGl_computeProperties(void)
{
return _CoreGl_mgr.hasComputeShaders ? &_CoreGl_mgr.computeProperties
                                     : (const CoreGlComputeProperties *)0;
}

INLINE
const char * const * // extension name array
CoreGl_extensions(int *out_extensionCount)
{
*out_extensionCount=_CoreGl_mgr.extensionCount;
return (const char * const *)_CoreGl_mgr.extensions;
}

TRANSGL_API
bool // success
CoreGl_hasExtension(const char *extensionName);

TRANSGL_API
void * // symbol or 0
CoreGl_extensionSymbol(const char *symbolName);

#define CoreGl_extensionFunction(out_function,fnctName) \
        do                                              \
          {                                             \
          void *_symb=CoreGl_extensionSymbol(fnctName); \
          memcpy((out_function),&_symb,sizeof(void *)); \
          } while(0)

TRANSGL_API
void * // symbol or 0
CoreGl_extensionSymbolWithSuffix(const char *symbolName,
                                 const char * const *suffixList);

#define CoreGl_extensionFunctionWithSuffix(out_function,fnctName,suffixList) \
        do                                                                   \
          {                                                                  \
          void *_symb=CoreGl_extensionSymbolWithSuffix(fnctName,suffixList); \
          memcpy((out_function),&_symb,sizeof(void *));                      \
          } while(0)

TRANSGL_API
const char * // last OpenGL error message if any
CoreGl_lastError(void);

TRANSGL_API
char * // allocated compilation/link error/warning message if any
CoreGl_shaderStatus(GLuint shader,
                    bool *out_error);

TRANSGL_API
GLuint // compiled shader-object or 0
CoreGl_compileShader(GLenum shaderKind,
                     const char *sourceCode,
                     bool addPreamble,
                     char **out_allocatedCompilationLog);

TRANSGL_API
GLuint // linked shader-program or 0
CoreGl_buildRenderShader(const char *vertexSrc,
                         const char *geometrySrc,
                         const char *fragmentSrc,
                         bool addPreamble,
                         char **out_allocatedBuildLog);

TRANSGL_API
GLuint // linked shader-program or 0
CoreGl_buildComputeShader(const char *computeSrc,
                          bool addPreamble,
                          char **out_allocatedBuildLog);

''')

source.write(r'''
#include "TransGl/coreGl.h"
//
// The preceding inclusion can be changed in order to find coreGl.h in its
// actual location (independant from the TransProg library).
//

#if defined TRANSGL_API_H
  #include "TransProg/memory.h"
  const char * _CoreGl_querySpecificExtensions(void); // defined elsewhere
#else // TransProg library is not used
  static
  const char * _CoreGl_querySpecificExtensions(void)
  { return (const char *)0; } // fake definition
#endif

#if defined _WIN32
# define GL_LIBRARY_NAME  "OPENGL32.DLL"
#elif defined __APPLE__
# define GL_LIBRARY_NAME  "/System/Library/Frameworks/OpenGL.framework" \
                          "/Libraries/libGL.dylib"
#elif defined __ANDROID__
# define GL_LIBRARY_NAME  "libGLESv2.so"
  #if !defined EGL_VERSION
    EXTERN_C_FUNCTION void (*eglGetProcAddress(const char *))(void);
  #endif
#else
# define GL_LIBRARY_NAME  "libGL.so"
  #if !defined GLX_VERSION
    EXTERN_C_FUNCTION void (*glXGetProcAddress(const GLubyte *))(void);
  #endif
#endif

TRANSGL_API_VAR struct _CoreGlManager _CoreGl_mgr;

static
void
CoreGl_addExtensionList(const char *extensionList)
{
if(extensionList)
  {
  for(int run=0;run<2;++run)
    {
    const char *beginList=extensionList;
    const char *endList=beginList+strlen(beginList);
    int extCount=_CoreGl_mgr.extensionCount;
    while(beginList<endList)
      {
      int extLen=(int)strcspn(beginList," ");
      if(extLen)
        {
        if(run!=0)
          {
          char *ext=Memory_alloc(extLen+1,char);
          memcpy(ext,beginList,extLen);
          ext[extLen]='\0';
          _CoreGl_mgr.extensions[extCount]=ext;
          }
        ++extCount;
        }
      beginList+=extLen+1;
      }
    if(run==0)
      {
      _CoreGl_mgr.extensions=Memory_reAlloc(_CoreGl_mgr.extensions,
                                            extCount+1,char *);
      }
    else
      {
      _CoreGl_mgr.extensions[extCount]=(char *)0;
      _CoreGl_mgr.extensionCount=extCount;
      }
    }
  }
}
''')

for e in extensions:
  source.write(r'''
static
bool // extension is available
CoreGl_load_%s(void)
{
bool hasExt='''%e.name)
  if e.name.startswith('VERSION_'):
    v=e.name.split('_')
    source.write(r'''CoreGl_isAtLeastVersion(%d,%d);
'''%(int(v[1]),int(v[2])))
  else:
    source.write(r'''CoreGl_hasExtension("GL_%s");
'''%e.name)
  for a in e.apiEntries:
    source.write(r'''CoreGl_extensionFunction(&CoreGl_%s,
                         "gl%s");
'''%(a.name,a.name))
    if e.name.startswith('VERSION_'):
      source.write(r'''if(!CoreGl_%s)
  {
  hasExt=false;
  }
'''%a.name)
    else:
      source.write(r'''ON_ERROR(hasExt&&!CoreGl_%s,
         "missing gl%s() in GL_%s extension\n",
         {});
'''%(a.name,a.name,e.name))
  source.write(r'''return hasExt;
}
''')

source.write(r'''
bool // success
CoreGl_init(void)
{
if(!_CoreGl_mgr.initDone)
  {
  _CoreGl_mgr.coreProfile=false;
  _CoreGl_mgr.embedded=false;
  _CoreGl_mgr.majorVersion=0;
  _CoreGl_mgr.minorVersion=0;
  _CoreGl_mgr.shadingLanguageVersion=0;
  _CoreGl_mgr.extensionCount=0;
  _CoreGl_mgr.extensions=Memory_alloc(1,char *);
  _CoreGl_mgr.extensions[0]=(char *)0;
  _CoreGl_mgr.hasTessControlShaders=false;
  _CoreGl_mgr.hasTessEvaluationShaders=false;
  _CoreGl_mgr.hasGeometryShaders=false;
  _CoreGl_mgr.hasComputeShaders=false;
  CoreGl_extensionFunction(&glGetError,"glGetError");
  ON_ERROR(!glGetError,"glGetError() not found",{ return false; });
  CoreGl_extensionFunction(&glGetIntegerv,"glGetIntegerv");
  ON_ERROR(!glGetIntegerv,"glGetIntegerv() not found",{ return false; });
  CoreGl_extensionFunction(&glGetString,"glGetString");
  ON_ERROR(!glGetString,"glGetString() not found",{ return false; });
  int value=-1;
  glGetIntegerv(GL_CONTEXT_PROFILE_MASK,&value); // may fail
  if(value==-1) glGetError(); // clear previous error
  else
    {
    if(value&GL_CONTEXT_CORE_PROFILE_BIT)
      {
      _CoreGl_mgr.coreProfile=true;
      }
    }
  bool success=false;
  const char *version=(const char *)glGetString(GL_VERSION);
  ON_GL_ERROR({ return false; });
  if(version)
    {
    int major,minor;
    if(sscanf(version," %d.%d ",&major,&minor)==2)
      {
      _CoreGl_mgr.majorVersion=major;
      _CoreGl_mgr.minorVersion=minor;
      success=true;
      }
    else if(sscanf(version," OpenGL ES %d.%d ",&major,&minor)==2)
      {
      _CoreGl_mgr.embedded=true;
      _CoreGl_mgr.majorVersion=major;
      _CoreGl_mgr.minorVersion=minor;
      success=true;
      }
    }
  ON_ERROR(!success,"OpenGL version not detected",{ return false; });
  success=false;
  const char *glslVersion=(const char *)
                          glGetString(GL_SHADING_LANGUAGE_VERSION);
  ON_GL_ERROR({ return false; });
  if(glslVersion)
    {
    int major,minor;
    if(((sscanf(glslVersion," %d.%d ",&major,&minor)==2)||
        (sscanf(glslVersion," OpenGL ES GLSL %d.%d ",&major,&minor)==2)||
        (sscanf(glslVersion," OpenGL ES GLSL ES %d.%d ",&major,&minor)==2))&&
       (minor<100))
      {
      _CoreGl_mgr.shadingLanguageVersion=100*major+minor;
      success=true;
      }
    }
  ON_ERROR(!success,"OpenGL GLSL version not detected",{ return false; });
  CoreGl_extensionFunction(&glGetStringi,"glGetStringi"); // may fail
  if(glGetStringi) // modern means to detect extensions
    {
    int extCount=0;
    glGetIntegerv(GL_NUM_EXTENSIONS,&extCount); // may fail
    if(!extCount) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.extensions=Memory_reAlloc(_CoreGl_mgr.extensions,
                                            extCount+1,char *);
      for(int i=0;i<extCount;++i)
        {
        const char *s=(const char *)glGetStringi(GL_EXTENSIONS,i); // may fail
        if(!s) glGetError(); // clear previous error
        else
          {
          int extLen=(int)strlen(s);
          char *ext=Memory_alloc(extLen+1,char);
          memcpy(ext,s,extLen+1);
          _CoreGl_mgr.extensions[_CoreGl_mgr.extensionCount++]=ext;
          }
        }
      _CoreGl_mgr.extensions[_CoreGl_mgr.extensionCount]=(char *)0;
      }
    }
  if(!_CoreGl_mgr.extensionCount) // fallback to legacy extension detection
    {
    const char *extList=(const char *)glGetString(GL_EXTENSIONS); // may fail
    if(!extList) glGetError(); // clear previous error
    else
      {
      CoreGl_addExtensionList(extList);
      }
    }
  CoreGl_addExtensionList(_CoreGl_querySpecificExtensions());
''')

for e in extensions:
  source.write(r'''  _CoreGl_mgr.has_%s=CoreGl_load_%s();
'''%(e.name,e.name))

source.write(r'''  if(glCreateShader&&glDeleteShader)
    {
    int featureTest;
    featureTest=glCreateShader(GL_TESS_CONTROL_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasTessControlShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      }
    featureTest=glCreateShader(GL_TESS_EVALUATION_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasTessEvaluationShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      }
    featureTest=glCreateShader(GL_GEOMETRY_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasGeometryShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      }
    featureTest=glCreateShader(GL_COMPUTE_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasComputeShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      CoreGlComputeProperties *prop=&_CoreGl_mgr.computeProperties;
      glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_COUNT,0,
                      prop->maxWorkGroupCount+0);
      ON_GL_ERROR({ return false; });
      glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_COUNT,1,
                      prop->maxWorkGroupCount+1);
      ON_GL_ERROR({ return false; });
      glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_COUNT,2,
                      prop->maxWorkGroupCount+2);
      ON_GL_ERROR({ return false; });
      glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_SIZE,0,
                      prop->maxWorkGroupSize+0);
      ON_GL_ERROR({ return false; });
      glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_SIZE,1,
                      prop->maxWorkGroupSize+1);
      ON_GL_ERROR({ return false; });
      glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_SIZE,2,
                      prop->maxWorkGroupSize+2);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_WORK_GROUP_INVOCATIONS,
                    &prop->maxWorkGroupInvocations);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_UNIFORM_BLOCKS,
                    &prop->maxUniformBlocks);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_TEXTURE_IMAGE_UNITS,
                    &prop->maxTextureImageUnits);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_ATOMIC_COUNTER_BUFFERS,
                    &prop->maxAtomicCounterBuffers);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_ATOMIC_COUNTERS,
                    &prop->maxAtomicCounters);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_SHARED_MEMORY_SIZE,
                    &prop->maxSharedMemorySize);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_UNIFORM_COMPONENTS,
                    &prop->maxUniformComponents);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_IMAGE_UNIFORMS,
                    &prop->maxImageUniforms);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMBINED_COMPUTE_UNIFORM_COMPONENTS,
                    &prop->maxCombinedUniformComponents);
      ON_GL_ERROR({ return false; });
      glGetIntegerv(GL_MAX_COMPUTE_SHADER_STORAGE_BLOCKS,
                    &prop->maxShaderStorageBlocks);
      ON_GL_ERROR({ return false; });
      bool useWarpInfo=false;
      prop->hasBallot=CoreGl_hasExtension("GL_NV_shader_thread_group");
      if(prop->hasBallot)
        {
        glGetIntegerv(0x9339, // WARP_SIZE_NV
                      &prop->warpSize);
        ON_GL_ERROR({ return false; });
        glGetIntegerv(0x933A, // WARPS_PER_SM_NV
                      &prop->warpsPerStreamingMultiprocessor);
        ON_GL_ERROR({ return false; });
        glGetIntegerv(0x933B, // SM_COUNT_NV
                      &prop->streamingMultiprocessorCount);
        ON_GL_ERROR({ return false; });
        prop->hasShuffle=CoreGl_hasExtension("GL_NV_shader_thread_shuffle");
        if((prop->warpSize>=2)&&
           !(prop->warpSize&(prop->warpSize-1)))
          {
          useWarpInfo=true; // ensure warpSize is a power of 2
          }
        }
      if(!useWarpInfo)
        {
        prop->warpSize=1;
        const char *vendor=(const char *)glGetString(GL_VENDOR);
        if(!vendor) glGetError(); // clear previous error
        else // try to guess warp size from the vendor
          {
          // FIXME: these values might change one day or another...
          if(strstr(vendor,"NVIDIA"))
            {
            prop->warpSize=32;
            }
          else if(strstr(vendor,"ATI"))
            {
            prop->warpSize=64;
            }
          }
        prop->warpsPerStreamingMultiprocessor=1;
        prop->streamingMultiprocessorCount=1;
        prop->hasBallot=false;
        prop->hasShuffle=false;
        }
      }
    }
  _CoreGl_mgr.initDone=true;
  }
return true;
}

bool // success
CoreGl_hasExtension(const char *extensionName)
{
int count;
const char * const *ext=CoreGl_extensions(&count);
for(int i=0;i<count;++i)
  {
  if(!strcmp(ext[i],extensionName))
    {
    return true;
    }
  }
return false;
}

void * // symbol or 0
CoreGl_extensionSymbol(const char *symbolName)
{
void *symbol=(void *)0;
#if defined _WIN32
  PROC procAddr=wglGetProcAddress(symbolName);
  memcpy(&symbol,&procAddr,sizeof(void *));
#elif defined __APPLE__
  // FIXME: no ???GetProcAddress() utility
#elif defined __ANDROID__
  // FIXME: eglGetProcAddress() only works for extension functions
  // void (*procAddr)(void)=eglGetProcAddress(symbolName);
  // memcpy(&symbol,&procAddr,sizeof(void *));
#else 
  void (*procAddr)(void)=glXGetProcAddress((const GLubyte *)symbolName);
  memcpy(&symbol,&procAddr,sizeof(void *));
#endif
if(!symbol) // not found with a specific utility... may be the application
  {         // is linked against GL...
  static void *currentBinary=(void *)0;
  if(!currentBinary)
    {
#if defined _WIN32
    currentBinary=(void *)GetModuleHandle((LPCTSTR)0);
#else
    currentBinary=dlopen((const char *)0,RTLD_LAZY);
#endif
    }
  if(currentBinary)
    {
#if defined _WIN32
    FARPROC procAddr=GetProcAddress((HMODULE)currentBinary,symbolName);
    memcpy(&symbol,&procAddr,sizeof(void *));
#else
    symbol=dlsym(currentBinary,symbolName);
#endif
    }
  }
if(!symbol) // not found in the current binary... may be the application
  {         // is a dynamic library which is linked against GL...
  static void *glLibrary=(void *)0;
  if(!glLibrary)
    {
#if defined _WIN32
    glLibrary=(void *)LoadLibrary(GL_LIBRARY_NAME);
#else
    glLibrary=dlopen(GL_LIBRARY_NAME,RTLD_LAZY);
#endif
    }
  if(glLibrary)
    {
#if defined _WIN32
    FARPROC procAddr=GetProcAddress((HMODULE)glLibrary,symbolName);
    memcpy(&symbol,&procAddr,sizeof(void *));
#else
    symbol=dlsym(glLibrary,symbolName);
#endif
    }
  }
return symbol;
}

void * // symbol or 0
CoreGl_extensionSymbolWithSuffix(const char *symbolName,
                                 const char * const *suffixList)
{
for(int i=0;suffixList[i];++i)
  {
  char fullName[0x100];
  snprintf(fullName,0x100,"%s%s",symbolName,suffixList[i]);
  fullName[0xFF]='\0';
  void *symbol=CoreGl_extensionSymbol(fullName);
  if(symbol)
    {
    return symbol;
    }
  }
return (void *)0;
}

const char * // last OpenGL error message if any
CoreGl_lastError(void)
{
int err=glGetError();
switch(err)
  {
  case GL_NO_ERROR:
    {
    return (const char *)0;
    }
  case GL_INVALID_ENUM:
    {
    return "GL_INVALID_ENUM";
    }
  case GL_INVALID_VALUE:
    {
    return "GL_INVALID_VALUE";
    }
  case GL_INVALID_OPERATION:
    {
    return "GL_INVALID_OPERATION";
    }
  case GL_STACK_OVERFLOW:
    {
    return "GL_STACK_OVERFLOW";
    }
  case GL_STACK_UNDERFLOW:
    {
    return "GL_STACK_UNDERFLOW";
    }
  case GL_OUT_OF_MEMORY:
    {
    return "GL_OUT_OF_MEMORY";
    }
  case GL_INVALID_FRAMEBUFFER_OPERATION:
    {
    return "GL_INVALID_FRAMEBUFFER_OPERATION";
    }
  default:
    {
    static char buffer[0x20];
    sprintf(buffer,"unknown OpenGL error 0x%.4X",err);
    return buffer;
    }
  }
}

char * // allocated compilation/link error/warning message if any
CoreGl_shaderStatus(GLuint shader,
                    bool *out_error)
{
bool isProgram;
GLint status=GL_FALSE;
GLint len=0;
if(glIsProgram(shader))
  {
  isProgram=true;
  glGetProgramiv(shader,GL_LINK_STATUS,&status);
  ON_GL_ERROR({});
  glGetProgramiv(shader,GL_INFO_LOG_LENGTH,&len);
  ON_GL_ERROR({});
  }
else
  {
  isProgram=false;
  glGetShaderiv(shader,GL_COMPILE_STATUS,&status);
  ON_GL_ERROR({});
  glGetShaderiv(shader,GL_INFO_LOG_LENGTH,&len);
  ON_GL_ERROR({});
  }
char *log=(char *)0;
*out_error=!status;
if(len>1)
  {
  log=Memory_initAlloc(len,char);
  GLint realLen=0;
  if(isProgram)
    {
    glGetProgramInfoLog(shader,len,&realLen,log);
    ON_GL_ERROR({ Memory_free(log); log=(char *)0; });
    }
  else
    {
    glGetShaderInfoLog(shader,len,&realLen,log);
    ON_GL_ERROR({ Memory_free(log); log=(char *)0; });
    }
  if(log)
    {
    int start=0;
    // sometimes the log just contains one of these useless information lines
    if(!start)
      {
      sscanf(log," %*d lines, 0 errors.%n",&start);
      }
    if(!start)
      {
      sscanf(log," Compiled%n",&start);
      }
    if(!start)
      {
      sscanf(log," --From Vertex Shader:"
                 " --From Fragment Shader:"
                 " Link was successful.%n",&start);
      }
    if(!start)
      {
      sscanf(log," %*s shader was successfully"
                 " compiled to run on hardware.%n",&start);
      }
    if(!start)
      {
      sscanf(log," Vertex shader(s) linked,"
                 " fragment shader(s) linked.%n",&start);
      }
    bool empty=true;
    for(int i=start;(i<len)&&log[i];++i)
      {
      if(!isspace(log[i]))
        {
        empty=false;
        break;
        }
      }
    if(empty)
      {
      Memory_free(log);
      log=(char *)0;
      }
    }
  }
if(!status&&!log)
  {
  log=Memory_alloc(64,char);
  if(isProgram)
    {
    sprintf(log,"shader program link failure\n");
    }
  else
    {
    sprintf(log,"shader compilation failure\n");
    }
  }
return log;
}

GLuint // compiled shader-object or 0
CoreGl_compileShader(GLenum shaderKind,
                     const char *sourceCode,
                     bool addPreamble,
                     char **out_allocatedCompilationLog)
{
*out_allocatedCompilationLog=(char *)0;
const char *esStr=CoreGl_embedded()&&(CoreGl_shadingLanguageVersion()>=300)
                  ? " es" : "";
char glslVersionStr[64];
sprintf(glslVersionStr," #version %u%s \n",
                       CoreGl_shadingLanguageVersion(),
                       esStr);
const char *preamble=""; // compose preamble to ease portability
const char *shaderKindStr="???";
switch(shaderKind)
  {
  case GL_VERTEX_SHADER:
    {
    shaderKindStr="VERTEX";
    preamble=" #if __VERSION__<130                                        \n"
             " # define in attribute                                      \n"
             " # define out varying                                       \n"
             " # define smooth                                            \n"
             " # define flat                                              \n"
             " # define noperspective                                     \n"
             " #endif                                                     \n";
    break;
    }
  case GL_GEOMETRY_SHADER:
    {
    shaderKindStr="GEOMETRY";
    preamble=" #if __VERSION__<130                                        \n"
             " # define in varying in                                     \n"
             " # define out varying out                                   \n"
             " # define smooth                                            \n"
             " # define flat                                              \n"
             " # define noperspective                                     \n"
             " #endif                                                     \n"
             " #if __VERSION__>=150                                       \n"
             " # define gl_PosIn(i) (gl_in[i].gl_Position)                \n"
             " #else                                                      \n"
             " # extension GL_EXT_geometry_shader4 : enable               \n"
             " # define gl_PosIn(i) (gl_PositionIn[i])                    \n"
             " #endif                                                     \n";
    break;
    }
  case GL_FRAGMENT_SHADER:
    {
    shaderKindStr="FRAGMENT";
    preamble=" #if defined GL_ES                                          \n"
             " # if defined GL_FRAGMENT_PRECISION_HIGH                    \n"
             "     precision highp float;                                 \n"
             " # else                                                     \n"
             "     precision mediump float;                               \n"
             " # endif                                                    \n"
             " #endif                                                     \n"
             " #if __VERSION__<130                                        \n"
             " # define in varying                                        \n"
             " # define smooth                                            \n"
             " # define flat                                              \n"
             " # define noperspective                                     \n"
             " #else                                                      \n"
             " # define texture2D texture                                 \n"
             " # define shadow2D texture                                  \n"
             " # define texture3D texture                                 \n"
             " # define textureCube texture                               \n"
             " # if __VERSION__>=330                                      \n"
             "     layout(location=0)                                     \n"
             " # endif                                                    \n"
             "   out vec4 defaultFragColour;                              \n"
             " # define gl_FragColor defaultFragColour                    \n"
             " #endif                                                     \n";
    break;
    }
  case GL_TESS_EVALUATION_SHADER:
    {
    shaderKindStr="TESS_EVALUATION";
    break;
    }
  case GL_TESS_CONTROL_SHADER:
    {
    shaderKindStr="TESS_CONTROL";
    break;
    }
  case GL_COMPUTE_SHADER:
    {
    shaderKindStr="COMPUTE";
    break;
    }
  }
GLuint shader=glCreateShader(shaderKind);
ON_GL_ERROR({ return 0; });
const char *strings[3]={ glslVersionStr, preamble, sourceCode };
if(!addPreamble)
  {
  strings[0]="";
  strings[1]="";
  }
glShaderSource(shader,3,strings,NULL);
ON_GL_ERROR({ glDeleteShader(shader); return 0; });
glCompileShader(shader);
ON_GL_ERROR({ glDeleteShader(shader); return 0; });
bool compilationError;
char *log=CoreGl_shaderStatus(shader,&compilationError);
if(log)
  {
  int srcCharCount=0,srcLineCount=1;
  for(int s=0;s<3;++s)
    {
    const char *src=strings[s];
    for(int i=0;src[i];++i)
      {
      ++srcCharCount;
      if(src[i]=='\n')
        {
        ++srcLineCount;
        }
      }
    }
  int lineCountWidth=1;
  int lineCountMax=10;
  while(lineCountMax<=srcLineCount)
    {
    ++lineCountWidth;
    lineCountMax*=10;
    }
  char *newLog=Memory_alloc(srcCharCount+srcLineCount*(lineCountWidth+1)+
                            strlen(log)+0x100,char);
  int pos=sprintf(newLog,"%s SHADER SOURCE:\n",shaderKindStr);
  srcLineCount=1;
  bool newLine=true;
  for(int s=0;s<3;++s)
    {
    const char *src=strings[s];
    for(int i=0;src[i];++i)
      {
      if(newLine)
        {
        newLine=false;
        pos+=sprintf(newLog+pos,"%*d:",lineCountWidth,srcLineCount);
        }
      newLog[pos++]=src[i];
      if(src[i]=='\n')
        {
        ++srcLineCount;
        newLine=true;
        }
      }
    }
  if(!newLine)
    {
    newLog[pos++]='\n';
    }
  sprintf(newLog+pos,"%s SHADER LOG:\n%s\n",shaderKindStr,log);
  Memory_free(log);
  *out_allocatedCompilationLog=newLog;
  }
if(compilationError)
  {
  glDeleteShader(shader);
  ON_GL_ERROR({});
  return 0;
  }
return shader; // sucessfully compiled shader-object
}

GLuint // linked shader-program or 0
CoreGl_buildRenderShader(const char *vertexSrc,
                         const char *geometrySrc,
                         const char *fragmentSrc,
                         bool addPreamble,
                         char **out_allocatedBuildLog)
{
*out_allocatedBuildLog=(char *)0;
const char *sources[3]={ vertexSrc, geometrySrc, fragmentSrc };
GLenum kinds[3]={ GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER };
GLuint shaders[3]={ 0, 0, 0 };
GLuint program=glCreateProgram();
ON_ERROR(!program,"glCreateProgram() failure", { goto undo; });
bool tryToLink=true;
for(int i=0;i<3;++i)
  {
  if(sources[i]&&*sources[i])
    {
    char *log=(char *)0;
    shaders[i]=CoreGl_compileShader(kinds[i],sources[i],addPreamble,&log);
    if(log)
      {
      if(*out_allocatedBuildLog)
        {
        char *oldLog=*out_allocatedBuildLog;
        *out_allocatedBuildLog=Memory_alloc(strlen(oldLog)+strlen(log)+1,
                                            char);
        sprintf(*out_allocatedBuildLog,"%s%s",oldLog,log);
        Memory_free(oldLog);
        }
      else
        {
        *out_allocatedBuildLog=log;
        }
      }
    if(shaders[i])
      {
      glAttachShader(program,shaders[i]);
      ON_GL_ERROR({ goto undo; });
      if((kinds[i]==GL_GEOMETRY_SHADER)&&
         (CoreGl_shadingLanguageVersion()<150))
        {
        // ugly workaround: parse source to detect layout parameters
        int inputType=-1;
        int outputType=-1;
        int verticesOut=-1;
        char line[0x400];
        const char *begin=sources[i];
        const char *end=begin+strlen(begin);
        while(begin<end)
          {
          const char *cr=strchr(begin,'\n');
          if(!cr) cr=end;
          int len=(int)(cr-begin);
          memcpy(line,begin,len);
          line[len]='\0';
          char txt[0x100];
          int value;
          int charCount=0;
          if((sscanf(line," layout ( %[^)] ) in ; %n",
                     txt,&charCount)==1)&&charCount)
            {
            if(!strcmp(txt,"points"))
              {
              inputType=GL_POINTS;
              }
            else if(!strcmp(txt,"lines"))
              {
              inputType=GL_LINES;
              }
            else if(!strcmp(txt,"lines_adjacency"))
              {
              inputType=GL_LINES_ADJACENCY;
              }
            else if(!strcmp(txt,"triangles"))
              {
              inputType=GL_TRIANGLES;
              }
            else if(!strcmp(txt,"triangles_adjacency"))
              {
              inputType=GL_TRIANGLES_ADJACENCY;
              }
            }
          charCount=0;
          if((sscanf(line," layout ( %[^),] , max_vertices = %d ) out ; %n",
                     txt,&value,&charCount)==2)&&charCount)
            {
            if(!strcmp(txt,"points"))
              {
              outputType=GL_POINTS;
              }
            else if(!strcmp(txt,"line_strip"))
              {
              outputType=GL_LINE_STRIP;
              }
            else if(!strcmp(txt,"triangle_strip"))
              {
              outputType=GL_TRIANGLE_STRIP;
              }
            verticesOut=value;
            }
          begin=cr+1;
          }
        // use GL_EXT_geometry_shader4 layout parameters
        if(inputType>=0)
          {
          glProgramParameteri(program,
                              0x8DDB, // GL_GEOMETRY_INPUT_TYPE_EXT
                              inputType);
          ON_GL_ERROR({});
          }
        if(outputType>=0)
          {
          glProgramParameteri(program,
                              0x8DDC, // GL_GEOMETRY_OUTPUT_TYPE_EXT
                              outputType);
          ON_GL_ERROR({});
          }
        if(verticesOut>0)
          {
          glProgramParameteri(program,
                              0x8DDA, // GL_GEOMETRY_VERTICES_OUT_EXT
                              verticesOut);
          ON_GL_ERROR({});
          }
        }
      }
    else
      {
      tryToLink=false;
      }
    }
  }
if(tryToLink)
  {
  glLinkProgram(program);
  ON_GL_ERROR({ goto undo; });
  bool linkError;
  char *log=CoreGl_shaderStatus(program,&linkError);
  if(log)
    {
    if(*out_allocatedBuildLog)
      {
      char *oldLog=*out_allocatedBuildLog;
      *out_allocatedBuildLog=Memory_alloc(strlen(oldLog)+strlen(log)+0x100,
                                          char);
      sprintf(*out_allocatedBuildLog,"%sSHADER PROGRAM LOG:\n%s\n",
                                     oldLog,log);
      Memory_free(oldLog);
      }
    else
      {
      *out_allocatedBuildLog=Memory_alloc(strlen(log)+0x100,char);
      sprintf(*out_allocatedBuildLog,"SHADER PROGRAM LOG:\n%s\n",log);
      }
    Memory_free(log);
    }
  if(linkError)
    {
    goto undo;
    }
  for(int i=0;i<3;++i)
    {
    if(shaders[i])
      {
      glDeleteShader(shaders[i]); // still referenced by program
      ON_GL_ERROR({});
      }
    }
  return program; // sucessfully built shader-program
  }
undo:
if(program)
  {
  glDeleteProgram(program);
  ON_GL_ERROR({});
  }
for(int i=0;i<3;++i)
  {
  if(shaders[i])
    {
    glDeleteShader(shaders[i]);
    ON_GL_ERROR({});
    }
  }
return 0;
}

GLuint // linked shader-program or 0
CoreGl_buildComputeShader(const char *computeSrc,
                          bool addPreamble,
                          char **out_allocatedBuildLog)
{
*out_allocatedBuildLog=(char *)0;
GLuint shader=0;
GLuint program=glCreateProgram();
ON_ERROR(!program,"glCreateProgram() failure", { goto undo; });
shader=CoreGl_compileShader(GL_COMPUTE_SHADER,computeSrc,addPreamble,
                            out_allocatedBuildLog);
if(!shader) { goto undo; }
glAttachShader(program,shader);
ON_GL_ERROR({ goto undo; });
glLinkProgram(program);
ON_GL_ERROR({ goto undo; });
bool linkError;
char *log=CoreGl_shaderStatus(program,&linkError);
if(log)
  {
  if(*out_allocatedBuildLog)
    {
    char *oldLog=*out_allocatedBuildLog;
    *out_allocatedBuildLog=Memory_alloc(strlen(oldLog)+strlen(log)+0x100,
                                        char);
    sprintf(*out_allocatedBuildLog,"%sSHADER PROGRAM LOG:\n%s\n",
                                   oldLog,log);
    Memory_free(oldLog);
    }
  else
    {
    *out_allocatedBuildLog=Memory_alloc(strlen(log)+0x100,char);
    sprintf(*out_allocatedBuildLog,"SHADER PROGRAM LOG:\n%s\n",log);
    }
  Memory_free(log);
  }
if(linkError)
  {
  goto undo;
  }
glDeleteShader(shader); // still referenced by program
ON_GL_ERROR({});
return program; // sucessfully built shader-program
undo:
if(program)
  {
  glDeleteProgram(program);
  ON_GL_ERROR({});
  }
if(shader)
  {
  glDeleteShader(shader);
  ON_GL_ERROR({});
  }
return 0;
}

''')

for e in extensions:
  header.write(r'''//---- %s ----

INLINE
bool
CoreGl_has_%s(void)
{
return _CoreGl_mgr.has_%s;
}

'''%(e.name,e.name,e.name))
  source.write(r'''//---- %s ----

'''%e.name)
  for a in e.apiEntries:
    header.write(r'''extern TRANSGL_API_VAR
PFNGL%sPROC
CoreGl_%s;
#define gl%s \
        CoreGl_%s

'''%(a.typeName,a.name,a.name,a.name))
    source.write(r'''TRANSGL_API_VAR
PFNGL%sPROC
CoreGl_%s;

'''%(a.typeName,a.name))

header.write(r'''#endif // TRANSGL_COREGL_H

//'''+'-'*76+'\n')
source.write('//'+'-'*76+'\n')
header.close()
source.close()

#-----------------------------------------------------------------------------
