#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
#   http://www.enib.fr/~harrouet/transprog.html
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

content='//'+'-'*76+r'''
// Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
// 
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef GUARD_MACRO
#define GUARD_MACRO 1
#line 44 THIS_FILE

#include "TransProg/memory.h"

SPECIFIC_CODE

typedef ELEM_TYPE DEQUE_TYPE_ElemType;

typedef struct _DEQUE_TYPE
{
DEQUE_TYPE_ElemType *data;
intptr_t offset;
intptr_t size;
intptr_t capacity;
intptr_t mask;
} DEQUE_TYPE;

INLINE
void
DEQUE_TYPE_init(DEQUE_TYPE *d)
{
d->data=(DEQUE_TYPE_ElemType *)0;
d->offset=0;
d->size=0;
d->capacity=0;
d->mask=0;
}

INLINE
DEQUE_TYPE *
DEQUE_TYPE_new(void)
{
DEQUE_TYPE *d=Memory_alloc(1,DEQUE_TYPE);
DEQUE_TYPE_init(d);
return d;
}

INLINE
void
DEQUE_TYPE_destroy(DEQUE_TYPE *d)
{
if(d->data)
  {
  Memory_free(d->data);
  }
}

INLINE
void
DEQUE_TYPE_delete(DEQUE_TYPE *d)
{
DEQUE_TYPE_destroy(d);
Memory_free(d);
}

INLINE
void
DEQUE_TYPE_numaUsage(const DEQUE_TYPE *d,
                     intptr_t *numaUsage)
{
if(d->data)
  {
  Memory_numaUsage(d->data,numaUsage);
  }
}

INLINE
intptr_t
DEQUE_TYPE_capacity(const DEQUE_TYPE *d)
{
return d->capacity;
}

INLINE
void
DEQUE_TYPE_shrink(DEQUE_TYPE *d)
{
if(d->size)
  {
  intptr_t newCapacity=1;
  while(newCapacity<d->size)
    {
    newCapacity<<=1;
    }
  if(newCapacity!=d->capacity)
    {
    DEQUE_TYPE_ElemType *data=Memory_alloc(newCapacity,
                                           DEQUE_TYPE_ElemType);
    intptr_t end=d->offset+d->size;
    if(end>d->capacity)
      {
      intptr_t firstPart=d->capacity-d->offset;
      memcpy(data,d->data+d->offset,firstPart*sizeof(DEQUE_TYPE_ElemType));
      memcpy(data+firstPart,d->data,
             (end-d->capacity)*sizeof(DEQUE_TYPE_ElemType));
      }
    else
      {
      memcpy(data,d->data+d->offset,d->size*sizeof(DEQUE_TYPE_ElemType));
      }
    Memory_free(d->data);
    d->data=data;
    d->offset=0;
    d->capacity=newCapacity;
    d->mask=newCapacity-1;
    }
  }
else if(d->data)
  {
  Memory_free(d->data);
  d->data=(DEQUE_TYPE_ElemType *)0;
  d->offset=0;
  d->capacity=0;
  d->mask=0;
  }
}

INLINE
void
DEQUE_TYPE_ensureMinCapacity(DEQUE_TYPE *d,
                             intptr_t capacity)
{
intptr_t newCapacity=d->capacity ? d->capacity : 16;
while(newCapacity<capacity)
  {
  newCapacity<<=1;
  }
if(newCapacity>d->capacity)
  {
  d->data=Memory_reAlloc(d->data,newCapacity,DEQUE_TYPE_ElemType);
  intptr_t end=d->offset+d->size;
  if(end>d->capacity)
    {
    memcpy(d->data+d->capacity,d->data,
           (end-d->capacity)*sizeof(DEQUE_TYPE_ElemType));
    }
  d->capacity=newCapacity;
  d->mask=newCapacity-1;
  }
}

INLINE
void
DEQUE_TYPE_clear(DEQUE_TYPE *d)
{
d->size=0;
}

INLINE
intptr_t
DEQUE_TYPE_size(const DEQUE_TYPE *d)
{
return d->size;
}

INLINE
DEQUE_TYPE_ElemType // index'th element
DEQUE_TYPE_get(const DEQUE_TYPE *d,
               intptr_t index)
{
return d->data[(d->offset+index)&d->mask];
}

INLINE
DEQUE_TYPE_ElemType // first element
DEQUE_TYPE_first(const DEQUE_TYPE *d)
{
return DEQUE_TYPE_get(d,0);
}

INLINE
DEQUE_TYPE_ElemType // last element
DEQUE_TYPE_last(const DEQUE_TYPE *d)
{
return DEQUE_TYPE_get(d,d->size-1);
}

INLINE
void
DEQUE_TYPE_set(DEQUE_TYPE *d,
               intptr_t index,
               DEQUE_TYPE_ElemType element)
{
d->data[(d->offset+index)&d->mask]=element;
}

INLINE
void
DEQUE_TYPE_pushBack(DEQUE_TYPE *d,
                    DEQUE_TYPE_ElemType element)
{
if(d->size==d->capacity)
  {
  DEQUE_TYPE_ensureMinCapacity(d,d->size+1);
  }
d->data[(d->offset+d->size++)&d->mask]=element;
}

INLINE
void
DEQUE_TYPE_pushFront(DEQUE_TYPE *d,
                     DEQUE_TYPE_ElemType element)
{
if(d->size==d->capacity)
  {
  DEQUE_TYPE_ensureMinCapacity(d,d->size+1);
  }
++(d->size);
--(d->offset);
d->data[d->offset&=d->mask]=element;
}

INLINE
DEQUE_TYPE_ElemType
DEQUE_TYPE_popBack(DEQUE_TYPE *d)
{
return d->data[(d->offset+--(d->size))&d->mask];
}

INLINE
DEQUE_TYPE_ElemType
DEQUE_TYPE_popFront(DEQUE_TYPE *d)
{
--(d->size);
intptr_t index=d->offset++;
d->offset&=d->mask;
return d->data[index];
}

#endif // GUARD_MACRO

//----------------------------------------------------------------------------
'''

import sys
import os

if len(sys.argv)<4:
  sys.stderr.write(
             'usage: %s file_name element_type container_type [precode ...]\n'
             % sys.argv[0])
  sys.exit(1)

fileName=sys.argv[1]
elementType=sys.argv[2]
containerType=sys.argv[3]

guardMacro=os.path.normpath(fileName).upper().replace(os.path.sep,'_') \
                                             .replace(os.path.extsep,'_') \
                                             .replace(':','_')

thisFile='"%s"' % os.path.normpath(sys.argv[0]).replace(os.path.sep,'/')

specificCode='\n'.join(sys.argv[4:])
if specificCode:
  specificCode='\n// specific code\n'+specificCode

open(fileName,'w').write(content.replace('ELEM_TYPE',elementType)
                                .replace('DEQUE_TYPE',containerType)
                                .replace('GUARD_MACRO',guardMacro)
                                .replace('THIS_FILE',thisFile)
                                .replace('SPECIFIC_CODE',specificCode))

