#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
#   http://www.enib.fr/~harrouet/transprog.html
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

content='//'+'-'*76+r'''
// Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
// 
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef GUARD_MACRO
#define GUARD_MACRO 1
#line 44 THIS_FILE

#include "TransProg/platform.h"

SPECIFIC_CODE

typedef ELEM_TYPE FIFO_TYPE_ElemType;

typedef struct _FIFO_TYPE
{
volatile FIFO_TYPE_ElemType *storage;
volatile intptr_t readIndex;
volatile intptr_t writeIndex;
intptr_t capacity;
} FIFO_TYPE;

INLINE
void
FIFO_TYPE_init(FIFO_TYPE *f,
               FIFO_TYPE_ElemType *storage,
               intptr_t capacity)
{
f->storage=storage;
f->readIndex=0;
f->writeIndex=0;
f->capacity=1;
while((f->capacity<<1)<=capacity)
  {
  f->capacity<<=1;
  }
--(f->capacity);
}

INLINE
void
FIFO_TYPE_destroy(FIFO_TYPE *f)
{
(void)f; // nothing to be done
}

INLINE
bool
FIFO_TYPE_empty(const FIFO_TYPE *f)
{
return f->writeIndex==f->readIndex;
}

INLINE
bool
FIFO_TYPE_full(const FIFO_TYPE *f)
{
return ((f->writeIndex+1)&f->capacity)==f->readIndex;
}

INLINE
intptr_t // number of elements stored
FIFO_TYPE_pending(const FIFO_TYPE *f)
{
return (f->writeIndex+f->capacity+1-f->readIndex)&f->capacity;
}

INLINE
bool // success
FIFO_TYPE_push(FIFO_TYPE *f,
               FIFO_TYPE_ElemType elem)
{
intptr_t nextWrite=(f->writeIndex+1)&f->capacity;
if(nextWrite!=f->readIndex)
  {
  f->storage[f->writeIndex]=elem;
  f->writeIndex=nextWrite;
  return true;
  }
return false;
}

INLINE
bool // success
FIFO_TYPE_pop(FIFO_TYPE *f,
              FIFO_TYPE_ElemType *out_elem)
{
if(f->writeIndex!=f->readIndex)
  {
  *out_elem=f->storage[f->readIndex];
  f->readIndex=(f->readIndex+1)&f->capacity;
  return true;
  }
return false;
}

#endif // GUARD_MACRO

//----------------------------------------------------------------------------
'''

import sys
import os

if len(sys.argv)<4:
  sys.stderr.write(
             'usage: %s file_name element_type container_type [precode ...]\n'
             % sys.argv[0])
  sys.exit(1)

fileName=sys.argv[1]
elementType=sys.argv[2]
containerType=sys.argv[3]

guardMacro=os.path.normpath(fileName).upper().replace(os.path.sep,'_') \
                                             .replace(os.path.extsep,'_') \
                                             .replace(':','_')

thisFile='"%s"' % os.path.normpath(sys.argv[0]).replace(os.path.sep,'/')

specificCode='\n'.join(sys.argv[4:])
if specificCode:
  specificCode='\n// specific code\n'+specificCode

open(fileName,'w').write(content.replace('ELEM_TYPE',elementType)
                                .replace('FIFO_TYPE',containerType)
                                .replace('GUARD_MACRO',guardMacro)
                                .replace('THIS_FILE',thisFile)
                                .replace('SPECIFIC_CODE',specificCode))

