#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
#   http://www.enib.fr/~harrouet/transprog.html
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

'''
  Windows tool to list DLL dependencies
  Similar to:
    ldd (Linux)
    otool -L (MacOSX)
'''

import sys
import os
import subprocess

class Structure:
  def __init__(self,**kw):
    self.__dict__=kw

def bytes_to_str(b):
  if type(b)!=str: return b.decode(sys.stdout.encoding)
  return b

recursion_level=1
target=''
target_dir=''
system_var=os.environ.get('SYSTEM')
system16_var=os.environ.get('SYSTEM16')
windir_var=os.environ.get('WINDIR')
path_var=os.environ.get('PATH').split(os.path.pathsep)
dumpbin_exe=''
objdump_exe=''
binList=[]

def findExecutable(name):
  name=os.path.normpath(name)
  dirs=[]
  if name.find(os.path.pathsep)!=-1: # path given
    dirs.append('')
  else:
    dirs.append(os.path.curdir)
    dirs+=path_var
  for i in dirs:
    i=os.path.join(i,name)
    if os.path.isfile(i):
      return i
  return None

def findLibrary(name):
  name=os.path.normpath(name)
  dirs=[]
  if name.find(os.path.pathsep)!=-1: # path given
    dirs.append('')
  else:
    if target_dir: dirs.append(target_dir)
    dirs.append(os.path.curdir)
    if system_var: dirs.append(system_var)
    if system16_var: dirs.append(system16_var)
    if windir_var: dirs.append(windir_var)
    dirs+=path_var
  for i in dirs:
    i=os.path.join(i,name)
    if os.path.isfile(i):
      return i
  return None

def listDependencies(binary):
  result=[]
  if dumpbin_exe:
    cmd=[dumpbin_exe,'/dependents',binary]
    p=subprocess.Popen(cmd,shell=False,stdout=subprocess.PIPE)
    for l in bytes_to_str(p.communicate()[0]).split('\n'):
      l=l.strip().split()
      if len(l)==1 and l[0].lower().endswith('.dll'):
        result.append(l[0])
  elif objdump_exe:
    cmd=[objdump_exe,'-p',binary]
    p=subprocess.Popen(cmd,shell=False,stdout=subprocess.PIPE)
    for l in bytes_to_str(p.communicate()[0]).split('\n'):
      l=l.strip().split()
      if len(l)==3 and l[0]=='DLL' and l[1]=='Name:' \
         and l[2].lower().endswith('.dll'):
        result.append(l[2])
  return result

def addDependency(name,level):
  for b in binList:
    if b.name==name:
      return False
  binList.append(Structure(name=name,path='',level=level))
  return True

for i in sys.argv[1:]:
  if i.startswith('-'):
    try:
      level=int(i[1:])
    except:
      level=None
    if level is None or level<0:
      sys.stdout.write('ignoring unknown option %s\n'%i)
    else:
      recursion_level=level
  else:
    target=i

if not target:
  sys.stdout.write('usage: %s [-recursion_level] exe_or_dll\n'%sys.argv[0])
  sys.exit(1)

dumpbin_exe=findExecutable('dumpbin.exe')
objdump_exe=findExecutable('objdump.exe')
addDependency(target,0)
b=binList[0]
b.path=findExecutable(b.name)
if b.path:
  target_dir=os.path.dirname(b.path)

level=1
index=0
while index<len(binList):
  b=binList[index]
  if b.level>level:
    sys.stdout.write('%s\n'%(78*'-'))
    level=b.level
  if not b.path:
    b.path=findLibrary(b.name)
  sys.stdout.write('%s => %s\n'%(b.name,b.path))
  if b.path and (not recursion_level or b.level<recursion_level):
    for d in listDependencies(b.path):
      addDependency(d,b.level+1)
  index+=1

