//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/glView.h"

#include "TransProg/ptrVector.h"

#include "glView.c"

#if !defined WGL_CONTEXT_MAJOR_VERSION_ARB
# define WGL_CONTEXT_MAJOR_VERSION_ARB 0x2091
#endif

#if !defined WGL_CONTEXT_MINOR_VERSION_ARB
# define WGL_CONTEXT_MINOR_VERSION_ARB 0x2092
#endif

#if !defined WGL_CONTEXT_LAYER_PLANE_ARB
# define WGL_CONTEXT_LAYER_PLANE_ARB 0x2093
#endif

#if !defined WGL_CONTEXT_FLAGS_ARB
# define WGL_CONTEXT_FLAGS_ARB 0x2094
#endif

#if !defined WGL_CONTEXT_PROFILE_MASK_ARB
# define WGL_CONTEXT_PROFILE_MASK_ARB 0x9126
#endif

#if !defined WGL_CONTEXT_DEBUG_BIT_ARB
# define WGL_CONTEXT_DEBUG_BIT_ARB 0x0001
#endif

#if !defined WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB
# define WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB 0x0002
#endif

#if !defined WGL_CONTEXT_CORE_PROFILE_BIT_ARB
# define WGL_CONTEXT_CORE_PROFILE_BIT_ARB 0x00000001
#endif

#if !defined WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB
# define WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB 0x00000002
#endif

static
HGLRC
(APIENTRY *createContextAttribsFnct)(HDC hDC,
                                     HGLRC,
				     const int *);

static
const char *
(APIENTRY *getExtensionsStringFnct)(HDC hDC);

typedef struct _Win32View
{
GlView *glView;
HWND window;
HDC hdc;
bool destroyFlag;
bool forceInitialConfig;
} Win32View;

typedef struct _Win32ViewManager
{
bool initDone;
int initCpuCtx;
Win32View *commonView;
PtrVector views;
Win32View *mouseView;
bool processingEvents;
int eventCount;
GfxStereoMode stereoMode;
HINSTANCE hInstance;
HCURSOR cursors[10];
HGLRC wglContext;
} Win32ViewManager;

static Win32ViewManager mgr;

//----------------------------------------------------------------------------

GfxStereoMode
GlView_stereoMode(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
return mgr.stereoMode;
}

int // screen width
GlView_screenWidth(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
RECT screen;
SystemParametersInfo(SPI_GETWORKAREA,0,&screen,0);
return screen.right-screen.left;
}

int // screen height
GlView_screenHeight(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
RECT screen;
SystemParametersInfo(SPI_GETWORKAREA,0,&screen,0);
return screen.bottom-screen.top;
}

static
void
Win32View_updateTitle(Win32View *v)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
SetWindowText(v->window,v->glView->title);
}

static
void
Win32View_updateCursor(Win32View *v)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
if(mgr.mouseView==v)
  {
  SetCursor(mgr.cursors[v->glView->cursor]);
  }
}

static
void
Win32View_processEvent(Win32View *v)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
if(!mgr.processingEvents)
  {
  // DROP_DEAD("event processed outside of event loop");
  return; // FIXME: should enqueue event
  }
if(v->destroyFlag)
  {
  return;
  }
GlView *view=v->glView;
RECT cr;
GetClientRect(v->window,&cr);
int width=cr.right-cr.left;
int height=cr.bottom-cr.top;
if(width||height) // avoid dummy 0x0 size during minimization
  {
  view->state.width=cr.right-cr.left;
  view->state.height=cr.bottom-cr.top;
  }
POINT pt;
GetCursorPos(&pt);
ScreenToClient(v->window,&pt);
view->state.mouseX=pt.x;
view->state.mouseY=pt.y;
view->state.modifiers=GFX_NO_MDF;
if(GetKeyState(VK_CONTROL)&0x8000) view->state.modifiers|=GFX_CONTROL_MDF;
if(GetKeyState(VK_SHIFT)&0x8000)   view->state.modifiers|=GFX_SHIFT_MDF;
if(GetKeyState(VK_CAPITAL)&0x0001) view->state.modifiers|=GFX_CAPSLOCK_MDF;
if(GetKeyState(VK_NUMLOCK)&0x0001) view->state.modifiers|=GFX_NUMLOCK_MDF;
if(GetKeyState(VK_LBUTTON)&0x8000) view->state.modifiers|=GFX_LEFT_MDF;
if(GetKeyState(VK_MBUTTON)&0x8000) view->state.modifiers|=GFX_MIDDLE_MDF;
if(GetKeyState(VK_RBUTTON)&0x8000) view->state.modifiers|=GFX_RIGHT_MDF;
if(!(*view->callback)(view))
  {
  v->destroyFlag=true;
  }
++mgr.eventCount;
}

static
LRESULT CALLBACK
Win32View_wndProc(HWND hWnd,
                  UINT uMsg,
                  WPARAM wParam,
                  LPARAM lParam)
{
// FIXME: ugly hack!
//        Window$ doesn't know enter/leave events!
//        Trying to track the mouse in order to detect
//        the window which is under the mouse.
if(uMsg==WM_NCHITTEST)
  {
  Win32View *old=mgr.mouseView;
  mgr.mouseView=(Win32View *)0;
  LRESULT hitTest=DefWindowProc(hWnd,uMsg,wParam,lParam);
  if(hitTest==HTCLIENT)
    {
    for(intptr_t i=0,i_end=PtrVector_size(&mgr.views);i<i_end;++i)
      {
      Win32View *v=(Win32View *)PtrVector_get(&mgr.views,i);
      if(v->window==hWnd)
        {
        mgr.mouseView=v;
        break;
        }
      }
    if(mgr.mouseView&&(mgr.mouseView!=old))
      {
      Win32View_updateCursor(mgr.mouseView);
      }
    }
  return hitTest;
  }
GlView *view=(GlView *)0;
for(intptr_t i=0,i_end=PtrVector_size(&mgr.views);i<i_end;++i)
  {
  Win32View *v=(Win32View *)PtrVector_get(&mgr.views,i);
  if(v->window==hWnd)
    {
    view=v->glView;
    break;
    }
  }
if(view)
  {
  Win32View *v=(Win32View *)view->specific;
  switch(uMsg)
    {
    case WM_MOUSEMOVE:
      {
      view->state.event=GFX_MOTION_EVT;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_PAINT:
      {
      if(v->forceInitialConfig)
        {
        v->forceInitialConfig=false;
        view->state.event=GFX_CONFIG_EVT;
        Win32View_processEvent(v);
        }
      if(!wglMakeCurrent(v->hdc,mgr.wglContext))
        {
        SYS_DROP_DEAD("wglMakeCurrent() failure");
        }
      view->state.event=GFX_DRAW_EVT;
      Win32View_processEvent(v);
      SwapBuffers(v->hdc); // implicit glFlush()
      break; // use default procedure too!
      }
    case WM_LBUTTONDOWN:
      {
      SetCapture(v->window);
      view->state.event=GFX_BUTTON_EVT;
      view->state.button=GFX_LEFT_BTN;
      view->state.pressed=true;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_MBUTTONDOWN:
      {
      SetCapture(v->window);
      view->state.event=GFX_BUTTON_EVT;
      view->state.button=GFX_MIDDLE_BTN;
      view->state.pressed=true;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_RBUTTONDOWN:
      {
      SetCapture(v->window);
      view->state.event=GFX_BUTTON_EVT;
      view->state.button=GFX_RIGHT_BTN;
      view->state.pressed=true;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_LBUTTONUP:
      {
      ReleaseCapture();
      view->state.event=GFX_BUTTON_EVT;
      view->state.button=GFX_LEFT_BTN;
      view->state.pressed=false;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_MBUTTONUP:
      {
      ReleaseCapture();
      view->state.event=GFX_BUTTON_EVT;
      view->state.button=GFX_MIDDLE_BTN;
      view->state.pressed=false;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_RBUTTONUP:
      {
      ReleaseCapture();
      view->state.event=GFX_BUTTON_EVT;
      view->state.button=GFX_RIGHT_BTN;
      view->state.pressed=false;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_MOUSEWHEEL:
      {
      view->state.event=GFX_BUTTON_EVT;
      view->state.button=GET_WHEEL_DELTA_WPARAM(wParam)>0
                         ? GFX_UP_BTN : GFX_DOWN_BTN;
      view->state.pressed=true;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_KEYDOWN:
    case WM_KEYUP:
      {
      if(!view->keyRepeat&&(uMsg==WM_KEYDOWN)&&(lParam&(1<<30)))
        {
        return 0; // bit 30 --> key was already down, ignore
        }
      char asciiKey[2]={'\0','\0'};
      const char *key=(const char *)0;
      switch(wParam)
        {
        case VK_UP:      key="Up";       break;
        case VK_DOWN:    key="Down";     break;
        case VK_LEFT:    key="Left";     break;
        case VK_RIGHT:   key="Right";    break;
        case VK_PRIOR:   key="Prior";    break;
        case VK_NEXT:    key="Next";     break;
        case VK_HOME:    key="Home";     break;
        case VK_END:     key="End";      break;
        case VK_CLEAR:   key="Begin";    break;
        case VK_INSERT:  key="Insert";   break;
        case VK_DELETE:  key="Delete";   break;
        case VK_F1:      key="F1";       break;
        case VK_F2:      key="F2";       break;
        case VK_F3:      key="F3";       break;
        case VK_F4:      key="F4";       break;
        case VK_F5:      key="F5";       break;
        case VK_F6:      key="F6";       break;
        case VK_F7:      key="F7";       break;
        case VK_F8:      key="F8";       break;
        case VK_F9:      key="F9";       break;
        case VK_F10:     key="F10";      break;
        case VK_F11:     key="F11";      break;
        case VK_F12:     key="F12";      break;
        default:
          {
          BYTE keyboardState[256];
          if(GetKeyboardState(keyboardState))
            {
            keyboardState[VK_CONTROL]=0;  // ignore [Control] key
            keyboardState[VK_LCONTROL]=0;
            keyboardState[VK_RCONTROL]=0;
            WORD buffer=0;
            if(ToAscii((UINT)wParam,(UINT)lParam,keyboardState,&buffer,0)==1)
              {
              asciiKey[0]=(char)buffer;
              if(asciiKey[0]=='\r')
                {
                asciiKey[0]='\n';
                }
              key=asciiKey;
              }
            }
          }
        }
      if(!key)
        {
        key="";
        }
      view->state.event=GFX_KEY_EVT;
      int maxLen=sizeof(view->state.key)-1;
      strncpy(view->state.key,key,maxLen);
      view->state.key[maxLen]='\0';
      view->state.pressed=uMsg==WM_KEYDOWN ? true : false;
      Win32View_processEvent(v);
      return 0;
      }
    case WM_SIZE:
      {
      if(v->glView->callback) // avoid during opening
        {
        InvalidateRgn(v->window,(HRGN)0,FALSE);
        view->state.event=GFX_CONFIG_EVT;
        Win32View_processEvent(v);
        }
      return 0;
      }
    case WM_SHOWWINDOW:
      {
      view->state.visible=wParam==TRUE ? true : false;
      if(v->glView->callback) // avoid during opening
        {
        view->state.event=GFX_CONFIG_EVT;
        Win32View_processEvent(v);
        }
      break; // use default procedure too!
      }
    case WM_SETFOCUS:
      {
      Win32View_updateCursor(v);
      return 0;
      }
    case WM_SYSCOMMAND:
      {
      bool process=false;
      switch(wParam)
        {
        case SC_MINIMIZE:
          {
          view->state.visible=false;
          process=true;
          break;
          }
        case SC_MAXIMIZE:
        case SC_RESTORE:
          {
          view->state.visible=true;
          process=true;
          break;
          }
        }
      if(process)
        {
        view->state.event=GFX_CONFIG_EVT;
        Win32View_processEvent(v);
        }
      break; // use default procedure too!
      }
    case WM_CLOSE:
      {
      view->state.event=GFX_CLOSE_EVT;
      Win32View_processEvent(v);
      return 0;
      }
    }
  }
return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

int // events occured
GlView_processEvents(int64_t usecTimeout)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
mgr.eventCount=0;
if(usecTimeout)
  {
  DWORD timeout=(usecTimeout>0) ? (DWORD)(usecTimeout/1000) : INFINITE;
  if(MsgWaitForMultipleObjects(0,NULL,FALSE,timeout,QS_ALLEVENTS)
     ==WAIT_FAILED)
    {
    SYS_DROP_DEAD("MsgWaitForMultipleObjects() failure");
    }
  }
mgr.processingEvents=true;
MSG msg;
while(PeekMessage(&msg,(HWND)0,0,0,PM_REMOVE))
  {
  TranslateMessage(&msg);
  DispatchMessage(&msg);
  }
mgr.processingEvents=false;
for(intptr_t i=PtrVector_size(&mgr.views);i--;)
  {
  Win32View *v=(Win32View *)PtrVector_get(&mgr.views,i);
  if(v->destroyFlag)
    {
    if(mgr.mouseView==v)
      {
      mgr.mouseView=(Win32View *)0;
      }
    PtrVector_erase(&mgr.views,
                    PtrVector_find(&mgr.views,(PtrVector_CmpFnct)0,v));
    GlView *view=v->glView;
    if(!wglMakeCurrent(mgr.commonView->hdc,mgr.wglContext))
      {
      SYS_DROP_DEAD("wglMakeCurrent() failure");
      }
    ReleaseDC(v->window,v->hdc);
    DestroyWindow(v->window);
    Memory_free(v);
    Memory_free(view);
    }
  }
return mgr.eventCount;
}

void
GlView_sync(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
glFinish();
GdiFlush(); // FIXME: isn't there any XSync equivalent?
}

static
void
Win32View_open(GlView *view,
               int initX,
               int initY,
               bool borderless)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
Win32View *v=Memory_alloc(1,Win32View);
v->glView=view;
v->destroyFlag=false;
v->forceInitialConfig=true;
DWORD dwStyle=borderless ? WS_POPUP|WS_SYSMENU|WS_MINIMIZEBOX|WS_MAXIMIZEBOX
                         : WS_OVERLAPPEDWINDOW;
DWORD dwExStyle=0;
v->window=CreateWindowEx(dwExStyle,"TransGl-View","",dwStyle,
                         initX,initY,
                         view ? view->state.width : 16,
                         view ? view->state.height : 16,
                         (HWND)0,(HMENU)0,mgr.hInstance,(LPVOID)0);
if(!v->window)
  {
  SYS_DROP_DEAD("CreateWindowEx() failure");
  }
PIXELFORMATDESCRIPTOR pfd;
pfd.nSize=sizeof(pfd);
pfd.nVersion=1;
pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
if(mgr.stereoMode==GFX_ACTIVE_STEREO)
  {
  pfd.dwFlags|=PFD_STEREO;
  }
pfd.iPixelType=PFD_TYPE_RGBA;
pfd.cColorBits=32;
pfd.cRedBits=0;
pfd.cRedShift=0;
pfd.cGreenBits=0;
pfd.cGreenShift=0;
pfd.cBlueBits=0;
pfd.cBlueShift=0;
pfd.cAlphaBits=0;
pfd.cAlphaShift=0;
pfd.cAccumBits=0;
pfd.cAccumRedBits=0;
pfd.cAccumGreenBits=0;
pfd.cAccumBlueBits=0;
pfd.cAccumAlphaBits=0;
pfd.cDepthBits=24;
pfd.cStencilBits=0;
pfd.cAuxBuffers=0;
pfd.iLayerType=PFD_MAIN_PLANE;
pfd.bReserved=0;
pfd.dwLayerMask=0;
pfd.dwVisibleMask=0;
pfd.dwDamageMask=0;
v->hdc=GetDC(v->window);
int pixelFormat=ChoosePixelFormat(v->hdc,&pfd);
if(!pixelFormat)
  {
  SYS_DROP_DEAD("ChoosePixelFormat() failure");
  }
if(!SetPixelFormat(v->hdc,pixelFormat,&pfd))
  {
  SYS_DROP_DEAD("SetPixelFormat() failure");
  }
if(view)
  {
  PtrVector_push(&mgr.views,v);
  view->specific=v;
  Win32View_updateTitle(v);
  Win32View_updateCursor(v);
  ShowWindow(v->window,SW_SHOWNORMAL); // callback ignored
  SetForegroundWindow(v->window);
  RECT cr;
  GetClientRect(v->window,&cr);
  view->state.width=cr.right-cr.left;
  view->state.height=cr.bottom-cr.top;
  }
else
  {
  mgr.commonView=v;
  }
}

GlView * // created view
GlView_open(int initX,
            int initY,
            int initWidth,
            int initHeight,
            bool borderless,
            GlViewCallback callback,
            void *data)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
GlView *view=Memory_alloc(1,GlView);
strcpy(view->title,"GlView");
view->cursor=GFX_ARROW_CSR;
GfxState_init(&view->state,initWidth,initHeight);
view->keyRepeat=true;
Win32View_open(view,initX,initY,borderless);
view->callback=callback;
view->data=data;
return view;
}

void
GlView_setTitle(GlView *view,
                 const char *title)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
int maxLen=sizeof(view->title)-1;
strncpy(view->title,title,maxLen);
view->title[maxLen]='\0';
Win32View_updateTitle((Win32View *)view->specific);
}

void
GlView_setCursor(GlView *view,
                  GfxCursor cursor)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
view->cursor=cursor;
Win32View_updateCursor((Win32View *)view->specific);
}

void
GlView_postRedraw(GlView *view)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
Win32View *v=(Win32View *)view->specific;
InvalidateRgn(v->window,(HRGN)0,FALSE);
}

void
GlView_makeCurrentGlTarget(GlView *view)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
Win32View *v=(Win32View *)view->specific;
if(!wglMakeCurrent(v->hdc,mgr.wglContext))
  {
  SYS_DROP_DEAD("wglMakeCurrent() failure");
  }
}

const intptr_t *
GlView_glContext(void)
{
static intptr_t params[2];
params[0]=(intptr_t)mgr.commonView->hdc;
params[1]=(intptr_t)mgr.wglContext;
return params;
}

//----------------------------------------------------------------------------

void
GlView_init(void)
{
if(!mgr.initDone)
  {
  mgr.initDone=true;
  mgr.initCpuCtx=Platform_cpuCtx();
  PtrVector_init(&mgr.views);
  mgr.mouseView=(Win32View *)0;
  mgr.processingEvents=false;
  mgr.eventCount=0;
  mgr.stereoMode=GlView_detectStereoMode();
  mgr.hInstance=(HINSTANCE)GetModuleHandle((LPCTSTR)0);
  if(!mgr.hInstance)
    {
    SYS_DROP_DEAD("GetModuleHandle() failure");
    }
  mgr.cursors[GFX_NO_CSR]=(HCURSOR)0;
  mgr.cursors[GFX_ARROW_CSR]=LoadCursor((HINSTANCE)0,IDC_ARROW);
  mgr.cursors[GFX_TEXT_CSR]=LoadCursor((HINSTANCE)0,IDC_IBEAM);
  mgr.cursors[GFX_HAND_CSR]=LoadCursor((HINSTANCE)0,IDC_HAND);
  mgr.cursors[GFX_CROSS_CSR]=LoadCursor((HINSTANCE)0,IDC_CROSS);
  mgr.cursors[GFX_WAIT_CSR]=LoadCursor((HINSTANCE)0,IDC_WAIT);
  mgr.cursors[GFX_NORTH_CSR]=LoadCursor((HINSTANCE)0,IDC_SIZENS);
  mgr.cursors[GFX_SOUTH_CSR]=LoadCursor((HINSTANCE)0,IDC_SIZENS);
  mgr.cursors[GFX_EAST_CSR]=LoadCursor((HINSTANCE)0,IDC_SIZEWE);
  mgr.cursors[GFX_WEST_CSR]=LoadCursor((HINSTANCE)0,IDC_SIZEWE);
  WNDCLASS wc;
  wc.style=CS_OWNDC;
  wc.lpfnWndProc=Win32View_wndProc;
  wc.cbClsExtra=0;
  wc.cbWndExtra=0;
  wc.hInstance=mgr.hInstance;
  wc.hIcon=LoadIcon((HINSTANCE)0,IDI_APPLICATION);
  wc.hCursor=(HCURSOR)0;
  wc.hbrBackground=(HBRUSH)0;
  wc.lpszMenuName=(LPCTSTR)0;
  wc.lpszClassName="TransGl-View";
  if(!RegisterClass(&wc))
    {
    SYS_DROP_DEAD("RegisterClass() failure");
    }
  Win32View_open((GlView *)0,0,0,false);
  // nb: an initial context is mandatory to load extensions
  HGLRC tmpContext=wglCreateContext(mgr.commonView->hdc);
  if(!tmpContext)
    {
    SYS_DROP_DEAD("wglCreateContext() failure");
    }
  if(!wglMakeCurrent(mgr.commonView->hdc,tmpContext))
    {
    SYS_DROP_DEAD("wglMakeCurrent() failure");
    }
  CoreGl_extensionFunctionWithSuffix(&createContextAttribsFnct,
                                     "wglCreateContextAttribs",
                                     extSuffixes);
  if(createContextAttribsFnct)
    {
    int gl3Attribs[]={WGL_CONTEXT_MAJOR_VERSION_ARB,
                      3,
                      WGL_CONTEXT_MINOR_VERSION_ARB,
                      2,
                      WGL_CONTEXT_PROFILE_MASK_ARB,
                      WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
                      // WGL_CONTEXT_FLAGS_ARB,
                      // WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB,
                      0};
    int *useGl3Attribs=(int *)0;
    int major=0,minor=0;
    if(GlView_detectGlVersion(&major,&minor)&&
       ((major>3)||((major==3)&&(minor>=2))))
      {
      gl3Attribs[1]=major;
      gl3Attribs[3]=minor;
      useGl3Attribs=gl3Attribs;
      }
    mgr.wglContext=createContextAttribsFnct(mgr.commonView->hdc,
                                            (HGLRC)0,useGl3Attribs);
    if(!mgr.wglContext)
      {
      SYS_DROP_DEAD("wglCreateContextAttribs() failure");
      }
    }
  else
    {
    mgr.wglContext=tmpContext;
    tmpContext=(HGLRC)0;
    }
  if(tmpContext)
    {
    wglMakeCurrent((HDC)0,(HGLRC)0);
    wglDeleteContext(tmpContext);
    }
  if(!wglMakeCurrent(mgr.commonView->hdc,mgr.wglContext))
    {
    SYS_DROP_DEAD("wglMakeCurrent() failure");
    }
  CoreGl_init();
  if(mgr.stereoMode==GFX_ACTIVE_STEREO)
    {
    int glIsStereo=0;
    glGetIntegerv(GL_STEREO,&glIsStereo);
    if(!glIsStereo)
      {
      DROP_DEAD("OpenGL context does not support active stereo");
      }
    }
  }
}

//----------------------------------------------------------------------------

const char * // specific OpenGL extensions
_CoreGl_querySpecificExtensions(void)
{
CoreGl_extensionFunctionWithSuffix(&getExtensionsStringFnct,
                                   "wglGetExtensionsString",
                                   extSuffixes);
return getExtensionsStringFnct
       ? getExtensionsStringFnct(mgr.commonView->hdc)
       : (const char *)0;
}

//----------------------------------------------------------------------------
