#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
#   http://www.enib.fr/~harrouet/transprog.html
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import sys ; sys.path.insert(0,'../../lib')
import TransProgConfig

import os
import glob

#-------- Create the construction tool ---------------------------------------

tool=TransProgConfig.Tool()

dynamicCompilerSettings=tool.copySettings()

topDir=tool.settings['TRANSPROG_DIR']
tool.settings['LIBS'].remove('TransProg')
tool.settings['LIBDIRS'].remove(os.path.join(topDir,'lib'))
tool.settings['DEFINES']['WIN32_TRANSPROG_DLL']=None

#-------- Enumerate targets and source files ---------------------------------

containers=[]
for container in ['vector','deque','fifo','rbTree']:
  suffix=container[0].upper()+container[1:]
  template='%s/lib/%sTemplate.py'%(topDir,container)
  for (typeName,CTypeName) in [('r64',  'real64_t'),
                               ('r32',  'real32_t'),
                               ('i',    'int'),
                               ('iPtr', 'intptr_t'),
                               ('ui8',  'uint8_t'),
                               ('str',  'char *'),
                               ('ptr',  'void *')]:
    containerType=typeName[0].upper()+typeName[1:]+suffix
    args=[CTypeName,containerType]
    if typeName=='str' and container in ['vector','rbTree']:
      args.append('#define %s_NO_DEFAULT_CMP 1'%containerType)
    containers.append(
      tool.Template('%s/include/TransProg/%s%s.h'%(topDir,typeName,suffix),
                    template,args))

tool.Template('skeletonHints.c',
              'skeletonHintGenerator.py')

def dynamicCompilerDefaults(cleaning,target,sources,arg,settings):
  if not cleaning:
    code='\n'
    for cmd in ['cc',
                'cflags-only-I','cflags-only-D','cflags-only-other',
                'libs-only-L','libs-only-l','libs-only-other']:
      code+='static\n'
      code+='const char * const\n'
      code+='_DynamicCompiler_%s[]={\n'%(cmd.replace('-','_'))
      cmd=tool.showSettings('--%s'%cmd,settings)[0]
      for i in cmd:
        code+='  "%s",\n'%(i.replace('\\','\\\\').replace('\"','\\\"'))
      code+='  (const char *)0\n  };\n\n'
    code+='static\n'
    code+='const char * const\n'
    code+='_DynamicCompiler_oflags[]={\n'
    if tool.testBoolOption('use_msc'):
      code+='  "/dll",\n'
      code+='  "/out:",\n'
    else:
      if tool.sysName().startswith('Darwin'):
        code+='  "-bundle",\n'
      else:
        code+='  "-shared",\n'
      code+='  "-o",\n'
    code+='  (const char *)0\n  };\n\n'
    open(target,'w').write(code)

tool.Generator('dynamicCompilerDefaults.c',[],
               dynamicCompilerDefaults,None,dynamicCompilerSettings)

tool.Library(topDir+'/lib/TransProg',
             ['platform.c',
              'platform_detect.c',
              'memory.c',
              'synchro.c',
              'math.c',
              'random.c',
              'text.c',
              'path.c',
              'socket.c',
              'plugin.c',
              'dynamicCompiler.c',
              'scheduler.c',
              'spatialUtils.c',
              'surface.c',
              'coordIndex.c',
              'gfxUtils.c',
              'examiner.c',
              'skeleton.c',
              'natnetClient.c'
              ])

tool.run()

#-----------------------------------------------------------------------------
