//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/coordIndex.h"

//----------------------------------------------------------------------------

void
SpatialAffinity_init(SpatialAffinity *sa,
                     real64_t xMin,
                     real64_t xMax,
                     real64_t yMin,
                     real64_t yMax,
                     real64_t zMin,
                     real64_t zMax)
{
if((xMax<=xMin)||(yMax<=yMin)||(zMax<=zMin))
  {
  DROP_DEAD("inconsistent bounds");
  }
sa->xMin=xMin;
sa->yMin=yMin;
sa->zMin=zMin;
sa->xSizeInv=0.999999/(xMax-xMin); // enlarge size to keep (x-xMin)/xSize < 1.0
sa->ySizeInv=0.999999/(yMax-yMin);
sa->zSizeInv=0.999999/(zMax-zMin);
int n=Platform_cpuCtxCount();
int split[3];
for(int i=0;i<3;++i)
  {
  real64_t r=pow(n,1.0/(3-i));
  split[i]=(int)r;
  if((r-split[i])>(split[i]+1-r))
    {
    ++split[i];
    }
  while(n%split[i])
    {
    --split[i];
    }
  n/=split[i];
  }
// FIXME: divide according to sizes, and consider using memory hierarchy
sa->xSplit=split[0];
sa->ySplit=split[1];
sa->zSplit=split[2];
sa->xySplit=split[0]*split[1];
}

//----------------------------------------------------------------------------

void
PlanarAffinity_init(PlanarAffinity *pa,
                    real64_t xMin,
                    real64_t xMax,
                    real64_t yMin,
                    real64_t yMax)
{
if((xMax<=xMin)||(yMax<=yMin))
  {
  DROP_DEAD("inconsistent bounds");
  }
pa->xMin=xMin;
pa->yMin=yMin;
pa->xSizeInv=0.999999/(xMax-xMin); // enlarge size to keep (x-xMin)/xSize < 1.0
pa->ySizeInv=0.999999/(yMax-yMin);
int n=Platform_cpuCtxCount();
int split[2];
for(int i=0;i<2;++i)
  {
  real64_t r=pow(n,1.0/(2-i));
  split[i]=(int)r;
  if((r-split[i])>(split[i]+1-r))
    {
    ++split[i];
    }
  while(n%split[i])
    {
    --split[i];
    }
  n/=split[i];
  }
// FIXME: divide according to sizes, and consider using memory hierarchy
pa->xSplit=split[0];
pa->ySplit=split[1];
}

//----------------------------------------------------------------------------
