//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

// included from platform_detect.c

static
bool // line read
firstLine(const char *path,
          char *out_line,
          int size)
{
out_line[0]='\0';
FILE *f=fopen(path,"r");
if(f&&fgets(out_line,size,f))
  {
  out_line[size-1]='\0';
  fclose(f);
  int begin=0;
  int end=strlen(out_line);
  while(isspace(out_line[begin])) ++begin;
  while((end>begin)&&isspace(out_line[end-1])) --end;
  for(int i=begin;i<end;++i)
    {
    out_line[i-begin]=(char)tolower(out_line[i]);
    }
  out_line[end-begin]='\0';
  return true;
  }
return false;
}

static
void *
CpuGroup_specificInit(void)
{
char path[0x100];
int pathLen=sprintf(path,"/sys/devices/system/node");
DIR *dir;
do
  {
  dir=opendir(path);
  } while(!dir&&(errno==EINTR));
if(dir)
  {
  struct dirent *entry;
  while((entry=readdir(dir)))
    {
    int id;
    if(!strncmp(entry->d_name,"node",4)&&
       (sscanf(entry->d_name+4,"%d",&id)==1)&&
       (id>_Platform_cpuMemInfo.maxNumaId))
      {
      sprintf(path+pathLen,"/%s",entry->d_name);
      struct stat st;
      int r;
      RESTART_SYSCALL(r,stat(path,&st));
      if((r!=-1)&&S_ISDIR(st.st_mode))
        {
        _Platform_cpuMemInfo.maxNumaId=id;
        }
      }
    }
  closedir(dir);
  }
pathLen=sprintf(path,"/sys/devices/system/cpu");
_Platform_cpuMemInfo.maxCpuId=-1;
do
  {
  dir=opendir(path);
  } while(!dir&&(errno==EINTR));
if(dir)
  {
  struct dirent *entry;
  while((entry=readdir(dir)))
    {
    int id;
    if(!strncmp(entry->d_name,"cpu",3)&&
       (sscanf(entry->d_name+3,"%d",&id)==1)&&
       (id>_Platform_cpuMemInfo.maxCpuId))
      {
      sprintf(path+pathLen,"/%s",entry->d_name);
      struct stat st;
      int r;
      RESTART_SYSCALL(r,stat(path,&st));
      if((r!=-1)&&S_ISDIR(st.st_mode))
        {
        _Platform_cpuMemInfo.maxCpuId=id;
        }
      }
    }
  closedir(dir);
  }
return (void *)0;
}

static
void
CpuGroup_specificDestroy(void *specif)
{
(void)specif;
}

static
bool
CpuGroup_cpuOnline(int cpuId,
                   void *specif)
{
(void)specif;
char path[0x100];
sprintf(path,"/sys/devices/system/cpu/cpu%d",cpuId);
struct stat st;
int r;
RESTART_SYSCALL(r,stat(path,&st));
if((r==-1)||!S_ISDIR(st.st_mode))
  {
  return false;
  }
strcat(path,"/online");
char line[0x100];
return !firstLine(path,line,0x100)|| // the main CPU is always online
       !strcmp(line,"1");
}

static
bool
CpuGroup_numaOnline(int numaId,
                    void *specif)
{
(void)specif;
char path[0x100];
sprintf(path,"/sys/devices/system/node/node%d/cpumap",numaId);
char line[0x100];
return firstLine(path,line,0x100);
}

static
void
parseCpuMap(CpuGroup *grp,
            const char *line)
{
int cpuId=0;
for(int i=strlen(line);i--;)
  {
  if(isalnum(line[i]))
    {
    int n=isalpha(line[i]) ? line[i]-'a'+10 : line[i]-'0';
    for(int j=0;j<4;++j)
      {
      if(n&(1<<j))
        {
        CpuGroup_setCpu(grp,cpuId);
        }
      ++cpuId;
      }
    }
  }
}

static
void
CpuGroup_collectNuma(CpuGroup *grp,
                     void *specif)
{
(void)specif;
char path[0x100];
sprintf(path,"/sys/devices/system/node/node%d/cpumap",grp->numaNode);
char line[0x100];
if(firstLine(path,line,0x100))
  {
  parseCpuMap(grp,line);
  }
}

static
void
iterateCache(int cpuId,
             bool
             (*fnct)(char *path,
                     int pathLen,
                     int level,
                     void **data),
             void **data)
{
char path[0x100];
int pathLen=sprintf(path,"/sys/devices/system/cpu/cpu%d/cache",cpuId);
DIR *dir;
do
  {
  dir=opendir(path);
  } while(!dir&&(errno==EINTR));
if(dir)
  {
  struct dirent *entry;
  while((entry=readdir(dir)))
    {
    if(!strncmp(entry->d_name,"index",5))
      {
      sprintf(path+pathLen,"/%s",entry->d_name);
      int pathLen2=pathLen+strlen(entry->d_name)+1;
      strcpy(path+pathLen2,"/type");
      char line[0x100];
      firstLine(path,line,0x100);
      if(!strcmp(line,"data")||!strcmp(line,"unified"))
        {
        strcpy(path+pathLen2,"/level");
        firstLine(path,line,0x100);
        int lvl;
        if((sscanf(line,"%d",&lvl)==1)&&
           !(*fnct)(path,pathLen2,lvl,data))
          {
          break;
          }
        }
      }
    }
  closedir(dir);
  }
}

static
bool
iterateCache_maxLevel(char *path,
                      int pathLen,
                      int level,
                      void **data)
{
(void)path;
(void)pathLen;
int *maxLevel=(int *)data[0];
if(level>*maxLevel)
  {
  *maxLevel=level;
  }
return true;
}

static
int
CpuGroup_maxCacheLevel(int cpuId,
                       void *specif)
{
(void)specif;
int level=-1;
void *data[1];
data[0]=&level;
iterateCache(cpuId,&iterateCache_maxLevel,data);
return level;
}

static
bool
iterateCache_size(char *path,
                  int pathLen,
                  int level,
                  void **data)
{
int *size=(int *)data[0];
if(level!=*((const int *)data[1]))
  {
  return true;
  }
else
  {
  strcpy(path+pathLen,"/size");
  char line[0x100];
  firstLine(path,line,0x100);
  if(sscanf(line,"%d",size)==1)
    {
    switch(line[strlen(line)-1])
      {
      case 'm': *size*=1024*1024; break;
      case 'k': *size*=1024; break;
      }
    }
  return false;
  }
}

static
int
CpuGroup_cacheSize(int cpuId,
                   int level,
                   void *specif)
{
(void)specif;
int size=0;
void *data[2];
data[0]=&size;
data[1]=&level;
iterateCache(cpuId,&iterateCache_size,data);
return size;
}

static
bool
iterateCache_lineSize(char *path,
                      int pathLen,
                      int level,
                      void **data)
{
int *lineSize=(int *)data[0];
if(level!=*((const int *)data[1]))
  {
  return true;
  }
else
  {
  strcpy(path+pathLen,"/coherency_line_size");
  char line[0x100];
  firstLine(path,line,0x100);
  sscanf(line,"%d",lineSize);
  return false;
  }
}

static
int
CpuGroup_cacheLineSize(int cpuId,
                       int level,
                       void *specif)
{
(void)specif;
int lineSize=0;
void *data[2];
data[0]=&lineSize;
data[1]=&level;
iterateCache(cpuId,&iterateCache_lineSize,data);
return lineSize;
}

static
bool
iterateCache_shared(char *path,
                    int pathLen,
                    int level,
                    void **data)
{
CpuGroup *grp=(CpuGroup *)data[0];
if(level!=*((const int *)data[1]))
  {
  return true;
  }
else
  {
  strcpy(path+pathLen,"/shared_cpu_map");
  char line[0x100];
  firstLine(path,line,0x100);
  parseCpuMap(grp,line);
  return false;
  }
}

static
void
CpuGroup_sharedCache(CpuGroup *grp,
                     int cpuId,
                     int level,
                     void *specif)
{
(void)specif;
void *data[2];
data[0]=grp;
data[1]=&level;
iterateCache(cpuId,&iterateCache_shared,data);
}

//----------------------------------------------------------------------------
