//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

// included from platform_detect.c

typedef struct _CpuGroupSpecificData
{
bool relInfoEx;
DWORD relInfoSize;
uint8_t *relInfo;
} CpuGroupSpecificData;

#define SPECIF ((CpuGroupSpecificData *)specif)

static
void
iterateRelInfo(void *specif,
               bool
               (*fnct)(void *specif,
                       const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
                       const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
                       void **data),
               void **data)
{
const uint8_t *ptr=SPECIF->relInfo;
const uint8_t *end=ptr+SPECIF->relInfoSize;
const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx=
  (const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *)0;
const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info=
  (const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *)0;
if(SPECIF->relInfoEx)
  {
  while(ptr<end)
    {
    infoEx=(const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *)ptr;
    if(!(*fnct)(specif,infoEx,info,data))
      {
      break;
      }
    ptr+=infoEx->Size;
    }
  }
else
  {
  while(ptr<end)
    {
    info=(const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *)ptr;
    if(!(*fnct)(specif,infoEx,info,data))
      {
      break;
      }
    ptr+=sizeof(SYSTEM_LOGICAL_PROCESSOR_INFORMATION);
    }
  }
}

static
bool
iterateRelInfo_show(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
(void)data;
if(infoEx)
  {
  switch(infoEx->Relationship)
    {
    case RelationGroup:
      {
      Platform_msg("Group");
      for(WORD i=0,i_end=infoEx->Group.ActiveGroupCount;i<i_end;++i)
        {
        Platform_msg(" mask[%d]=%016"F64"x",
                     i,
                     infoEx->Group.GroupInfo[i].ActiveProcessorMask);
        
        }
      Platform_msg("\n");
      break;
      }
    case RelationNumaNode:
      {
      Platform_msg("NumaNode node=%d group=%d mask=%016"F64"x\n",
                   (int)infoEx->NumaNode.NodeNumber,
                   infoEx->NumaNode.GroupMask.Group,
                   infoEx->NumaNode.GroupMask.Mask);
      break;
      }
    case RelationProcessorPackage:
      {
      Platform_msg("ProcessorPackage");
      for(WORD i=0,i_end=infoEx->Processor.GroupCount;i<i_end;++i)
        {
        Platform_msg(" group[%d]=%d mask[%d]=%016"F64"x",
                     i,
                     infoEx->Processor.GroupMask[i].Group,
                     i,
                     infoEx->Processor.GroupMask[i].Mask);
        }
      Platform_msg("\n");
      break;
      }
    case RelationProcessorCore:
      {
      Platform_msg("ProcessorCore smt=%d",
                   infoEx->Processor.Flags);
      for(WORD i=0,i_end=infoEx->Processor.GroupCount;i<i_end;++i)
        {
        Platform_msg(" group[%d]=%d mask[%d]=%016"F64"x",
                     i,
                     infoEx->Processor.GroupMask[i].Group,
                     i,
                     infoEx->Processor.GroupMask[i].Mask);
        }
      Platform_msg("\n");
      break;
      }
    case RelationCache:
      {
      const char *cacheType;
      switch(infoEx->Cache.Type)
        {
        case CacheUnified:     cacheType="unified";     break;
        case CacheInstruction: cacheType="instruction"; break;
        case CacheData:        cacheType="data";        break;
        case CacheTrace:       cacheType="trace";       break;
        default:               cacheType="???";         break;
        }
      Platform_msg("Cache level=%d line=%d size=%d type=%s "
                   "group=%d mask=%016"F64"x\n",
                   infoEx->Cache.Level,
                   infoEx->Cache.LineSize,
                   (int)infoEx->Cache.CacheSize,
                   cacheType,
                   infoEx->Cache.GroupMask.Group,
                   infoEx->Cache.GroupMask.Mask);
      break;
      }
    default:
      {
      Platform_msg("Relation ??? 0x%x\n",infoEx->Relationship);
      break;
      }
    }
  }
else if(info)
  {
  Platform_msg("Mask=%016"F64"x ",info->ProcessorMask);
  switch(info->Relationship)
    {
    case RelationNumaNode:
      {
      Platform_msg("NumaNode %d\n",(int)info->NumaNode.NodeNumber);
      break;
      }
    case RelationProcessorCore:
      {
      Platform_msg("ProcessorCore flags=0x%x\n",info->ProcessorCore.Flags);
      break;
      }
    case RelationCache:
      {
      const char *cacheType;
      switch(info->Cache.Type)
        {
        case CacheUnified:     cacheType="unified";     break;
        case CacheInstruction: cacheType="instruction"; break;
        case CacheData:        cacheType="data";        break;
        case CacheTrace:       cacheType="trace";       break;
        default:               cacheType="???";         break;
        }
      Platform_msg("Cache level=%d line=%d size=%d type=%s\n",
                   info->Cache.Level,
                   info->Cache.LineSize,
                   (int)info->Cache.Size,
                   cacheType);
      break;
      }
    case RelationProcessorPackage:
      {
      Platform_msg("ProcessorPackage\n");
      break;
      }
    default:
      {
      Platform_msg("Relation ??? 0x%x\n",info->Relationship);
      break;
      }
    }
  }
return true;
}

static
bool
iterateRelInfo_maxNumaId(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
int *maxNumaId=(int *)data[0];
if(infoEx)
  {
  if((infoEx->Relationship==RelationNumaNode)&&
     ((int)infoEx->NumaNode.NodeNumber>*maxNumaId))
    {
    *maxNumaId=infoEx->NumaNode.NodeNumber;
    }
  }
else if(info)
  {
  if((info->Relationship==RelationNumaNode)&&
     ((int)info->NumaNode.NodeNumber>*maxNumaId))
    {
    *maxNumaId=info->NumaNode.NodeNumber;
    }
  }
return true;
}

static
int
maxCpuIdInMask(WORD group,
               KAFFINITY mask)
{
int maxId=-1;
for(int i=0;mask;++i,mask>>=1)
  {
  if(mask&1)
    {
    maxId=(group<<6)+i;
    }
  }
return maxId;
}

static
bool
iterateRelInfo_maxCpuId(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
int *maxCpuId=(int32_t *)data[0];
if(infoEx)
  {
  switch(infoEx->Relationship)
    {
    case RelationGroup:
      {
      for(WORD i=0,i_end=infoEx->Group.ActiveGroupCount;i<i_end;++i)
        {
        int maxId=maxCpuIdInMask(i,
                  infoEx->Group.GroupInfo[i].ActiveProcessorMask);
        if(maxId>*maxCpuId)
          {
          *maxCpuId=maxId;
          }
        }
      break;
      }
    case RelationNumaNode:
      {
      int maxId=maxCpuIdInMask(infoEx->NumaNode.GroupMask.Group,
                               infoEx->NumaNode.GroupMask.Mask);
      if(maxId>*maxCpuId)
        {
        *maxCpuId=maxId;
        }
      break;
      }
    case RelationProcessorPackage:
    case RelationProcessorCore:
      {
      for(WORD i=0,i_end=infoEx->Processor.GroupCount;i<i_end;++i)
        {
        int maxId=maxCpuIdInMask(infoEx->Processor.GroupMask[i].Group,
                                 infoEx->Processor.GroupMask[i].Mask);
        if(maxId>*maxCpuId)
          {
          *maxCpuId=maxId;
          }
        }
      break;
      }
    case RelationCache:
      {
      int maxId=maxCpuIdInMask(infoEx->Cache.GroupMask.Group,
                               infoEx->Cache.GroupMask.Mask);
      if(maxId>*maxCpuId)
        {
        *maxCpuId=maxId;
        }
      break;
      }
    default:
      {
      break;
      }
    }
  }
else if(info)
  {
  int maxId=maxCpuIdInMask(0,info->ProcessorMask);
  if(maxId>*maxCpuId)
    {
    *maxCpuId=maxId;
    }
  }
return true;
}

static
void *
CpuGroup_specificInit(void)
{
void *specif=malloc(sizeof(CpuGroupSpecificData));
SPECIF->relInfoEx=false;
SPECIF->relInfoSize=0;
SPECIF->relInfo=(uint8_t *)0;
do
  {
  if(!_Platform_w32Fncts.GetLogicalProcessorInformationEx)
    {
    break;
    }
  SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *buffer=
    (SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *)0;
  DWORD sz=0;
  if(!(*_Platform_w32Fncts.GetLogicalProcessorInformationEx)
      (RelationAll,buffer,&sz)&&(GetLastError()==ERROR_INSUFFICIENT_BUFFER))
    {
    buffer=(SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *)malloc(sz);
    }
  if(!buffer)
    {
    break;
    }
  if(!(*_Platform_w32Fncts.GetLogicalProcessorInformationEx)
      (RelationAll,buffer,&sz))
    {
    free(buffer);
    break;
    }
  SPECIF->relInfoEx=true;
  SPECIF->relInfoSize=sz;
  SPECIF->relInfo=(uint8_t *)buffer;
  } while(false);
do
  {
  if(SPECIF->relInfo||
     !_Platform_w32Fncts.GetLogicalProcessorInformation)
    {
    break;
    }
  SYSTEM_LOGICAL_PROCESSOR_INFORMATION *buffer=
    (SYSTEM_LOGICAL_PROCESSOR_INFORMATION *)0;
  DWORD sz=0;
  if(!(*_Platform_w32Fncts.GetLogicalProcessorInformation)(buffer,&sz)&&
     (GetLastError()==ERROR_INSUFFICIENT_BUFFER))
    {
    buffer=(SYSTEM_LOGICAL_PROCESSOR_INFORMATION *)malloc(sz);
    }
  if(!buffer)
    {
    break;
    }
  if(!(*_Platform_w32Fncts.GetLogicalProcessorInformation)(buffer,&sz))
    {
    free(buffer);
    break;
    }
  SPECIF->relInfoEx=false;
  SPECIF->relInfoSize=sz;
  SPECIF->relInfo=(uint8_t *)buffer;
  } while(false);
if(SPECIF->relInfo)
  {
  void *data[1];
  if(false)
    {
    data[0]=(void *)0;
    iterateRelInfo(specif,&iterateRelInfo_show,data);
    }
  data[0]=&_Platform_cpuMemInfo.maxNumaId;
  iterateRelInfo(specif,&iterateRelInfo_maxNumaId,data);
  data[0]=&_Platform_cpuMemInfo.maxCpuId;
  iterateRelInfo(specif,&iterateRelInfo_maxCpuId,data);
  }
return specif;
}

static
void
CpuGroup_specificDestroy(void *specif)
{
if(SPECIF->relInfo)
  {
  free(SPECIF->relInfo);
  }
free(specif);
}

static
bool
cpuIdInMask(int cpuId,
            WORD group,
            KAFFINITY mask)
{
return ((cpuId>>6)==group)&&(mask&(1ULL<<(cpuId&0x0000003F)));
}

static
bool
iterateRelInfo_cpuOnline(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
int *cpuId=(int *)data[0];
bool *online=(bool *)data[1];
if(infoEx)
  {
  switch(infoEx->Relationship)
    {
    case RelationGroup:
      {
      for(WORD i=0,i_end=infoEx->Group.ActiveGroupCount;i<i_end;++i)
        {
        if(cpuIdInMask(*cpuId,i,
                       infoEx->Group.GroupInfo[i].ActiveProcessorMask))
          {
          *online=true;
          return false;
          }
        }
      break;
      }
    case RelationNumaNode:
      {
      if(cpuIdInMask(*cpuId,infoEx->NumaNode.GroupMask.Group,
                     infoEx->NumaNode.GroupMask.Mask))
        {
        *online=true;
        return false;
        }
      break;
      }
    case RelationProcessorPackage:
    case RelationProcessorCore:
      {
      for(WORD i=0,i_end=infoEx->Processor.GroupCount;i<i_end;++i)
        {
        if(cpuIdInMask(*cpuId,infoEx->Processor.GroupMask[i].Group,
                       infoEx->Processor.GroupMask[i].Mask))
          {
          *online=true;
          return false;
          }
        }
      break;
      }
    case RelationCache:
      {
      if(cpuIdInMask(*cpuId,infoEx->Cache.GroupMask.Group,
                     infoEx->Cache.GroupMask.Mask))
        {
        *online=true;
        return false;
        }
      break;
      }
    default:
      {
      break;
      }
    }
  }
else if(info)
  {
  if(cpuIdInMask(*cpuId,0,info->ProcessorMask))
    {
    *online=true;
    return false;
    }
  }
return true;
}

static
bool
CpuGroup_cpuOnline(int cpuId,
                   void *specif)
{
bool online=false;
void *data[2];
data[0]=&cpuId;
data[1]=&online;
iterateRelInfo(specif,&iterateRelInfo_cpuOnline,data);
return online;
}

static
bool
iterateRelInfo_numaOnline(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
int *numaId=(int *)data[0];
bool *online=(bool *)data[1];
if(infoEx)
  {
  if((infoEx->Relationship==RelationNumaNode)&&
     ((int)infoEx->NumaNode.NodeNumber==*numaId))
    {
    *online=true;
    return false;
    }
  }
else if(info)
  {
  if((info->Relationship==RelationNumaNode)&&
     ((int)info->NumaNode.NodeNumber==*numaId))
    {
    *online=true;
    return false;
    }
  }
return true;
}

static
bool
CpuGroup_numaOnline(int numaId,
                    void *specif)
{
bool online=false;
void *data[2];
data[0]=&numaId;
data[1]=&online;
iterateRelInfo(specif,&iterateRelInfo_numaOnline,data);
return online;
}

static
void
parseCpuMask(CpuGroup *grp,
             WORD group,
             KAFFINITY mask)
{
for(int i=0;mask;++i,mask>>=1)
  {
  if(mask&1)
    {
    CpuGroup_setCpu(grp,(group<<6)+i);
    }
  }
}

static
bool
iterateRelInfo_collectNuma(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
CpuGroup *grp=(CpuGroup *)data[0];
if(infoEx)
  {
  if((infoEx->Relationship==RelationNumaNode)&&
     ((int)infoEx->NumaNode.NodeNumber==grp->numaNode))
    {
    parseCpuMask(grp,infoEx->NumaNode.GroupMask.Group,
                 infoEx->NumaNode.GroupMask.Mask);
    return false;
    }
  }
else if(info)
  {
  if((info->Relationship==RelationNumaNode)&&
     ((int)info->NumaNode.NodeNumber==grp->numaNode))
    {
    parseCpuMask(grp,0,info->ProcessorMask);
    return false;
    }
  }
return true;
}

static
void
CpuGroup_collectNuma(CpuGroup *grp,
                     void *specif)
{
void *data[1];
data[0]=grp;
iterateRelInfo(specif,&iterateRelInfo_collectNuma,data);
}

static
bool
iterateRelInfo_maxCacheLevel(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
int *cpuId=(int *)data[0];
int *maxLevel=(int *)data[1];
if(infoEx)
  {
  if((infoEx->Relationship==RelationCache)&&
     ((infoEx->Cache.Type==CacheUnified)||(infoEx->Cache.Type==CacheData))&&
     (infoEx->Cache.Level>=*maxLevel)&&
     cpuIdInMask(*cpuId,infoEx->Cache.GroupMask.Group,
                 infoEx->Cache.GroupMask.Mask))
    {
    *maxLevel=infoEx->Cache.Level;
    }
  }
else if(info)
  {
  if((info->Relationship==RelationCache)&&
     ((info->Cache.Type==CacheUnified)||(info->Cache.Type==CacheData))&&
     (info->Cache.Level>*maxLevel)&&
     cpuIdInMask(*cpuId,0,info->ProcessorMask))
    {
    *maxLevel=info->Cache.Level;
    }
  }
return true;
}

static
int
CpuGroup_maxCacheLevel(int cpuId,
                       void *specif)
{
int level=-1;
void *data[2];
data[0]=&cpuId;
data[1]=&level;
iterateRelInfo(specif,&iterateRelInfo_maxCacheLevel,data);
return level;
}

static
bool
iterateRelInfo_cacheSize(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
int *cpuId=(int *)data[0];
int *level=(int *)data[1];
int *size=(int *)data[2];
if(infoEx)
  {
  if((infoEx->Relationship==RelationCache)&&
     ((infoEx->Cache.Type==CacheUnified)||(infoEx->Cache.Type==CacheData))&&
     (infoEx->Cache.Level==*level)&&
     cpuIdInMask(*cpuId,infoEx->Cache.GroupMask.Group,
                 infoEx->Cache.GroupMask.Mask))
    {
    *size=infoEx->Cache.CacheSize;
    return false;
    }
  }
else if(info)
  {
  if((info->Relationship==RelationCache)&&
     ((info->Cache.Type==CacheUnified)||(info->Cache.Type==CacheData))&&
     (info->Cache.Level==*level)&&
     cpuIdInMask(*cpuId,0,info->ProcessorMask))
    {
    *size=info->Cache.Size;
    return false;
    }
  }
return true;
}

static
int
CpuGroup_cacheSize(int cpuId,
                   int level,
                   void *specif)
{
int size=0;
void *data[3];
data[0]=&cpuId;
data[1]=&level;
data[2]=&size;
iterateRelInfo(specif,&iterateRelInfo_cacheSize,data);
return size;
}

static
bool
iterateRelInfo_cacheLineSize(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
int *cpuId=(int *)data[0];
int *level=(int *)data[1];
int *lineSize=(int *)data[2];
if(infoEx)
  {
  if((infoEx->Relationship==RelationCache)&&
     ((infoEx->Cache.Type==CacheUnified)||(infoEx->Cache.Type==CacheData))&&
     (infoEx->Cache.Level==*level)&&
     cpuIdInMask(*cpuId,infoEx->Cache.GroupMask.Group,
                 infoEx->Cache.GroupMask.Mask))
    {
    *lineSize=infoEx->Cache.LineSize;
    return false;
    }
  }
else if(info)
  {
  if((info->Relationship==RelationCache)&&
     ((info->Cache.Type==CacheUnified)||(info->Cache.Type==CacheData))&&
     (info->Cache.Level==*level)&&
     cpuIdInMask(*cpuId,0,info->ProcessorMask))
    {
    *lineSize=info->Cache.LineSize;
    return false;
    }
  }
return true;
}

static
int
CpuGroup_cacheLineSize(int cpuId,
                       int level,
                       void *specif)
{
int lineSize=0;
void *data[3];
data[0]=&cpuId;
data[1]=&level;
data[2]=&lineSize;
iterateRelInfo(specif,&iterateRelInfo_cacheLineSize,data);
return lineSize;
}

static
bool
iterateRelInfo_sharedCache(
               void *specif,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *infoEx,
               const SYSTEM_LOGICAL_PROCESSOR_INFORMATION *info,
               void **data)
{
(void)specif;
CpuGroup *grp=(CpuGroup *)data[0];
int *cpuId=(int *)data[1];
int *level=(int *)data[2];
if(infoEx)
  {
  if((infoEx->Relationship==RelationCache)&&
     ((infoEx->Cache.Type==CacheUnified)||(infoEx->Cache.Type==CacheData))&&
     (infoEx->Cache.Level==*level)&&
     cpuIdInMask(*cpuId,infoEx->Cache.GroupMask.Group,
                 infoEx->Cache.GroupMask.Mask))
    {
    parseCpuMask(grp,infoEx->Cache.GroupMask.Group,
                 infoEx->Cache.GroupMask.Mask);
    return false;
    }
  }
else if(info)
  {
  if((info->Relationship==RelationCache)&&
     ((info->Cache.Type==CacheUnified)||(info->Cache.Type==CacheData))&&
     (info->Cache.Level==*level)&&
     cpuIdInMask(*cpuId,0,info->ProcessorMask))
    {
    parseCpuMask(grp,0,info->ProcessorMask);
    return false;
    }
  }
return true;
}

static
void
CpuGroup_sharedCache(CpuGroup *grp,
                     int cpuId,
                     int level,
                     void *specif)
{
void *data[3];
data[0]=grp;
data[1]=&cpuId;
data[2]=&level;
iterateRelInfo(specif,&iterateRelInfo_sharedCache,data);
}

//----------------------------------------------------------------------------
