//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/spatialUtils.h"

#define SPATIALBASIS_SIN_PRECISION 0.999999999 // 1.0-1.0e-9

TRANSPROG_API_VAR real64_t _SpatialUtils_idMtx[16]={ 1.0 , 0.0 , 0.0 , 0.0 ,
                                                     0.0 , 1.0 , 0.0 , 0.0 ,
                                                     0.0 , 0.0 , 1.0 , 0.0 ,
                                                     0.0 , 0.0 , 0.0 , 1.0 };

//----------------------------------------------------------------------------

bool // success
SpatialBasis_adjustOrthonormality(SpatialBasis *sb)
{
real64_t *mtx=sb->matrix;
real64_t d=mtx[MTX_R00]*mtx[MTX_R00]+
           mtx[MTX_R10]*mtx[MTX_R10]+
           mtx[MTX_R20]*mtx[MTX_R20];
if(d)
  {
  d=1.0/sqrt(d);
  mtx[MTX_R00]*=d;
  mtx[MTX_R10]*=d;
  mtx[MTX_R20]*=d;
  mtx[MTX_R02]=mtx[MTX_R10]*mtx[MTX_R21]-mtx[MTX_R20]*mtx[MTX_R11];
  mtx[MTX_R12]=mtx[MTX_R20]*mtx[MTX_R01]-mtx[MTX_R00]*mtx[MTX_R21];
  mtx[MTX_R22]=mtx[MTX_R00]*mtx[MTX_R11]-mtx[MTX_R10]*mtx[MTX_R01];
  d=mtx[MTX_R02]*mtx[MTX_R02]+
    mtx[MTX_R12]*mtx[MTX_R12]+
    mtx[MTX_R22]*mtx[MTX_R22];
  if(d)
    {
    d=1.0/sqrt(d);
    mtx[MTX_R02]*=d;
    mtx[MTX_R12]*=d;
    mtx[MTX_R22]*=d;
    mtx[MTX_R01]=mtx[MTX_R12]*mtx[MTX_R20]-mtx[MTX_R22]*mtx[MTX_R10],
    mtx[MTX_R11]=mtx[MTX_R22]*mtx[MTX_R00]-mtx[MTX_R02]*mtx[MTX_R20],
    mtx[MTX_R21]=mtx[MTX_R02]*mtx[MTX_R10]-mtx[MTX_R12]*mtx[MTX_R00];
    return true;
    }
  }
return false;
}

void
SpatialBasis_globalToLocalLocation(const SpatialBasis *sb,
                                   SpatialBasis *inout_location)
{
const real64_t *mtx=sb->matrix;
real64_t *mtx2=inout_location->matrix;
real64_t r00=mtx2[MTX_R00], r01=mtx2[MTX_R01], r02=mtx2[MTX_R02];
real64_t r10=mtx2[MTX_R10], r11=mtx2[MTX_R11], r12=mtx2[MTX_R12];
real64_t r20=mtx2[MTX_R20], r21=mtx2[MTX_R21], r22=mtx2[MTX_R22];
real64_t tx=mtx2[MTX_TX]-mtx[MTX_TX];
real64_t ty=mtx2[MTX_TY]-mtx[MTX_TY];
real64_t tz=mtx2[MTX_TZ]-mtx[MTX_TZ];
mtx2[MTX_R00]=mtx[MTX_R00]*r00+mtx[MTX_R10]*r10+mtx[MTX_R20]*r20;
mtx2[MTX_R01]=mtx[MTX_R00]*r01+mtx[MTX_R10]*r11+mtx[MTX_R20]*r21;
mtx2[MTX_R02]=mtx[MTX_R00]*r02+mtx[MTX_R10]*r12+mtx[MTX_R20]*r22;
mtx2[MTX_TX]=mtx[MTX_R00]*tx+mtx[MTX_R10]*ty+mtx[MTX_R20]*tz;
mtx2[MTX_R10]=mtx[MTX_R01]*r00+mtx[MTX_R11]*r10+mtx[MTX_R21]*r20;
mtx2[MTX_R11]=mtx[MTX_R01]*r01+mtx[MTX_R11]*r11+mtx[MTX_R21]*r21;
mtx2[MTX_R12]=mtx[MTX_R01]*r02+mtx[MTX_R11]*r12+mtx[MTX_R21]*r22;
mtx2[MTX_TY]=mtx[MTX_R01]*tx+mtx[MTX_R11]*ty+mtx[MTX_R21]*tz;
mtx2[MTX_R20]=mtx[MTX_R02]*r00+mtx[MTX_R12]*r10+mtx[MTX_R22]*r20;
mtx2[MTX_R21]=mtx[MTX_R02]*r01+mtx[MTX_R12]*r11+mtx[MTX_R22]*r21;
mtx2[MTX_R22]=mtx[MTX_R02]*r02+mtx[MTX_R12]*r12+mtx[MTX_R22]*r22;
mtx2[MTX_TZ]=mtx[MTX_R02]*tx+mtx[MTX_R12]*ty+mtx[MTX_R22]*tz;
}

void
SpatialBasis_localToGlobalLocation(const SpatialBasis *sb,
                                   SpatialBasis *inout_location)
{
const real64_t *mtx=sb->matrix;
real64_t *mtx2=inout_location->matrix;
real64_t r00=mtx2[MTX_R00], r01=mtx2[MTX_R01], r02=mtx2[MTX_R02];
real64_t r10=mtx2[MTX_R10], r11=mtx2[MTX_R11], r12=mtx2[MTX_R12];
real64_t r20=mtx2[MTX_R20], r21=mtx2[MTX_R21], r22=mtx2[MTX_R22];
real64_t tx=mtx2[MTX_TX], ty=mtx2[MTX_TY], tz=mtx2[MTX_TZ];
mtx2[MTX_R00]=mtx[MTX_R00]*r00+mtx[MTX_R01]*r10+mtx[MTX_R02]*r20;
mtx2[MTX_R01]=mtx[MTX_R00]*r01+mtx[MTX_R01]*r11+mtx[MTX_R02]*r21;
mtx2[MTX_R02]=mtx[MTX_R00]*r02+mtx[MTX_R01]*r12+mtx[MTX_R02]*r22;
mtx2[MTX_TX]=mtx[MTX_R00]*tx+mtx[MTX_R01]*ty+mtx[MTX_R02]*tz+mtx[MTX_TX];
mtx2[MTX_R10]=mtx[MTX_R10]*r00+mtx[MTX_R11]*r10+mtx[MTX_R12]*r20;
mtx2[MTX_R11]=mtx[MTX_R10]*r01+mtx[MTX_R11]*r11+mtx[MTX_R12]*r21;
mtx2[MTX_R12]=mtx[MTX_R10]*r02+mtx[MTX_R11]*r12+mtx[MTX_R12]*r22;
mtx2[MTX_TY]=mtx[MTX_R10]*tx+mtx[MTX_R11]*ty+mtx[MTX_R12]*tz+mtx[MTX_TY];
mtx2[MTX_R20]=mtx[MTX_R20]*r00+mtx[MTX_R21]*r10+mtx[MTX_R22]*r20;
mtx2[MTX_R21]=mtx[MTX_R20]*r01+mtx[MTX_R21]*r11+mtx[MTX_R22]*r21;
mtx2[MTX_R22]=mtx[MTX_R20]*r02+mtx[MTX_R21]*r12+mtx[MTX_R22]*r22;
mtx2[MTX_TZ]=mtx[MTX_R20]*tx+mtx[MTX_R21]*ty+mtx[MTX_R22]*tz+mtx[MTX_TZ];
}

void
SpatialBasis_globalToLocalOrientation(const SpatialBasis *sb,
                                      real64_t *inout_roll,
                                      real64_t *inout_pitch,
                                      real64_t *inout_yaw)
{
const real64_t *mtx=sb->matrix;
real64_t s1,c1,s2,c2,s3,c3;
sincos64(*inout_roll,&s1,&c1);
sincos64(*inout_pitch,&s2,&c2);
sincos64(*inout_yaw,&s3,&c3);
real64_t g00=c2*c3, g01=s1*s2*c3-c1*s3, g02=s1*s3+c1*s2*c3; // [global]
real64_t g10=c2*s3, g11=c1*c3+s1*s2*s3, g12=c1*s2*s3-s1*c3;
real64_t g20=-s2,   g21=s1*c2,          g22=c1*c2;
real64_t r00=mtx[MTX_R00]*g00+mtx[MTX_R10]*g10+mtx[MTX_R20]*g20; // [-self]*[g]
// real64_t r01=mtx[MTX_R00]*g01+mtx[MTX_R10]*g11+mtx[MTX_R20]*g21;
// real64_t r02=mtx[MTX_R00]*g02+mtx[MTX_R10]*g12+mtx[MTX_R20]*g22;
real64_t r10=mtx[MTX_R01]*g00+mtx[MTX_R11]*g10+mtx[MTX_R21]*g20;
// real64_t r11=mtx[MTX_R01]*g01+mtx[MTX_R11]*g11+mtx[MTX_R21]*g21;
// real64_t r12=mtx[MTX_R01]*g02+mtx[MTX_R11]*g12+mtx[MTX_R21]*g22;
real64_t r20=mtx[MTX_R02]*g00+mtx[MTX_R12]*g10+mtx[MTX_R22]*g20;
real64_t r21=mtx[MTX_R02]*g01+mtx[MTX_R12]*g11+mtx[MTX_R22]*g21;
real64_t r22=mtx[MTX_R02]*g02+mtx[MTX_R12]*g12+mtx[MTX_R22]*g22;
if(r20<=-SPATIALBASIS_SIN_PRECISION)
  {
  real64_t r01=mtx[MTX_R00]*g01+mtx[MTX_R10]*g11+mtx[MTX_R20]*g21;
  real64_t r11=mtx[MTX_R01]*g01+mtx[MTX_R11]*g11+mtx[MTX_R21]*g21;
  *inout_roll=0.0; // make a choice
  *inout_pitch=M_PI_2;
  *inout_yaw=atan2(-r01,r11);
  }
else if(r20>=SPATIALBASIS_SIN_PRECISION)
  {
  real64_t r01=mtx[MTX_R00]*g01+mtx[MTX_R10]*g11+mtx[MTX_R20]*g21;
  real64_t r11=mtx[MTX_R01]*g01+mtx[MTX_R11]*g11+mtx[MTX_R21]*g21;
  *inout_roll=0.0; // make a choice
  *inout_pitch=-M_PI_2;
  *inout_yaw=atan2(-r01,r11);
  }
else
  {
  *inout_roll=atan2(r21,r22);
  *inout_pitch=asin(-r20);
  *inout_yaw=atan2(r10,r00);
  }
}

void
SpatialBasis_localToGlobalOrientation(const SpatialBasis *sb,
                                      real64_t *inout_roll,
                                      real64_t *inout_pitch,
                                      real64_t *inout_yaw)
{
const real64_t *mtx=sb->matrix;
real64_t s1,c1,s2,c2,s3,c3;
sincos64(*inout_roll,&s1,&c1);
sincos64(*inout_pitch,&s2,&c2);
sincos64(*inout_yaw,&s3,&c3);
real64_t l00=c2*c3, l01=s1*s2*c3-c1*s3, l02=s1*s3+c1*s2*c3; // [local]
real64_t l10=c2*s3, l11=c1*c3+s1*s2*s3, l12=c1*s2*s3-s1*c3;
real64_t l20=-s2,   l21=s1*c2,          l22=c1*c2;
real64_t r00=mtx[MTX_R00]*l00+mtx[MTX_R01]*l10+mtx[MTX_R02]*l20; // [self]*[l]
// real64_t r01=mtx[MTX_R00]*l01+mtx[MTX_R01]*l11+mtx[MTX_R02]*l21;
// real64_t r02=mtx[MTX_R00]*l02+mtx[MTX_R01]*l12+mtx[MTX_R02]*l22;
real64_t r10=mtx[MTX_R10]*l00+mtx[MTX_R11]*l10+mtx[MTX_R12]*l20;
// real64_t r11=mtx[MTX_R10]*l01+mtx[MTX_R11]*l11+mtx[MTX_R12]*l21;
// real64_t r12=mtx[MTX_R10]*l02+mtx[MTX_R11]*l12+mtx[MTX_R12]*l22;
real64_t r20=mtx[MTX_R20]*l00+mtx[MTX_R21]*l10+mtx[MTX_R22]*l20;
real64_t r21=mtx[MTX_R20]*l01+mtx[MTX_R21]*l11+mtx[MTX_R22]*l21;
real64_t r22=mtx[MTX_R20]*l02+mtx[MTX_R21]*l12+mtx[MTX_R22]*l22;
if(r20<=-SPATIALBASIS_SIN_PRECISION)
  {
  real64_t r01=mtx[MTX_R00]*l01+mtx[MTX_R01]*l11+mtx[MTX_R02]*l21;
  real64_t r11=mtx[MTX_R10]*l01+mtx[MTX_R11]*l11+mtx[MTX_R12]*l21;
  *inout_roll=0.0; // make a choice
  *inout_pitch=M_PI_2;
  *inout_yaw=atan2(-r01,r11);
  }
else if(r20>=SPATIALBASIS_SIN_PRECISION)
  {
  real64_t r01=mtx[MTX_R00]*l01+mtx[MTX_R01]*l11+mtx[MTX_R02]*l21;
  real64_t r11=mtx[MTX_R10]*l01+mtx[MTX_R11]*l11+mtx[MTX_R12]*l21;
  *inout_roll=0.0; // make a choice
  *inout_pitch=-M_PI_2;
  *inout_yaw=atan2(-r01,r11);
  }
else
  {
  *inout_roll=atan2(r21,r22);
  *inout_pitch=asin(-r20);
  *inout_yaw=atan2(r10,r00);
  }
}

void
SpatialBasis_move(SpatialBasis *sb,
                  const SpatialBasis *motion)
{
real64_t *mtx=sb->matrix;
const real64_t *mtx2=motion->matrix;
real64_t r00=mtx[MTX_R00], r01=mtx[MTX_R01], r02=mtx[MTX_R02];
real64_t r10=mtx[MTX_R10], r11=mtx[MTX_R11], r12=mtx[MTX_R12];
real64_t r20=mtx[MTX_R20], r21=mtx[MTX_R21], r22=mtx[MTX_R22];
mtx[MTX_R00]=r00*mtx2[MTX_R00]+r01*mtx2[MTX_R10]+r02*mtx2[MTX_R20];
mtx[MTX_R01]=r00*mtx2[MTX_R01]+r01*mtx2[MTX_R11]+r02*mtx2[MTX_R21];
mtx[MTX_R02]=r00*mtx2[MTX_R02]+r01*mtx2[MTX_R12]+r02*mtx2[MTX_R22];
mtx[MTX_TX]+=r00*mtx2[MTX_TX]+r01*mtx2[MTX_TY]+r02*mtx2[MTX_TZ];
mtx[MTX_R10]=r10*mtx2[MTX_R00]+r11*mtx2[MTX_R10]+r12*mtx2[MTX_R20];
mtx[MTX_R11]=r10*mtx2[MTX_R01]+r11*mtx2[MTX_R11]+r12*mtx2[MTX_R21];
mtx[MTX_R12]=r10*mtx2[MTX_R02]+r11*mtx2[MTX_R12]+r12*mtx2[MTX_R22];
mtx[MTX_TY]+=r10*mtx2[MTX_TX]+r11*mtx2[MTX_TY]+r12*mtx2[MTX_TZ];
mtx[MTX_R20]=r20*mtx2[MTX_R00]+r21*mtx2[MTX_R10]+r22*mtx2[MTX_R20];
mtx[MTX_R21]=r20*mtx2[MTX_R01]+r21*mtx2[MTX_R11]+r22*mtx2[MTX_R21];
mtx[MTX_R22]=r20*mtx2[MTX_R02]+r21*mtx2[MTX_R12]+r22*mtx2[MTX_R22];
mtx[MTX_TZ]+=r20*mtx2[MTX_TX]+r21*mtx2[MTX_TY]+r22*mtx2[MTX_TZ];
}

void
SpatialBasis_invert(SpatialBasis *sb)
{
real64_t *mtx=sb->matrix;
real64_t tx=-mtx[MTX_TX], ty=-mtx[MTX_TY], tz=-mtx[MTX_TZ];
mtx[MTX_TX]=mtx[MTX_R00]*tx+mtx[MTX_R10]*ty+mtx[MTX_R20]*tz;
mtx[MTX_TY]=mtx[MTX_R01]*tx+mtx[MTX_R11]*ty+mtx[MTX_R21]*tz;
mtx[MTX_TZ]=mtx[MTX_R02]*tx+mtx[MTX_R12]*ty+mtx[MTX_R22]*tz;
tx=mtx[MTX_R01]; mtx[MTX_R01]=mtx[MTX_R10]; mtx[MTX_R10]=tx;
tx=mtx[MTX_R02]; mtx[MTX_R02]=mtx[MTX_R20]; mtx[MTX_R20]=tx;
tx=mtx[MTX_R12]; mtx[MTX_R12]=mtx[MTX_R21]; mtx[MTX_R21]=tx;
}

void
SpatialBasis_extractZyxOrientation(const SpatialBasis *sb,
                                   real64_t *out_roll,
                                   real64_t *out_pitch,
                                   real64_t *out_yaw)
{
//      _                                _
//     |  c2c3  s1s2c3-c1s3  s1s3+c1s2c3  |  s1,c1 : roll
//     |  c2s3  c1c3+s1s2s3  c1s2s3-s1c3  |  s2,c2 : pitch
//     |_  -s2         s1c2         c1c2 _|  s3,c3 : yaw
//
const real64_t *mtx=sb->matrix;
if(mtx[MTX_R20]<=-SPATIALBASIS_SIN_PRECISION)
  {
  *out_roll=0.0; // make a choice
  *out_pitch=M_PI_2;
  *out_yaw=atan2(-mtx[MTX_R01],mtx[MTX_R11]);
  }
else if(mtx[MTX_R20]>=SPATIALBASIS_SIN_PRECISION)
  {
  *out_roll=0.0; // make a choice
  *out_pitch=-M_PI_2;
  *out_yaw=atan2(-mtx[MTX_R01],mtx[MTX_R11]);
  }
else
  {
  *out_roll=atan2(mtx[MTX_R21],mtx[MTX_R22]);
  *out_pitch=asin(-mtx[MTX_R20]);
  *out_yaw=atan2(mtx[MTX_R10],mtx[MTX_R00]);
  }
}

void
SpatialBasis_extractXyzOrientation(const SpatialBasis *sb,
                                   real64_t *out_roll,
                                   real64_t *out_pitch,
                                   real64_t *out_yaw)
{
//      _                                 _
//     |         c2c3        -c2s3     s2  |  s1,c1 : roll
//     |  s1s2c3+c1s3  c1c3-s1s2s3  -s1c2  |  s2,c2 : pitch
//     |_ s1s3-c1s2c3  c1s2s3+s1c3   c1c2 _|  s3,c3 : yaw
//
const real64_t *mtx=sb->matrix;
if(mtx[MTX_R02]<=-SPATIALBASIS_SIN_PRECISION)
  {
  *out_roll=0.0; // make a choice
  *out_pitch=-M_PI_2;
  *out_yaw=atan2(mtx[MTX_R10],mtx[MTX_R11]);
  }
else if(mtx[MTX_R02]>=SPATIALBASIS_SIN_PRECISION)
  {
  *out_roll=0.0; // make a choice
  *out_pitch=M_PI_2;
  *out_yaw=atan2(mtx[MTX_R10],mtx[MTX_R11]);
  }
else
  {
  *out_roll=atan2(-mtx[MTX_R12],mtx[MTX_R22]);
  *out_pitch=asin(mtx[MTX_R02]);
  *out_yaw=atan2(-mtx[MTX_R01],mtx[MTX_R00]);
  }
}

void
SpatialBasis_extractRotation(const SpatialBasis *sb,
                             real64_t *out_x,
                             real64_t *out_y,
                             real64_t *out_z,
                             real64_t *out_angle)
{
const real64_t *mtx=sb->matrix;
real64_t m[3][3];
m[0][0]=mtx[MTX_R00]-1.0; m[0][1]=mtx[MTX_R01];     m[0][2]=mtx[MTX_R02];
m[1][0]=mtx[MTX_R10];     m[1][1]=mtx[MTX_R11]-1.0; m[1][2]=mtx[MTX_R12];
m[2][0]=mtx[MTX_R20];     m[2][1]=mtx[MTX_R21];     m[2][2]=mtx[MTX_R22]-1.0;
real64_t maxAbs=-1.0;
int maxLine1=0;
int maxCol1=0;
for(int i=0;i<3;++i)
  {
  for(int j=0;j<3;++j)
    {
    real64_t abs=fabs(m[i][j]);
    if(abs>maxAbs)
      {
      maxAbs=abs;
      maxLine1=i;
      maxCol1=j;
      }
    }
  }
if(!maxAbs)
  {
  *out_x=*out_y=*out_z=*out_angle=0.0; // identity
  return;
  }
real64_t max=m[maxLine1][maxCol1];
for(int j=0;j<3;++j)
  {
  m[maxLine1][j]/=max;
  }
for(int i=0;i<3;++i)
  {
  if(i!=maxLine1)
    {
    real64_t coef=m[i][maxCol1];
    for(int j=0;j<3;++j)
      {
      if(j!=maxCol1)
        {
        m[i][j]-=coef*m[maxLine1][j];
        }
      else
        {
        m[i][j]=0.0;
        }
      }
    }
  }
maxAbs=-1.0;
int maxLine2=0;
int maxCol2=0;
for(int i=0;i<3;++i)
  {
  if(i!=maxLine1)
    {
    for(int j=0;j<3;++j)
      {
      if(j!=maxCol1)
        {
        real64_t abs=fabs(m[i][j]);
        if(abs>maxAbs)
          {
          maxAbs=abs;
          maxLine2=i;
          maxCol2=j;
          }
        }
      }
    }
  }
real64_t x=0.0,y=0.0,z=0.0;
switch(maxCol1)
  {
  case 0:
    if(maxCol2==1)
      {
      z=1.0;
      y=-m[maxLine2][2]/m[maxLine2][maxCol2];
      x=-(m[maxLine1][2]+y*m[maxLine1][maxCol2]);
      }
    else // 2
      {
      y=1.0;
      z=-m[maxLine2][1]/m[maxLine2][maxCol2];
      x=-(m[maxLine1][1]+z*m[maxLine1][maxCol2]);
      }
    break;
  case 1:
    if(maxCol2==0)
      {
      z=1.0;
      x=-m[maxLine2][2]/m[maxLine2][maxCol2];
      y=-(m[maxLine1][2]+x*m[maxLine1][maxCol2]);
      }
    else // 2
      {
      x=1.0;
      z=-m[maxLine2][0]/m[maxLine2][maxCol2];
      y=-(m[maxLine1][0]+z*m[maxLine1][maxCol2]);
      }
    break;
  case 2:
    if(maxCol2==0)
      {
      y=1.0;
      x=-m[maxLine2][1]/m[maxLine2][maxCol2];
      z=-(m[maxLine1][1]+x*m[maxLine1][maxCol2]);
      }
    else // 1
      {
      x=1.0;
      y=-m[maxLine2][0]/m[maxLine2][maxCol2];
      z=-(m[maxLine1][0]+y*m[maxLine1][maxCol2]);
      }
    break;
  }
real64_t x2=x*x;
real64_t y2=y*y;
real64_t z2=z*z;
real64_t r=sqrt(x2+y2+z2);
real64_t r2=r*r;
x/=r;
y/=r;
z/=r;
real64_t c,s;
if((x2<=y2)&&(x2<=z2))
  {
  c=(mtx[MTX_R00]*r2-x2)/(r2-x2);
  if(y2>=z2)
    {
    s=((1.0-c)*x*z-mtx[MTX_R20])/y;
    }
  else
    {
    s=((1.0-c)*x*y-mtx[MTX_R01])/z;
    }
  }
else if((y2<=x2)&&(y2<=z2))
  {
  c=(mtx[MTX_R11]*r2-y2)/(r2-y2);
  if(x2>=z2)
    {
    s=((1.0-c)*y*z-mtx[MTX_R12])/x;
    }
  else
    {
    s=((1.0-c)*x*y-mtx[MTX_R01])/z;
    }
  }
else // (z2<=x2)&&(z2<=y2)
  {
  c=(mtx[MTX_R22]*r2-z2)/(r2-z2);
  if(x2>=y2)
    {
    s=((1.0-c)*y*z-mtx[MTX_R12])/x;
    }
  else
    {
    s=((1.0-c)*x*z-mtx[MTX_R20])/y;
    }
  }
*out_x=x;
*out_y=y;
*out_z=z;
*out_angle=atan2(s,c);
}

void
SpatialBasis_setZyxOrientation(SpatialBasis *sb,
                               real64_t roll,
                               real64_t pitch,
                               real64_t yaw)
{
real64_t *mtx=sb->matrix;
real64_t s1,c1,s2,c2,s3,c3;
sincos64(roll,&s1,&c1);
sincos64(pitch,&s2,&c2);
sincos64(yaw,&s3,&c3);
mtx[MTX_R00]=c2*c3;  mtx[MTX_R01]=s1*s2*c3-c1*s3;  mtx[MTX_R02]=s1*s3+c1*s2*c3;
mtx[MTX_R10]=c2*s3;  mtx[MTX_R11]=c1*c3+s1*s2*s3;  mtx[MTX_R12]=c1*s2*s3-s1*c3;
mtx[MTX_R20]=-s2;    mtx[MTX_R21]=s1*c2;           mtx[MTX_R22]=c1*c2;
}

void
SpatialBasis_setXyzOrientation(SpatialBasis *sb,
                               real64_t roll,
                               real64_t pitch,
                               real64_t yaw)
{
real64_t *mtx=sb->matrix;
real64_t s1,c1,s2,c2,s3,c3;
sincos64(roll,&s1,&c1);
sincos64(pitch,&s2,&c2);
sincos64(yaw,&s3,&c3);
mtx[MTX_R00]=c2*c3;           mtx[MTX_R01]=-c2*s3;          mtx[MTX_R02]=s2;
mtx[MTX_R10]=s1*s2*c3+c1*s3;  mtx[MTX_R11]=c1*c3-s1*s2*s3;  mtx[MTX_R12]=-s1*c2;
mtx[MTX_R20]=s1*s3-c1*s2*c3;  mtx[MTX_R21]=c1*s2*s3+s1*c3;  mtx[MTX_R22]=c1*c2;
}

void
SpatialBasis_setRotation(SpatialBasis *sb,
                         real64_t x,
                         real64_t y,
                         real64_t z,
                         real64_t angle)
{
real64_t *mtx=sb->matrix;
real64_t r=sqrt(x*x+y*y+z*z);
if(r)
  {
  x/=r;
  y/=r;
  z/=r;
  real64_t s,c;
  sincos64(angle,&s,&c);
  real64_t t=1.0-c;
  mtx[MTX_R00]=t*x*x+c;    mtx[MTX_R01]=t*x*y-s*z;  mtx[MTX_R02]=t*x*z+s*y;
  mtx[MTX_R10]=t*x*y+s*z;  mtx[MTX_R11]=t*y*y+c;    mtx[MTX_R12]=t*y*z-s*x;
  mtx[MTX_R20]=t*x*z-s*y;  mtx[MTX_R21]=t*y*z+s*x;  mtx[MTX_R22]=t*z*z+c;
  }
}

void
SpatialBasis_rotate(SpatialBasis *sb,
                    real64_t x,
                    real64_t y,
                    real64_t z,
                    real64_t dAngle)
{
real64_t *mtx=sb->matrix;
real64_t r=sqrt(x*x+y*y+z*z);
if(r)
  {
  x/=r;
  y/=r;
  z/=r;
  real64_t s,c;         //  tx2+c   txy-sz  txz+sy
  sincos64(dAngle,&s,&c); //  txy+sz  ty2+c   tyz-sx
  real64_t t=1.0-c;     //  txz-sy  tyz+sx  tz2+c
  real64_t l00=t*x*x+c,   l01=t*x*y-s*z, l02=t*x*z+s*y;
  real64_t l10=t*x*y+s*z, l11=t*y*y+c,   l12=t*y*z-s*x;
  real64_t l20=t*x*z-s*y, l21=t*y*z+s*x, l22=t*z*z+c;
  real64_t g00=mtx[MTX_R00], g01=mtx[MTX_R01], g02=mtx[MTX_R02];
  real64_t g10=mtx[MTX_R10], g11=mtx[MTX_R11], g12=mtx[MTX_R12];
  real64_t g20=mtx[MTX_R20], g21=mtx[MTX_R21], g22=mtx[MTX_R22];
  mtx[MTX_R00]=g00*l00+g01*l10+g02*l20;  // [Global]*[Local]
  mtx[MTX_R01]=g00*l01+g01*l11+g02*l21;
  mtx[MTX_R02]=g00*l02+g01*l12+g02*l22;
  mtx[MTX_R10]=g10*l00+g11*l10+g12*l20;
  mtx[MTX_R11]=g10*l01+g11*l11+g12*l21;
  mtx[MTX_R12]=g10*l02+g11*l12+g12*l22;
  mtx[MTX_R20]=g20*l00+g21*l10+g22*l20;
  mtx[MTX_R21]=g20*l01+g21*l11+g22*l21;
  mtx[MTX_R22]=g20*l02+g21*l12+g22*l22;
  }
}

void
SpatialBasis_roll(SpatialBasis *sb, //  1  0  0
                  real64_t dRoll)   //  0  c -s
{                                   //  0  s  c
real64_t *mtx=sb->matrix;
real64_t s,c;
sincos64(dRoll,&s,&c);
real64_t r01=+c*mtx[MTX_R01]+s*mtx[MTX_R02];  // [Global]*[Local]
real64_t r02=-s*mtx[MTX_R01]+c*mtx[MTX_R02];
real64_t r11=+c*mtx[MTX_R11]+s*mtx[MTX_R12];
real64_t r12=-s*mtx[MTX_R11]+c*mtx[MTX_R12];
real64_t r21=+c*mtx[MTX_R21]+s*mtx[MTX_R22];
real64_t r22=-s*mtx[MTX_R21]+c*mtx[MTX_R22];
mtx[MTX_R01]=r01;
mtx[MTX_R02]=r02;
mtx[MTX_R11]=r11;
mtx[MTX_R12]=r12;
mtx[MTX_R21]=r21;
mtx[MTX_R22]=r22;
}

void
SpatialBasis_pitch(SpatialBasis *sb, //  c  0  s
                   real64_t dPitch)  //  0  1  0
{                                    // -s  0  c
real64_t *mtx=sb->matrix;
real64_t s,c;
sincos64(dPitch,&s,&c);
real64_t r00=+c*mtx[MTX_R00]-s*mtx[MTX_R02];  // [Global]*[Local]
real64_t r02=+s*mtx[MTX_R00]+c*mtx[MTX_R02];
real64_t r10=+c*mtx[MTX_R10]-s*mtx[MTX_R12];
real64_t r12=+s*mtx[MTX_R10]+c*mtx[MTX_R12];
real64_t r20=+c*mtx[MTX_R20]-s*mtx[MTX_R22];
real64_t r22=+s*mtx[MTX_R20]+c*mtx[MTX_R22];
mtx[MTX_R00]=r00;
mtx[MTX_R02]=r02;
mtx[MTX_R10]=r10;
mtx[MTX_R12]=r12;
mtx[MTX_R20]=r20;
mtx[MTX_R22]=r22;
}

void
SpatialBasis_yaw(SpatialBasis *sb, //  c -s  0
                 real64_t dYaw)    //  s  c  0
{                                  //  0  0  1
real64_t *mtx=sb->matrix;
real64_t s,c;
sincos64(dYaw,&s,&c);
real64_t r00=+c*mtx[MTX_R00]+s*mtx[MTX_R01];  // [Global]*[Local]
real64_t r01=-s*mtx[MTX_R00]+c*mtx[MTX_R01];
real64_t r10=+c*mtx[MTX_R10]+s*mtx[MTX_R11];
real64_t r11=-s*mtx[MTX_R10]+c*mtx[MTX_R11];
real64_t r20=+c*mtx[MTX_R20]+s*mtx[MTX_R21];
real64_t r21=-s*mtx[MTX_R20]+c*mtx[MTX_R21];
mtx[MTX_R00]=r00;
mtx[MTX_R01]=r01;
mtx[MTX_R10]=r10;
mtx[MTX_R11]=r11;
mtx[MTX_R20]=r20;
mtx[MTX_R21]=r21;
}

//----------------------------------------------------------------------------

void
SpatialTransform_invert(SpatialTransform *st)
{
if(!st->identity)
  {
  real64_t *mtx=st->matrix;
  real64_t d=1.0/(-mtx[MTX_R00]*mtx[MTX_R11]*mtx[MTX_R22]
                  +mtx[MTX_R00]*mtx[MTX_R12]*mtx[MTX_R21]
                  +mtx[MTX_R01]*mtx[MTX_R10]*mtx[MTX_R22]
                  -mtx[MTX_R01]*mtx[MTX_R12]*mtx[MTX_R20]
                  -mtx[MTX_R02]*mtx[MTX_R10]*mtx[MTX_R21]
                  +mtx[MTX_R02]*mtx[MTX_R11]*mtx[MTX_R20]);
  real64_t n00=mtx[MTX_R12]*mtx[MTX_R21]-mtx[MTX_R11]*mtx[MTX_R22];
  real64_t n01=mtx[MTX_R01]*mtx[MTX_R22]-mtx[MTX_R02]*mtx[MTX_R21];
  real64_t n02=mtx[MTX_R02]*mtx[MTX_R11]-mtx[MTX_R01]*mtx[MTX_R12];
  real64_t n10=mtx[MTX_R10]*mtx[MTX_R22]-mtx[MTX_R12]*mtx[MTX_R20];
  real64_t n11=mtx[MTX_R02]*mtx[MTX_R20]-mtx[MTX_R00]*mtx[MTX_R22];
  real64_t n12=mtx[MTX_R00]*mtx[MTX_R12]-mtx[MTX_R02]*mtx[MTX_R10];
  real64_t n20=mtx[MTX_R11]*mtx[MTX_R20]-mtx[MTX_R10]*mtx[MTX_R21];
  real64_t n21=mtx[MTX_R00]*mtx[MTX_R21]-mtx[MTX_R01]*mtx[MTX_R20];
  real64_t n22=mtx[MTX_R01]*mtx[MTX_R10]-mtx[MTX_R00]*mtx[MTX_R11];
  real64_t ntx=
           (mtx[MTX_R01]*mtx[MTX_R12]-mtx[MTX_R02]*mtx[MTX_R11])*mtx[MTX_TZ]+
           (mtx[MTX_R02]*mtx[MTX_R21]-mtx[MTX_R01]*mtx[MTX_R22])*mtx[MTX_TY]+
           (mtx[MTX_R11]*mtx[MTX_R22]-mtx[MTX_R12]*mtx[MTX_R21])*mtx[MTX_TX];
  real64_t nty=
           (mtx[MTX_R02]*mtx[MTX_R10]-mtx[MTX_R00]*mtx[MTX_R12])*mtx[MTX_TZ]+
           (mtx[MTX_R00]*mtx[MTX_R22]-mtx[MTX_R02]*mtx[MTX_R20])*mtx[MTX_TY]+
           (mtx[MTX_R12]*mtx[MTX_R20]-mtx[MTX_R10]*mtx[MTX_R22])*mtx[MTX_TX];
  real64_t ntz=
           (mtx[MTX_R00]*mtx[MTX_R11]-mtx[MTX_R01]*mtx[MTX_R10])*mtx[MTX_TZ]+
           (mtx[MTX_R01]*mtx[MTX_R20]-mtx[MTX_R00]*mtx[MTX_R21])*mtx[MTX_TY]+
           (mtx[MTX_R10]*mtx[MTX_R21]-mtx[MTX_R11]*mtx[MTX_R20])*mtx[MTX_TX];
  mtx[MTX_R00]=n00*d; mtx[MTX_R01]=n01*d; mtx[MTX_R02]=n02*d;
  mtx[MTX_R10]=n10*d; mtx[MTX_R11]=n11*d; mtx[MTX_R12]=n12*d;
  mtx[MTX_R20]=n20*d; mtx[MTX_R21]=n21*d; mtx[MTX_R22]=n22*d;
  mtx[MTX_TX]=ntx*d; mtx[MTX_TY]=nty*d; mtx[MTX_TZ]=ntz*d;
  }
}

void
SpatialTransform_preTransform(SpatialTransform *st,
                              const SpatialTransform *other)
{
if(!other->identity)
  {
  if(st->identity)
    {
    *st=*other;
    }
  else
    {
    real64_t *mtx=st->matrix;
    const real64_t *mtx2=other->matrix;                    // [Other]*[Global]
    real64_t r00=mtx[MTX_R00], r01=mtx[MTX_R01], r02=mtx[MTX_R02];
    real64_t r10=mtx[MTX_R10], r11=mtx[MTX_R11], r12=mtx[MTX_R12];
    real64_t r20=mtx[MTX_R20], r21=mtx[MTX_R21], r22=mtx[MTX_R22];
    real64_t  tx=mtx[MTX_TX],   ty=mtx[MTX_TY],   tz=mtx[MTX_TZ];
    mtx[MTX_R00]=mtx2[MTX_R00]*r00+mtx2[MTX_R01]*r10+mtx2[MTX_R02]*r20;
    mtx[MTX_R01]=mtx2[MTX_R00]*r01+mtx2[MTX_R01]*r11+mtx2[MTX_R02]*r21;
    mtx[MTX_R02]=mtx2[MTX_R00]*r02+mtx2[MTX_R01]*r12+mtx2[MTX_R02]*r22;
    mtx[MTX_TX]=mtx2[MTX_R00]*tx+mtx2[MTX_R01]*ty+mtx2[MTX_R02]*tz+
                mtx2[MTX_TX];
    mtx[MTX_R10]=mtx2[MTX_R10]*r00+mtx2[MTX_R11]*r10+mtx2[MTX_R12]*r20;
    mtx[MTX_R11]=mtx2[MTX_R10]*r01+mtx2[MTX_R11]*r11+mtx2[MTX_R12]*r21;
    mtx[MTX_R12]=mtx2[MTX_R10]*r02+mtx2[MTX_R11]*r12+mtx2[MTX_R12]*r22;
    mtx[MTX_TY]=mtx2[MTX_R10]*tx+mtx2[MTX_R11]*ty+mtx2[MTX_R12]*tz+
                mtx2[MTX_TY];
    mtx[MTX_R20]=mtx2[MTX_R20]*r00+mtx2[MTX_R21]*r10+mtx2[MTX_R22]*r20;
    mtx[MTX_R21]=mtx2[MTX_R20]*r01+mtx2[MTX_R21]*r11+mtx2[MTX_R22]*r21;
    mtx[MTX_R22]=mtx2[MTX_R20]*r02+mtx2[MTX_R21]*r12+mtx2[MTX_R22]*r22;
    mtx[MTX_TZ]=mtx2[MTX_R20]*tx+mtx2[MTX_R21]*ty+mtx2[MTX_R22]*tz+
                mtx2[MTX_TZ];
    }
  }
}

void
SpatialTransform_postTransform(SpatialTransform *st,
                               const SpatialTransform *other)
{
if(!other->identity)
  {
  if(st->identity)
    {
    *st=*other;
    }
  else
    {
    real64_t *mtx=st->matrix;
    const real64_t *mtx2=other->matrix;                    // [Global]*[Other]
    real64_t r00=mtx[MTX_R00], r01=mtx[MTX_R01], r02=mtx[MTX_R02];
    real64_t r10=mtx[MTX_R10], r11=mtx[MTX_R11], r12=mtx[MTX_R12];
    real64_t r20=mtx[MTX_R20], r21=mtx[MTX_R21], r22=mtx[MTX_R22];
    mtx[MTX_R00]=r00*mtx2[MTX_R00]+r01*mtx2[MTX_R10]+r02*mtx2[MTX_R20];
    mtx[MTX_R01]=r00*mtx2[MTX_R01]+r01*mtx2[MTX_R11]+r02*mtx2[MTX_R21];
    mtx[MTX_R02]=r00*mtx2[MTX_R02]+r01*mtx2[MTX_R12]+r02*mtx2[MTX_R22];
    mtx[MTX_TX]+=r00*mtx2[MTX_TX]+r01*mtx2[MTX_TY]+r02*mtx2[MTX_TZ];
    mtx[MTX_R10]=r10*mtx2[MTX_R00]+r11*mtx2[MTX_R10]+r12*mtx2[MTX_R20];
    mtx[MTX_R11]=r10*mtx2[MTX_R01]+r11*mtx2[MTX_R11]+r12*mtx2[MTX_R21];
    mtx[MTX_R12]=r10*mtx2[MTX_R02]+r11*mtx2[MTX_R12]+r12*mtx2[MTX_R22];
    mtx[MTX_TY]+=r10*mtx2[MTX_TX]+r11*mtx2[MTX_TY]+r12*mtx2[MTX_TZ];
    mtx[MTX_R20]=r20*mtx2[MTX_R00]+r21*mtx2[MTX_R10]+r22*mtx2[MTX_R20];
    mtx[MTX_R21]=r20*mtx2[MTX_R01]+r21*mtx2[MTX_R11]+r22*mtx2[MTX_R21];
    mtx[MTX_R22]=r20*mtx2[MTX_R02]+r21*mtx2[MTX_R12]+r22*mtx2[MTX_R22];
    mtx[MTX_TZ]+=r20*mtx2[MTX_TX]+r21*mtx2[MTX_TY]+r22*mtx2[MTX_TZ];
    }
  }
}

void
SpatialTransform_preMove(SpatialTransform *st,
                         const SpatialBasis *motion)
{
if(st->identity)
  {
  st->identity=false;
  memcpy(st->matrix,motion->matrix,16*sizeof(real64_t));
  }
else
  {
  real64_t *mtx=st->matrix;
  const real64_t *mtx2=motion->matrix;                    // [Motion]*[Global]
  real64_t r00=mtx[MTX_R00], r01=mtx[MTX_R01], r02=mtx[MTX_R02];
  real64_t r10=mtx[MTX_R10], r11=mtx[MTX_R11], r12=mtx[MTX_R12];
  real64_t r20=mtx[MTX_R20], r21=mtx[MTX_R21], r22=mtx[MTX_R22];
  real64_t  tx=mtx[MTX_TX],   ty=mtx[MTX_TY],   tz=mtx[MTX_TZ];
  mtx[MTX_R00]=mtx2[MTX_R00]*r00+mtx2[MTX_R01]*r10+mtx2[MTX_R02]*r20;
  mtx[MTX_R01]=mtx2[MTX_R00]*r01+mtx2[MTX_R01]*r11+mtx2[MTX_R02]*r21;
  mtx[MTX_R02]=mtx2[MTX_R00]*r02+mtx2[MTX_R01]*r12+mtx2[MTX_R02]*r22;
  mtx[MTX_TX]=mtx2[MTX_R00]*tx+mtx2[MTX_R01]*ty+mtx2[MTX_R02]*tz+
              mtx2[MTX_TX];
  mtx[MTX_R10]=mtx2[MTX_R10]*r00+mtx2[MTX_R11]*r10+mtx2[MTX_R12]*r20;
  mtx[MTX_R11]=mtx2[MTX_R10]*r01+mtx2[MTX_R11]*r11+mtx2[MTX_R12]*r21;
  mtx[MTX_R12]=mtx2[MTX_R10]*r02+mtx2[MTX_R11]*r12+mtx2[MTX_R12]*r22;
  mtx[MTX_TY]=mtx2[MTX_R10]*tx+mtx2[MTX_R11]*ty+mtx2[MTX_R12]*tz+
              mtx2[MTX_TY];
  mtx[MTX_R20]=mtx2[MTX_R20]*r00+mtx2[MTX_R21]*r10+mtx2[MTX_R22]*r20;
  mtx[MTX_R21]=mtx2[MTX_R20]*r01+mtx2[MTX_R21]*r11+mtx2[MTX_R22]*r21;
  mtx[MTX_R22]=mtx2[MTX_R20]*r02+mtx2[MTX_R21]*r12+mtx2[MTX_R22]*r22;
  mtx[MTX_TZ]=mtx2[MTX_R20]*tx+mtx2[MTX_R21]*ty+mtx2[MTX_R22]*tz+
              mtx2[MTX_TZ];
  }
}

void
SpatialTransform_postMove(SpatialTransform *st,
                          const SpatialBasis *motion)
{
if(st->identity)
  {
  st->identity=false;
  memcpy(st->matrix,motion->matrix,16*sizeof(real64_t));
  }
else
  {
  real64_t *mtx=st->matrix;
  const real64_t *mtx2=motion->matrix;                    // [Motion]*[Global]
  real64_t r00=mtx[MTX_R00], r01=mtx[MTX_R01], r02=mtx[MTX_R02];
  real64_t r10=mtx[MTX_R10], r11=mtx[MTX_R11], r12=mtx[MTX_R12];
  real64_t r20=mtx[MTX_R20], r21=mtx[MTX_R21], r22=mtx[MTX_R22];
  mtx[MTX_R00]=r00*mtx2[MTX_R00]+r01*mtx2[MTX_R10]+r02*mtx2[MTX_R20];
  mtx[MTX_R01]=r00*mtx2[MTX_R01]+r01*mtx2[MTX_R11]+r02*mtx2[MTX_R21];
  mtx[MTX_R02]=r00*mtx2[MTX_R02]+r01*mtx2[MTX_R12]+r02*mtx2[MTX_R22];
  mtx[MTX_TX]+=r00*mtx2[MTX_TX]+r01*mtx2[MTX_TY]+r02*mtx2[MTX_TZ];
  mtx[MTX_R10]=r10*mtx2[MTX_R00]+r11*mtx2[MTX_R10]+r12*mtx2[MTX_R20];
  mtx[MTX_R11]=r10*mtx2[MTX_R01]+r11*mtx2[MTX_R11]+r12*mtx2[MTX_R21];
  mtx[MTX_R12]=r10*mtx2[MTX_R02]+r11*mtx2[MTX_R12]+r12*mtx2[MTX_R22];
  mtx[MTX_TY]+=r10*mtx2[MTX_TX]+r11*mtx2[MTX_TY]+r12*mtx2[MTX_TZ];
  mtx[MTX_R20]=r20*mtx2[MTX_R00]+r21*mtx2[MTX_R10]+r22*mtx2[MTX_R20];
  mtx[MTX_R21]=r20*mtx2[MTX_R01]+r21*mtx2[MTX_R11]+r22*mtx2[MTX_R21];
  mtx[MTX_R22]=r20*mtx2[MTX_R02]+r21*mtx2[MTX_R12]+r22*mtx2[MTX_R22];
  mtx[MTX_TZ]+=r20*mtx2[MTX_TX]+r21*mtx2[MTX_TY]+r22*mtx2[MTX_TZ];
  }
}

void
SpatialTransform_preRotate(SpatialTransform *st,
                           real64_t x,
                           real64_t y,
                           real64_t z,
                           real64_t angle)
{
real64_t r=sqrt(x*x+y*y+z*z);
if(r)
  {
  r=1.0/r;
  x*=r;
  y*=r;
  z*=r;
  real64_t s,c;          //  tx2+c   txy-sz  txz+sy
  sincos64(angle,&s,&c); //  txy+sz  ty2+c   tyz-sx
  real64_t t=1.0-c;      //  txz-sy  tyz+sx  tz2+c
  if(st->identity)
    {
    st->identity=false;
    real64_t *mtx=st->matrix;
    mtx[MTX_R00]=t*x*x+c;
    mtx[MTX_R01]=t*x*y-s*z;
    mtx[MTX_R02]=t*x*z+s*y;
    mtx[MTX_R10]=t*x*y+s*z;
    mtx[MTX_R11]=t*y*y+c;
    mtx[MTX_R12]=t*y*z-s*x;
    mtx[MTX_R20]=t*x*z-s*y;
    mtx[MTX_R21]=t*y*z+s*x;
    mtx[MTX_R22]=t*z*z+c;
    }
  else
    {
    SpatialTransform rot; // initialized right after
    rot.identity=false;
    rot.matrix[MTX_R00]=t*x*x+c;
    rot.matrix[MTX_R01]=t*x*y-s*z;
    rot.matrix[MTX_R02]=t*x*z+s*y;
    rot.matrix[MTX_R10]=t*x*y+s*z;
    rot.matrix[MTX_R11]=t*y*y+c;
    rot.matrix[MTX_R12]=t*y*z-s*x;
    rot.matrix[MTX_R20]=t*x*z-s*y;
    rot.matrix[MTX_R21]=t*y*z+s*x;
    rot.matrix[MTX_R22]=t*z*z+c;
    rot.matrix[MTX_TX]=rot.matrix[MTX_TY]=rot.matrix[MTX_TZ]=0.0;
    // rot.matrix[MTX_Z0]=rot.matrix[MTX_Z1]=rot.matrix[MTX_Z2]=0.0;
    // rot.matrix[MTX_ONE]=1.0;
    SpatialTransform_preTransform(st,&rot);
    }
  }
}

void
SpatialTransform_postRotate(SpatialTransform *st,
                            real64_t x,
                            real64_t y,
                            real64_t z,
                            real64_t angle)
{
real64_t r=sqrt(x*x+y*y+z*z);
if(r)
  {
  r=1.0/r;
  x*=r;
  y*=r;
  z*=r;
  real64_t s,c;          //  tx2+c   txy-sz  txz+sy
  sincos64(angle,&s,&c); //  txy+sz  ty2+c   tyz-sx
  real64_t t=1.0-c;      //  txz-sy  tyz+sx  tz2+c
  real64_t *mtx=st->matrix;
  if(st->identity)
    {
    st->identity=false;
    mtx[MTX_R00]=t*x*x+c;
    mtx[MTX_R01]=t*x*y-s*z;
    mtx[MTX_R02]=t*x*z+s*y;
    mtx[MTX_R10]=t*x*y+s*z;
    mtx[MTX_R11]=t*y*y+c;
    mtx[MTX_R12]=t*y*z-s*x;
    mtx[MTX_R20]=t*x*z-s*y;
    mtx[MTX_R21]=t*y*z+s*x;
    mtx[MTX_R22]=t*z*z+c;
    }
  else
    {
    real64_t l00=t*x*x+c,    l01=t*x*y-s*z,  l02=t*x*z+s*y;
    real64_t l10=t*x*y+s*z,  l11=t*y*y+c,    l12=t*y*z-s*x;
    real64_t l20=t*x*z-s*y,  l21=t*y*z+s*x,  l22=t*z*z+c;
    real64_t g00=mtx[MTX_R00],  g01=mtx[MTX_R01],  g02=mtx[MTX_R02];
    real64_t g10=mtx[MTX_R10],  g11=mtx[MTX_R11],  g12=mtx[MTX_R12];
    real64_t g20=mtx[MTX_R20],  g21=mtx[MTX_R21],  g22=mtx[MTX_R22];
    mtx[MTX_R00]=g00*l00+g01*l10+g02*l20;  // [Global]*[Local]
    mtx[MTX_R01]=g00*l01+g01*l11+g02*l21;
    mtx[MTX_R02]=g00*l02+g01*l12+g02*l22;
    mtx[MTX_R10]=g10*l00+g11*l10+g12*l20;
    mtx[MTX_R11]=g10*l01+g11*l11+g12*l21;
    mtx[MTX_R12]=g10*l02+g11*l12+g12*l22;
    mtx[MTX_R20]=g20*l00+g21*l10+g22*l20;
    mtx[MTX_R21]=g20*l01+g21*l11+g22*l21;
    mtx[MTX_R22]=g20*l02+g21*l12+g22*l22;
    }
  }
}

void
SpatialTransform_preRoll(SpatialTransform *st, //  1  0  0
                         real64_t roll)        //  0  c -s
{                                              //  0  s  c
real64_t s,c;
sincos64(roll,&s,&c);
if(st->identity)
  {
  st->identity=false;
  real64_t *mtx=st->matrix;
  mtx[MTX_R11]=c;
  mtx[MTX_R12]=-s;
  mtx[MTX_R21]=s;
  mtx[MTX_R22]=c;
  }
else
  {
  SpatialTransform rot; // initialized right after
  rot.identity=false;
  rot.matrix[MTX_R11]=c;
  rot.matrix[MTX_R12]=-s;
  rot.matrix[MTX_R21]=s;
  rot.matrix[MTX_R22]=c;
  rot.matrix[MTX_R00]=1.0;
  rot.matrix[MTX_R10]=rot.matrix[MTX_R20]=
  rot.matrix[MTX_R01]=rot.matrix[MTX_R02]=
  rot.matrix[MTX_TX]=rot.matrix[MTX_TY]=rot.matrix[MTX_TZ]=0.0;
  // rot.matrix[MTX_Z0]=rot.matrix[MTX_Z1]=rot.matrix[MTX_Z2]=0.0;
  // rot.matrix[MTX_ONE]=1.0;
  SpatialTransform_preTransform(st,&rot);
  }
}

void
SpatialTransform_postRoll(SpatialTransform *st, //  1  0  0
                          real64_t roll)        //  0  c -s
{                                               //  0  s  c
real64_t s,c;
sincos64(roll,&s,&c);
real64_t *mtx=st->matrix;
if(st->identity)
  {
  st->identity=false;
  mtx[MTX_R11]=c;
  mtx[MTX_R12]=-s;
  mtx[MTX_R21]=s;
  mtx[MTX_R22]=c;
  }
else
  {
  real64_t r01=+c*mtx[MTX_R01]+s*mtx[MTX_R02];  // [Global]*[Local]
  real64_t r02=-s*mtx[MTX_R01]+c*mtx[MTX_R02];
  real64_t r11=+c*mtx[MTX_R11]+s*mtx[MTX_R12];
  real64_t r12=-s*mtx[MTX_R11]+c*mtx[MTX_R12];
  real64_t r21=+c*mtx[MTX_R21]+s*mtx[MTX_R22];
  real64_t r22=-s*mtx[MTX_R21]+c*mtx[MTX_R22];
  mtx[MTX_R01]=r01;
  mtx[MTX_R02]=r02;
  mtx[MTX_R11]=r11;
  mtx[MTX_R12]=r12;
  mtx[MTX_R21]=r21;
  mtx[MTX_R22]=r22;
  }
}

void
SpatialTransform_prePitch(SpatialTransform *st, //  c  0  s
                          real64_t pitch)       //  0  1  0
{                                               // -s  0  c
real64_t s,c;
sincos64(pitch,&s,&c);
if(st->identity)
  {
  st->identity=false;
  real64_t *mtx=st->matrix;
  mtx[MTX_R22]=c;
  mtx[MTX_R20]=-s;
  mtx[MTX_R02]=s;
  mtx[MTX_R00]=c;
  }
else
  {
  SpatialTransform rot; // initialized right after
  rot.identity=false;
  rot.matrix[MTX_R22]=c;
  rot.matrix[MTX_R20]=-s;
  rot.matrix[MTX_R02]=s;
  rot.matrix[MTX_R00]=c;
  rot.matrix[MTX_R11]=1.0;
  rot.matrix[MTX_R10]=rot.matrix[MTX_R01]=
  rot.matrix[MTX_R21]=rot.matrix[MTX_R12]=
  rot.matrix[MTX_TX]=rot.matrix[MTX_TY]=rot.matrix[MTX_TZ]=0.0;
  // rot.matrix[MTX_Z0]=rot.matrix[MTX_Z1]=rot.matrix[MTX_Z2]=0.0;
  // rot.matrix[MTX_ONE]=1.0;
  SpatialTransform_preTransform(st,&rot);
  }
}

void
SpatialTransform_postPitch(SpatialTransform *st, //  c  0  s
                           real64_t pitch)       //  0  1  0
{                                                // -s  0  c
real64_t s,c;
sincos64(pitch,&s,&c);
real64_t *mtx=st->matrix;
if(st->identity)
  {
  st->identity=false;
  mtx[MTX_R22]=c;
  mtx[MTX_R20]=-s;
  mtx[MTX_R02]=s;
  mtx[MTX_R00]=c;
  }
else
  {
  real64_t r00=+c*mtx[MTX_R00]-s*mtx[MTX_R02];  // [Global]*[Local]
  real64_t r02=+s*mtx[MTX_R00]+c*mtx[MTX_R02];
  real64_t r10=+c*mtx[MTX_R10]-s*mtx[MTX_R12];
  real64_t r12=+s*mtx[MTX_R10]+c*mtx[MTX_R12];
  real64_t r20=+c*mtx[MTX_R20]-s*mtx[MTX_R22];
  real64_t r22=+s*mtx[MTX_R20]+c*mtx[MTX_R22];
  mtx[MTX_R00]=r00;
  mtx[MTX_R02]=r02;
  mtx[MTX_R10]=r10;
  mtx[MTX_R12]=r12;
  mtx[MTX_R20]=r20;
  mtx[MTX_R22]=r22;
  }
}

void
SpatialTransform_preYaw(SpatialTransform *st, //  c -s  0
                        real64_t yaw)         //  s  c  0
{                                             //  0  0  1
real64_t s,c;
sincos64(yaw,&s,&c);
if(st->identity)
  {
  st->identity=false;
  real64_t *mtx=st->matrix;
  mtx[MTX_R00]=c;
  mtx[MTX_R01]=-s;
  mtx[MTX_R10]=s;
  mtx[MTX_R11]=c;
  }
else
  {
  SpatialTransform rot; // initialized right after
  rot.identity=false;
  rot.matrix[MTX_R00]=c;
  rot.matrix[MTX_R01]=-s;
  rot.matrix[MTX_R10]=s;
  rot.matrix[MTX_R11]=c;
  rot.matrix[MTX_R22]=1.0;
  rot.matrix[MTX_R20]=rot.matrix[MTX_R21]=
  rot.matrix[MTX_R02]=rot.matrix[MTX_R12]=
  rot.matrix[MTX_TX]=rot.matrix[MTX_TY]=rot.matrix[MTX_TZ]=0.0;
  // rot.matrix[MTX_Z0]=rot.matrix[MTX_Z1]=rot.matrix[MTX_Z2]=0.0;
  // rot.matrix[MTX_ONE]=1.0;
  SpatialTransform_preTransform(st,&rot);
  }
}

void
SpatialTransform_postYaw(SpatialTransform *st, //  c -s  0
                         real64_t yaw)         //  s  c  0
{                                              //  0  0  1
real64_t s,c;
sincos64(yaw,&s,&c);
real64_t *mtx=st->matrix;
if(st->identity)
  {
  st->identity=false;
  mtx[MTX_R00]=c;
  mtx[MTX_R01]=-s;
  mtx[MTX_R10]=s;
  mtx[MTX_R11]=c;
  }
else
  {
  real64_t r00=+c*mtx[MTX_R00]+s*mtx[MTX_R01];  // [Global]*[Local]
  real64_t r01=-s*mtx[MTX_R00]+c*mtx[MTX_R01];
  real64_t r10=+c*mtx[MTX_R10]+s*mtx[MTX_R11];
  real64_t r11=-s*mtx[MTX_R10]+c*mtx[MTX_R11];
  real64_t r20=+c*mtx[MTX_R20]+s*mtx[MTX_R21];
  real64_t r21=-s*mtx[MTX_R20]+c*mtx[MTX_R21];
  mtx[MTX_R00]=r00;
  mtx[MTX_R01]=r01;
  mtx[MTX_R10]=r10;
  mtx[MTX_R11]=r11;
  mtx[MTX_R20]=r20;
  mtx[MTX_R21]=r21;
  }
}

//----------------------------------------------------------------------------

void
BoundingBox_addMovedBox(BoundingBox *bb,
                        const SpatialBasis *motion,
                        const BoundingBox *other)
{
if(BoundingBox_valid(other))
  {
  real64_t px[8],py[8],pz[8];
  px[0]=other->xmin; py[0]=other->ymin; pz[0]=other->zmin;
  px[1]=other->xmax; py[1]=other->ymin; pz[1]=other->zmin;
  px[2]=other->xmax; py[2]=other->ymax; pz[2]=other->zmin;
  px[6]=other->xmax; py[6]=other->ymax; pz[6]=other->zmax;
  SpatialBasis_localToGlobalPosition(motion,px+0,py+0,pz+0);
  SpatialBasis_localToGlobalPosition(motion,px+1,py+1,pz+1);
  SpatialBasis_localToGlobalPosition(motion,px+2,py+2,pz+2);
  SpatialBasis_localToGlobalPosition(motion,px+6,py+6,pz+6);
  real64_t vx[3],vy[3],vz[3];
  vx[0]=px[1]-px[0]; vy[0]=py[1]-py[0]; vz[0]=pz[1]-pz[0];
  vx[1]=px[2]-px[1]; vy[1]=py[2]-py[1]; vz[1]=pz[2]-pz[1];
  vx[2]=px[6]-px[2]; vy[2]=py[6]-py[2]; vz[2]=pz[6]-pz[2];
  px[3]=px[0]+vx[1]; py[3]=py[0]+vy[1]; pz[3]=pz[0]+vz[1];
  px[4]=px[0]+vx[2]; py[4]=py[0]+vy[2]; pz[4]=pz[0]+vz[2];
  px[5]=px[1]+vx[2]; py[5]=py[1]+vy[2]; pz[5]=pz[1]+vz[2];
  px[7]=px[6]-vx[0]; py[7]=py[6]-vy[0]; pz[7]=pz[6]-vz[0];
  BoundingBox_addVertex(bb,px[0],py[0],pz[0]);
  BoundingBox_addVertex(bb,px[1],py[1],pz[1]);
  BoundingBox_addVertex(bb,px[2],py[2],pz[2]);
  BoundingBox_addVertex(bb,px[3],py[3],pz[3]);
  BoundingBox_addVertex(bb,px[4],py[4],pz[4]);
  BoundingBox_addVertex(bb,px[5],py[5],pz[5]);
  BoundingBox_addVertex(bb,px[6],py[6],pz[6]);
  BoundingBox_addVertex(bb,px[7],py[7],pz[7]);
  }
}

void
BoundingBox_addTransformedBox(BoundingBox *bb,
                              const SpatialTransform *transform,
                              const BoundingBox *other)
{
if(BoundingBox_valid(other))
  {
  real64_t px[8],py[8],pz[8];
  px[0]=other->xmin; py[0]=other->ymin; pz[0]=other->zmin;
  px[1]=other->xmax; py[1]=other->ymin; pz[1]=other->zmin;
  px[2]=other->xmax; py[2]=other->ymax; pz[2]=other->zmin;
  px[6]=other->xmax; py[6]=other->ymax; pz[6]=other->zmax;
  SpatialTransform_applyToPosition(transform,px+0,py+0,pz+0);
  SpatialTransform_applyToPosition(transform,px+1,py+1,pz+1);
  SpatialTransform_applyToPosition(transform,px+2,py+2,pz+2);
  SpatialTransform_applyToPosition(transform,px+6,py+6,pz+6);
  real64_t vx[3],vy[3],vz[3];
  vx[0]=px[1]-px[0]; vy[0]=py[1]-py[0]; vz[0]=pz[1]-pz[0];
  vx[1]=px[2]-px[1]; vy[1]=py[2]-py[1]; vz[1]=pz[2]-pz[1];
  vx[2]=px[6]-px[2]; vy[2]=py[6]-py[2]; vz[2]=pz[6]-pz[2];
  px[3]=px[0]+vx[1]; py[3]=py[0]+vy[1]; pz[3]=pz[0]+vz[1];
  px[4]=px[0]+vx[2]; py[4]=py[0]+vy[2]; pz[4]=pz[0]+vz[2];
  px[5]=px[1]+vx[2]; py[5]=py[1]+vy[2]; pz[5]=pz[1]+vz[2];
  px[7]=px[6]-vx[0]; py[7]=py[6]-vy[0]; pz[7]=pz[6]-vz[0];
  BoundingBox_addVertex(bb,px[0],py[0],pz[0]);
  BoundingBox_addVertex(bb,px[1],py[1],pz[1]);
  BoundingBox_addVertex(bb,px[2],py[2],pz[2]);
  BoundingBox_addVertex(bb,px[3],py[3],pz[3]);
  BoundingBox_addVertex(bb,px[4],py[4],pz[4]);
  BoundingBox_addVertex(bb,px[5],py[5],pz[5]);
  BoundingBox_addVertex(bb,px[6],py[6],pz[6]);
  BoundingBox_addVertex(bb,px[7],py[7],pz[7]);
  }
}

//----------------------------------------------------------------------------

bool // bounding box is visible
Frustum_boxVisibility(const Frustum *f,
                      const SpatialBasis *viewPoint,
                      const BoundingBox *box,
                      const SpatialBasis *location,
                      const SpatialTransform *transform)
{
if(BoundingBox_valid(box))
  {
  int outFront=8;
  int outBack=8;
  int outLeft=8;
  int outRight=8;
  int outUp=8;
  int outDown=8;
  for(int i=0;i<8;++i)
    {
    real64_t x=(i&1 ? box->xmax : box->xmin);
    real64_t y=(i&2 ? box->ymax : box->ymin);
    real64_t z=(i&4 ? box->zmax : box->zmin);
    if(transform) SpatialTransform_applyToPosition(transform,&x,&y,&z);
    if(location) SpatialBasis_localToGlobalPosition(location,&x,&y,&z);
    if(viewPoint) SpatialBasis_globalToLocalPosition(viewPoint,&x,&y,&z);
    VertexLocation loc=Frustum_locateVertex(f,x,y,z);
    if(loc==LOC_MIDDLE)   return true; // inside perspective
    if(loc&LOC_XPOS)      --outFront;
    else if(loc&LOC_XNEG) --outBack;
    if(loc&LOC_YPOS)      --outLeft;
    if(loc&LOC_YNEG)      --outRight;
    if(loc&LOC_ZPOS)      --outUp;
    if(loc&LOC_ZNEG)      --outDown;
    }
  return outFront&&outBack&&
         outLeft&&outRight&&
         outUp&&outDown; // not on the same side
  }
else
  {
  return false;
  }
}

//----------------------------------------------------------------------------

INLINE
void
Perspective_computeSlopes(Perspective *p)
{
real64_t slope=tan(0.5*p->upDownFieldOfView);
p->frustum.upSlope=p->upDownSlopeOffset+slope;
p->frustum.downSlope=p->upDownSlopeOffset-slope;
slope*=p->aspectRatio;
p->frustum.leftSlope=p->leftRightSlopeOffset+slope;
p->frustum.rightSlope=p->leftRightSlopeOffset-slope;
}

void
Perspective_init(Perspective *p)
{
p->frustum.nearDistance=0.1;
p->frustum.farDistance=100.0;
p->upDownFieldOfView=M_PI/3.0;
p->aspectRatio=1.0;
p->leftRightSlopeOffset=0.0;
p->upDownSlopeOffset=0.0;
p->stereoEyeSeparation=p->frustum.nearDistance;
p->stereoViewDistance=p->frustum.nearDistance+
                      0.25*(p->frustum.farDistance-p->frustum.nearDistance);
Perspective_computeSlopes(p);
}

void
Perspective_setUpDownFieldOfView(Perspective *p,
                                 real64_t upDownFieldOfView)
{
p->upDownFieldOfView=upDownFieldOfView;
Perspective_computeSlopes(p);
}

void
Perspective_setAspectRatio(Perspective *p,
                           real64_t aspectRatio)
{
p->aspectRatio=aspectRatio;
Perspective_computeSlopes(p);
}

void
Perspective_setLeftRightSlopeOffset(Perspective *p,
                                    real64_t leftRightSlopeOffset)
{
p->leftRightSlopeOffset=leftRightSlopeOffset;
Perspective_computeSlopes(p);
}

void
Perspective_setUpDownSlopeOffset(Perspective *p,
                                 real64_t upDownSlopeOffset)
{
p->upDownSlopeOffset=upDownSlopeOffset;
Perspective_computeSlopes(p);
}

//----------------------------------------------------------------------------
